/*-----------------------------------------------------------------------------

 PROJECT:  SIMULATION GENERATION FRAMEWORK (SGFramework)

 SUBSYSTEM:  sgxlat.exe
 FILE:	     actions.h
 AUTHOR:     Kevin M. Kramer

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 02/18/94    KMK      initial implementation
 04/13/94    KMK      conversion to SGL syntax
 10/20/94    KMK      modified for use in command line version

 DESCRIPTION:

 This module contains the user action type definitions.

-----------------------------------------------------------------------------*/


#ifndef __actions_h
#define __actions_h

#ifndef __tokens_h
#include "tokens.h"
#endif


#define ACT_	      0x0000
#define ACT_DROP      0x0001
#define ACT_PUSH      0x0002
#define ACT_POP	      0x0003
#define ACT_CALL      0x0004
#define ACT_REM1      0x0005
#define ACT_REM2      0x0006
#define ACT_NEW	      0x0007

#define ACT_EXPR      TOK_EXPR
#define ACT_LEXPR     TOK_LEXPR
#define ACT_AEXPR     TOK_AEXPR
#define ACT_MEXPR     TOK_MEXPR
#define ACT_UEXPR     TOK_UEXPR
#define ACT_EXPRS     TOK_EXPRS
#define ACT_INDEX     TOK_INDEX
#define ACT_INDICES   TOK_INDICES
#define ACT_RANGES    TOK_RANGES
#define ACT_KITEM     TOK_KITEM
#define ACT_MSTAT     TOK_MSTAT
#define ACT_MSTATS    TOK_MSTATS
#define ACT_FARG      TOK_FARG
#define ACT_FARGS     TOK_FARGS


#define USR_	      0x0000
#define USR_INIT      0x0001
#define USR_PROG      0x0002
#define USR_MAIN      0x0003
#define USR_RCONST    0x0004
#define USR_ICONST    0x0005
#define USR_VECTOR    0x0006
#define USR_ID	      0x0007
#define USR_UNOP1     0x0008
#define USR_UNOP2     0x0009
#define USR_UNOP2A    0x000a
#define USR_BINOP1    0x000b
#define USR_BINOP2    0x000c
#define USR_BINOP2A   0x000d
#define USR_NSUM      0x000e
#define USR_LSUM      0x000f
#define USR_REAL      0x0010
#define USR_FUNC      0x0011
#define USR_ARRAY     0x0012
#define USR_EXPRS     0x0013
#define USR_ADDEXPR   0x0014
#define USR_PAREN     0x0015
#define USR_INDEX1    0x0016
#define USR_INDEX2    0x0017
#define USR_INDEX3    0x0018
#define USR_INDEX4    0x0019
#define USR_INDEXL    0x001a
#define USR_INDICES   0x001b
#define USR_ADDINDX   0x001c
#define USR_DVAR      0x001d
#define USR_DARRAY    0x001e
#define USR_DCONST    0x001f
#define USR_DINDEX    0x0020
#define USR_DRANGE    0x0021
#define USR_KNOWN1    0x0022
#define USR_KNOWN2    0x0023
#define USR_UNKNO1    0x0024
#define USR_UNKNO2    0x0025
#define USR_EQN1      0x0026
#define USR_EQN2      0x0027
#define USR_SOLVE     0x0028
#define USR_WRITE     0x0029
#define USR_ASS1      0x002a
#define USR_ASS2      0x002b
#define USR_WHILE1    0x002c
#define USR_WHILE2    0x002d
#define USR_DO1	      0x002e
#define USR_DO2	      0x002f
#define USR_IF1	      0x0030
#define USR_IF2	      0x0031
#define USR_IF3	      0x0032
#define USR_IF4	      0x0033
#define USR_SET	      0x0034
#define USR_FARG1     0x0035
#define USR_FARG2     0x0036
#define USR_DECFUNC   0x0037
#define USR_DLCONST   0x0038
#define USR_LASS      0x0039
#define USR_RETURN    0x003a
#define USR_OPEN      0x003b
#define USR_CLOSE     0x003c
#define USR_VREAD     0x003d
#define USR_VWRITE    0x003e
#define USR_IO	      0x003f
#define USR_GOTO      0x0040
#define USR_LABEL     0x0041
#define USR_COMMENT   0x0042
#define USR_EXIT      0x0043
#define USR_PROC1     0x0044
#define USR_PROC2     0x0045
#define USR_CALL      0x0046
#define USR_MESH      0x0047
#define USR_LABTEST   0x0048
#define USR_DECEXT    0x0049
#define USR_REF	      0x004a
#define USR_PCNODE    0x004b
#define USR_PCEDGE    0x004c
#define USR_PRECOMP   0x004d
#define USR_ELEN      0x004e
#define USR_EM1	      0x004f
#define USR_EM2	      0x0050

#endif
