#include <stdio.h>
#include <fcntl.h>

#if defined(UNIX)
#include <unistd.h>
#include <sys/stat.h>
#define	 O_BINARY 0x0000
#else
#include <io.h>
#include <sys\stat.h>
#endif


typedef unsigned int UINT;
typedef int	     BOOL;
typedef int	     I3ARRAY[3];

typedef struct tagTOPHEAD
{ char szLogo[64];			// text logo
  char szSign[16];			// signature
  UINT cElement;			// number of variables & array elements
  UINT cTag;				// number of unknown elements
  UINT cTopology;			// number of nonzeroes excluding pivots
  UINT cEquation;			// number of equations
} TOPHEAD;

typedef struct tagTAG			// tag
{ UINT iVar;				// element array index (i.e. offset)
  UINT iEqu;				// equation index
  UINT iArray;				// array index ... -1u == variable
} TAG;

typedef struct tagTOP			// top
{ UINT i;				// index in topology array
  UINT c;				// number of nonzeroes in equation
} TOP;

#define TRUE  1
#define FALSE 0

UINT cBlock;
UINT *aiOrder;
UINT *auBlock;
BOOL *afBlock;

UINT *aiTop;
UINT *aiTag;
TAG *atag;
TOP *atop;
I3ARRAY *aanEI;

void Mark (UINT i)
{
  UINT j;
  afBlock[i] = TRUE;
  auBlock[i] = cBlock;
  for(j = atop[i].i; j < atop[i].c + atop[i].i; ++j)
    if (!afBlock[aiTop[j]]) Mark(aiTop[j]);
}


int main(int cArg, char *apszArg[])
{
  fprintf(stderr, "Block Perm Vector	   Version 1.0	Copyright (c) 1995 Kevin M. Kramer\n");
  if (cArg != 2)
  { fprintf(stderr, "usage:  blkperm top_file <in_perm >out_perm\n");
    return 1;
  }

  int nFile = open(apszArg[1], O_BINARY | O_RDONLY);
  if (nFile == -1)
  { fprintf(stderr, "cannot open file \"%s\"\n", apszArg[1]);
    return 2;
  }

  TOPHEAD  head;
  read(nFile, &head, sizeof(TOPHEAD));
  aiTop = new UINT[head.cTopology];
  aiTag = new UINT[head.cElement];
  atag = new TAG[head.cTag];
  atop = new TOP[head.cTag];
  aanEI = new I3ARRAY[head.cTag];
  read(nFile, aiTag, head.cElement  * sizeof(UINT));
  read(nFile, aiTop, head.cTopology * sizeof(UINT));
  read(nFile, atag,  head.cTag * sizeof(TAG));
  read(nFile, atop,  head.cTag * sizeof(TOP));
  read(nFile, aanEI, head.cTag * sizeof(I3ARRAY));
  aiOrder = new UINT[head.cTag];
  auBlock = new UINT[head.cTag];
  afBlock = new BOOL[head.cTag];
  cBlock = 0;

  close(nFile);

  UINT i,j;
  for(i = 0; i < head.cTag; ++i)
    afBlock[i] = FALSE;

  for(i = 0; i < head.cTag; ++i)
    if (!afBlock[i]) { Mark(i); ++cBlock; }

  UINT c;
  fscanf(stdin, "%u\n", &c);
  if (c != head.cTag)
  { fprintf(stderr, "size of perm vector and topology file do not agree!\n");
    return 2;
  }

  for(i = 0; i < head.cTag; ++i)
    fscanf(stdin, "%u", aiOrder + i);

  fprintf(stdout, "%u\n", head.cTag);
  for(i = 0; i < cBlock; ++i)
    for(j = 0; j < head.cTag; ++j)
      if (auBlock[aiOrder[j]-1] == i)
	fprintf(stdout, "%u ", aiOrder[j]);

  delete [] aiTop;
  delete [] aiTag;
  delete [] atag;
  delete [] atop;
  delete [] aanEI;
  delete [] aiOrder;
  delete [] auBlock;
  delete [] afBlock;
  return 0;
}

