/*-----------------------------------------------------------------------------

 PROJECT:  SIMULATION GENERATION FRAMEWORK (SGFramework)

 SUBSYSTEM:  sgxlat.exe
 FILE:	     codebuf.cpp
 AUTHOR:     Kevin M. Kramer
             J. Joshua Feng

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 09/10/94    KMK      initial implementation
 10/20/94    KMK      modified for use in command line version
 07/06/00    JJF      modify AddText

 DESCRIPTION:

 This module implements the code buffer class.

-----------------------------------------------------------------------------*/


#ifndef __codebuf_h
#define __codebuf_h

#define LEN_BUFFER 512
#define LEN_PREFIX 3

#ifndef __mathnode_h
#include "mathnode.h"
#endif


// code buffer class
class CodeBuffer
{ protected:
    char szFileName[LEN_FILENAME];	// file name
    char szUCPrefix[LEN_PREFIX + 1];	// uppercase prefix
    char szLCPrefix[LEN_PREFIX + 1];	// lowercase prefix
    char szBuffer[LEN_BUFFER];		// buffer
    int	 nFile;				// file handle
    UINT iLeftMargin;			// left margin
    UINT iRightMargin;			// right margin
    UINT cPrecision;			// number of significant digits
    UINT iBuffer;			// buffer index

  protected:
    void Flush ();			// flush the buffer
    void AddToken (char *);		// add token to buffer

  public:
    inline void SetRightMargin (UINT u) { iRightMargin = u; }
    void SetLeftMargin	(UINT);
    void AddChar (char);
    void AddIndex (UINT, BOOL = TRUE);
    void AddArgument (UINT, BOOL = TRUE);
    void AddInteger (int, BOOL = TRUE);
    void AddUInteger (UINT, BOOL = TRUE);
    void AddReal (REAL, BOOL = TRUE);
    void AddVariable (char *, BOOL, BOOL = TRUE);
    void AddVariable (UINT, BOOL, BOOL = TRUE);
    void AddArray (char *, BOOL, BOOL = FALSE);
    void AddArray (UINT, BOOL, BOOL = FALSE);
    void AddLocalVariable (UINT, BOOL = TRUE);
    void AddLocalDerivative (UINT, BOOL = TRUE);
    void AddFunction (char *, UINT, BOOL = FALSE);
    void AddFunction (UINT, UINT, BOOL = FALSE);
    void AddUDFunction (char *, UINT, BOOL = FALSE);
    void AddUDFunction (UINT, UINT, BOOL = FALSE);
    void AddPrecUDFunc (char *, BOOL = FALSE);
    void AddPrecUDFunc (UINT, BOOL = FALSE);
    void AddProcedure (char *, BOOL = TRUE);
    void AddProcedure (UINT, BOOL = TRUE);
    void AddLabel (UINT, BOOL, BOOL = TRUE);
    void AddString (char *);
    void AddText (const char *);
    void AddEquation (UINT, BOOL = FALSE);
    void AddPartial (UINT, BOOL = FALSE);
    void AddSum (UINT, BOOL = FALSE);
    inline void AddMathNode (MathNode *pmn, BOOL fTag = FALSE, BOOL fPrecompute = FALSE)
    { pmn->WriteCode((void *) this, fTag, fPrecompute); }

  public:
    CodeBuffer (char *, char *, UINT = 7, UINT = 0, UINT = 78);
   ~CodeBuffer ();
};


#endif
