/*-----------------------------------------------------------------------------

 PROJECT:  SIMULATION GENERATION FRAMEWORK (SGFramework)

 FILE:	     error.h
 AUTHOR:     Kevin M. Kramer
             J. Joshua Feng
 
 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 02/25/94    KMK      initial implementation
 07/06/00    JJF      modify AddMessage function

 DESCRIPTION:

 This module implements the SGFramework error handler.

-----------------------------------------------------------------------------*/


#ifndef __error_h
#define __error_h

class ErrorHandler
{ private:
    UINT cMaxErrors;			// maximum number of errors
    UINT cMaxWarnings;			// maximum number of warnings
    UINT cErrors;			// number of errors in message array
    UINT cWarnings;			// number of warnings in message array
    UINT cFatal;			// number of fatal errors
    UINT cTotal;			// total number of messages

    typedef struct tagMESSAGE		// msg
    { char *psz;			// pointer to message
      UINT  iLine;			// line index
      UINT  iCol;			// column index
    } MESSAGE;

    MESSAGE *amsg;			// message array
    void *p;				// corresponding source client
    UINT iBuf;				// buffer index

    UINT uMessagePrefix (UINT, UINT, UINT &);

  public:
    inline BOOL fFatal	    ()	     { return cFatal
                                           || (cErrors >= cMaxErrors)
                                           || (cWarnings >= cMaxWarnings); }
    inline BOOL fRunTime    ()	     { return cTotal; }
    inline BOOL fError	    ()	     { return (cErrors) ? TRUE : FALSE; }
    inline BOOL fWarning    ()	     { return (cWarnings) ? TRUE : FALSE; }
    inline UINT cNumMsg	    ()	     { return cErrors + cWarnings + cFatal; }
    inline void SetBufPos   (UINT i) { iBuf = i; }
    // inline void * GetSource ()	     { return p; }

    void Initialize (void *);
    void AddMessage (UINT,                             UINT = -1u);
    void AddMessage (UINT, const char *,               UINT = -1u);
    void AddMessage (UINT, const char *, const char *, UINT = -1u);
    void ShowMessages ();

  public:
     ErrorHandler (UINT = 25, UINT = 25);
    ~ErrorHandler ();
};


// global reference of error instance
extern ErrorHandler *error;


#define MSG_WARNING   0
#define MSG_ERROR     1
#define MSG_FATAL     2


typedef struct tagEHM			// ehm
{ UINT uType;				// type of message
  BOOL fLineInfo;			// include line information
  char *pszMsg;				// message text
} EHM;

extern EHM aehm[];

#define EHM_STRLF	  0		// string cannot span multiple lines
#define EHM_UNEXPEOF	  1		// unexpected end of file
#define EHM_SYNTAX	  2		// %s expected
#define EHM_SYMREDEF	  3		// redefinition of symbol "?"
#define EHM_SYMNOTDEF	  4		// symbol "?" not defined
#define EHM_SYMEXP	  5		// "?" is not a ? symbol
#define EHM_MAXVAR	  6		// max number of variables exceeded
#define EHM_MAXDENSITY	  7		// max Jacobian matrix density exceeded
#define EHM_NOTDEC1	  8		// ? symbol "?" not declared
#define EHM_NOTDEC2	  9		// ? statement not declared
#define EHM_NOTDEC3	  10		// function "?" not declared
#define EHM_CONST	  11		// "?" does not evaluate to a constant
#define EHM_ARRAYDIM1	  12		// array "?" has too many dimensions
#define EHM_ARRAYDIM2	  13		// dim ? of arr "?" does not eval to IC
#define EHM_ARRAYDIM3	  14		// dim ? of arr "?" is not an int expr
#define EHM_ARRAYDIM4	  15		// dim ? of arr "?" must be > 0
#define EHM_ARRAYDIM5	  16		// dim ? of arr "?" is out of bounds
#define EHM_ARRAYDIM6	  17		// array "?" has wrong # of dimensions
#define EHM_NUMARGS	  18		// func "?" has the wrong # of args
#define EHM_INDEX1	  19		// ? value of idx ? does not eval to IC
#define EHM_INDEX2	  20		// ? value of idx ? is illegal
#define EHM_USELESS	  21		// equ corr. to ? "?" is useless
#define EHM_NOPIVOT	  22		// equ corr. to ? "?" has no pivot
#define EHM_NOMAINPROC	  23		// no main procedure statements
#define EHM_NOEQUATION	  24		// no equations specified
#define EHM_DIVBYZERO	  25		// division by zero in ?
#define EHM_DOMAIN	  26		// domain error in ? function
#define EHM_OVERFLOW	  27		// overflow in ? function
#define EHM_UNDERFLOW	  28		// underflow in ? function
#define EHM_INTOVER	  29		// integer overflow
#define EHM_BADMESHFILE	  30		// corrupt or missing mesh file
#define EHM_NEARSINGULAR  31		// near singularity encountered
#define EHM_NOEQU	  32		// unknown "?" has no corresponding equ
#define EHM_UNDERSPEC	  33		// underspecified system of equations
#define EHM_OVERRELAX	  34		// overrelax param not in range 1 to 2
#define EHM_SYMNOTALLOWED 35		// sym "?" not allowed in UD func body
#define EHM_RECURSION	  36		// function "?" may not call itself
#define EHM_FIOSTATE1	  37		// a file must be open for ? prior ...
#define EHM_FIOSTATE2	  38		// file already open, must issue a ...
#define EHM_FILEOPEN	  39		// cannot open file "?"
#define EHM_FILECLOSE	  40		// the data file has not been closed
#define EHM_FILEEOF	  41		// unexpected end of data file
#define EHM_FILESYN	  42		// syntax error in data file
#define EHM_FILEFIND	  43		// could not find label "?' in data file
#define EHM_MAINPROC	  44		// main procedure already declared
#define EHM_INTEXPR	  45		// ? is not an integer expression
#define EHM_SUMEMBED	  46		// ? statement cannot be embedded
#define EHM_TRIMESH	  47		// triangular mesh not declared
#define EHM_PRECOMP	  48		// the ? operator can ? be used in pc
#define EHM_ODDEVEN	  49		// odd or even expression?

#endif
