/*-----------------------------------------------------------------------------

 PROJECT:  SIMULATION GENERATION FRAMEWORK (SGFramework)

 SUBSYSTEM:  sgxlat.exe
 FILE:	     fileio.cpp
 AUTHOR:     Kevin M. Kramer

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 04/05/94    KMK      initial implementation
 10/20/94    KMK      modified for use in command line version

 DESCRIPTION:

 This module implements the FileInputOutput class.

-----------------------------------------------------------------------------*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "global.h"

#ifndef UNIX
#pragma hdrstop
#endif

#if defined(UNIX)
#include <unistd.h>
#include <sys/stat.h>
#else
#include <io.h>
#include <sys\stat.h>
#endif

#include <fcntl.h>
#include "statemnt.h"
#include "fileio.h"


///////////////////////////////////////////////////////////////////////////////
// Function:	 FileInputOutput::FileInputOutput
// Scope:	 Public, Constructor
FileInputOutput::FileInputOutput ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("fileio.cpp FileInputOutput::FileInputOutput\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

}


///////////////////////////////////////////////////////////////////////////////
// Function:	 FileInputOutput::FormFileName
// Scope:	 Public
// Description:	 This function forms the translate and result file name by
//		 modifying the source file name's extension.  The file name
//		 is formed in the global buffer.
void FileInputOutput::FormFileName (char *pszSource, UINT uFileType, BOOL fNoPath)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("fileio.cpp FileInputOutput::FormFileName\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  // copy the source file name into the global buffer.
  char *psz = (fNoPath) ? psz = strrchr(pszSource, '\\') : NULL;
  if (psz) ++psz;
  else psz = pszSource;
  strcpy(pszGlobalBuf, psz);

  // form the file name in the global buffer
  if (uFileType != FILE_SOURCE)
  { char *psz = strchr(pszGlobalBuf, '.');
    if (psz) *psz = '\0';
    switch (uFileType)
    { case FILE_XLAT:	 strcat(pszGlobalBuf, EXT_XLAT);     break;
      case FILE_RESULT:	 strcat(pszGlobalBuf, EXT_RESULT);   break;
      case FILE_HEADER:	 strcat(pszGlobalBuf, EXT_HEADER);   break;
      case FILE_CODE:	 strcat(pszGlobalBuf, EXT_CODE);     break;
      case FILE_PERM:	 strcat(pszGlobalBuf, EXT_PERM);     break;
      case FILE_TOP:	 strcat(pszGlobalBuf, EXT_TOP);	     break;
    }
  }
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 FileInputOutput::nWriteTopFile
// Scope:	 Public
// Description:	 This function writes the topology file.
int FileInputOutput::nWriteTopFile (char *pszSource)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("fileio.cpp FileInputOutput::nWriteTopFile\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  ////////////////////////
  //  Open Output File	//
  ////////////////////////

  FormFileName(pszSource, FILE_TOP);
  nFile = open(pszGlobalBuf, O_CREAT | O_TRUNC | O_BINARY | O_WRONLY, S_IREAD | S_IWRITE);
  if (nFile == -1) return FILEIO_BADFILE;

  ///////////////////////////
  //  Write Topology File  //
  ///////////////////////////

  TOPHEAD head;
  strcpy(head.szLogo, "SGFramework Topology File  Version 1.0\n\x1a");
  strcpy(head.szSign, SIGNATURE_TOP);
  pstatements->WriteTopFile(nFile, &head);

  //////////////////////
  //  Close the File  //
  //////////////////////

  close(nFile);
  return FILEIO_SUCCESS;

}
