/*-----------------------------------------------------------------------------

 PROJECT:  SIMULATION GENERATION FRAMEWORK (SGFramework)

 SUBSYSTEM:  sgxlat.exe
 FILE:	     gencode.cpp
 AUTHOR:     Kevin M. Kramer

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 09/10/94    KMK      initial implementation
 10/20/94    KMK      modified for use in command line version

 DESCRIPTION:

 This module implements the code generation functions of the FileInputOutput
 class.

-----------------------------------------------------------------------------*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "global.h"

#ifndef UNIX
#pragma hdrstop
#endif

#include "symbol.h"
#include "gencode.h"
#include "codebuf.h"
#include "fileio.h"
#include "mathnode.h"
#include "numalg.h"
#include "statemnt.h"


static UINT cSum;
static BOOL fTag;
static BOOL fLabelSum;
static CodeBuffer *pcb;


///////////////////////////////////////////////////////////////////////////////
// Function:	 MathNode::MarkSums
// Scope:	 Public
// Description:	 This function marks the sums in mathnode expressions.
void MathNode::MarkSums ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("gencode.cpp MathNode::MarkSums\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;

  if	  (uNode == TYPE_LSUM) data.lsum.iExpr = ++cSum;
  else if (uNode == TYPE_NSUM) data.nsum.iExpr = ++cSum;

  if (uExpr & FLAG_SUM)
    for(i = 0; i < c; ++i)
      apmn[i]->MarkSums();
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 MathNode::WriteSums
// Scope:	 Public
// Description:	 This function writes C++ source code for each sum in the
//		 math node expression.
void MathNode::WriteSums ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("gencode.cpp MathNode::WriteSums\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;

  if	  (uNode == TYPE_LSUM)
  { pcb->SetLeftMargin(2);
    pcb->AddSum(data.lsum.iExpr, TRUE);
    pcb->AddText("= 0.0;\n");
    pcb->AddText("for(");
    pcb->AddIndex(data.lsum.iIndex, TRUE);
    pcb->AddText("= _label_init(0,");
    pcb->AddInteger(data.lsum.iLabel, FALSE);
    pcb->AddText("); ");
    pcb->AddIndex(data.lsum.iIndex, TRUE);
    pcb->AddText("!= -1u; ");
    pcb->AddIndex(data.lsum.iIndex, TRUE);
    pcb->AddText("= _label_next(0))\n{ ");
    pcb->SetLeftMargin(4);
    fLabelSum = TRUE;
    if (apmn[0]->uExpr & FLAG_NSUM) apmn[0]->WriteSums();
    fLabelSum = FALSE;
    pcb->SetLeftMargin(4);
    pcb->AddSum(data.lsum.iExpr, TRUE);
    pcb->AddText("+= ");
    pcb->SetLeftMargin(6);
    apmn[0]->WriteCode();
    pcb->AddText(";\n");
    pcb->SetLeftMargin(2);
    pcb->AddText("}\n\n");
    return;
  }
  else if (uNode == TYPE_NSUM)
  { pcb->SetLeftMargin((fLabelSum) ? 4 : 2);
    UINT iLabel = data.nsum.iLabel;
    pcb->AddSum(data.nsum.iExpr, TRUE);
    pcb->AddText("= 0.0;\n");
    pcb->AddText("for(_nsum_init(");
    if (iLabel < cRegLabel)
    { pcb->AddText("_bnode_map(");
      apmn[0]->WriteCode();
      pcb->AddChar(',');
      pcb->AddUInteger(iLabel, FALSE);
      pcb->AddChar(')');
      iLabel = -1u;
    }
    else apmn[0]->WriteCode();
    pcb->AddChar(',');
    if (iLabel == -1u) pcb->AddText("-1u");
    else pcb->AddUInteger(iLabel, FALSE);
    pcb->AddChar(',');
    pcb->AddText((data.nsum.fTriSum) ? "TRUE" : "FALSE");
    pcb->AddText("); !_nsum_loopover(); _nsum_next())\n{ ");
    pcb->AddIndex(data.nsum.iIndex, TRUE);
    pcb->AddText("= _nsum_index();\n");
    pcb->SetLeftMargin((fLabelSum) ? 6 : 4);
    pcb->AddSum(data.nsum.iExpr, TRUE);
    pcb->AddText("+= ");
    pcb->SetLeftMargin((fLabelSum) ? 8 : 6);
    apmn[1]->WriteCode();
    pcb->AddText(";\n");
    pcb->SetLeftMargin((fLabelSum) ? 4 : 2);
    pcb->AddText("}\n\n");
    return;
  }

  if (uExpr & FLAG_SUM)
    for(i = 0; i < c; ++i)
      apmn[i]->WriteSums();
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 MathNode::WriteSums
// Scope:	 Public
// Description:	 This function writes C++ source code for each sum in the
//		 math node expression.
void MathNode::WriteSums (void *pv, BOOL fTagStructure)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("gencode.cpp MathNode::WriteSums\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;

  // initialize global variables and mark the sums
  cSum = 0;
  fTag = fTagStructure;
  fLabelSum = FALSE;
  pcb  = (CodeBuffer *) pv;
  MarkSums();

  // declare the sum variables
  pcb->SetLeftMargin(2);
  for(i = 1; i <= cSum; ++i)
  { pcb->AddText("REAL ");
    pcb->AddSum(i, FALSE);
    pcb->AddText(";\n");
  }
  if (cSum) pcb->AddText("\n");

  // write the sums
  WriteSums();
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 WriteContribSums
// Scope:	 Public
// Description:	 This function writes C++ source code for each sum in the
//		 math node contribution expression.
void WriteContribSums (void *pv, UINT cContrib, MathNode **apmn)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("gencode.cpp WriteContribSums\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;

  // initialize global variables and mark the summations
  cSum = 0;
  fLabelSum = FALSE;
  pcb  = (CodeBuffer *) pv;
  for(i = 0; i < cContrib; ++i)
    apmn[i]->MarkSums();

  // declare the sum variables
  pcb->SetLeftMargin(2);
  for(i = 1; i <= cSum; ++i)
  { pcb->AddText("REAL ");
    pcb->AddSum(i, FALSE);
    pcb->AddText(";\n");
  }
  if (cSum) pcb->AddText("\n");

  // write the sums
  for(i = 0; i < cContrib; ++i)
    apmn[i]->WriteSums();
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Statements::DeclareProcedureSums
// Scope:	 Public
// Description:	 This function codes the procedure starting at the specified
//		 statement.
void Statements::DeclareProcedureSums(void *pv, UINT iFirst)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("gencode.cpp Statements::DeclareProcedureSums\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  // initialize global variables
  cSum = 0;
  fLabelSum = FALSE;
  pcb  = (CodeBuffer *) pv;

  // initialize local variables
  UINT cStatement = DAPMainProc.GetElements();
  BOOL fReturn;
  UINT i;
  Statement   *ps;
  VarAssign   *pva;
  ArrayAssign *paa;
  Goto	      *pgoto;


  // mark the summations
  for(i = iFirst, fReturn = FALSE; (i < cStatement) && !fReturn; ++i)
  { ps = DAPMainProc[i];
    switch (ps->uType)
    { // variable assignment statement
      case TYPE_VASSIGN:
	pva = TYPESAFE_DOWNCAST(ps, VarAssign);
	pva->pmn->MarkSums();
	break;

      // array assignment statement
      case TYPE_AASSIGN:
	paa = TYPESAFE_DOWNCAST(ps, ArrayAssign);
	paa->pmn->MarkSums();
	break;

      // conditional goto statement
      case TYPE_GOTO:
	pgoto = TYPESAFE_DOWNCAST(ps, Goto);
	if (pgoto->pmn) pgoto->pmn->MarkSums();
	break;

      // return statement
      case TYPE_RETURN:
	fReturn = TRUE;
	break;
    }
  }

  // declare the sum variables
  pcb->SetLeftMargin(2);
  for(i = 1; i <= cSum; ++i)
  { pcb->AddText("REAL ");
    pcb->AddSum(i, FALSE);
    pcb->AddText(";\n");
  }
  if (cSum) pcb->AddText("\n");

  // write the summations ... the sums for each statement needs to be
  // written after the statement's label
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 MathNode::WriteCode
// Scope:	 Public
// Description:	 This function writes C++ source code for the math node
//		 expression.
static BOOL fPrecompute;
void MathNode::WriteCode (void *pv, BOOL fTagStructure, BOOL fPreUDFunc)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("gencode.cpp MathNode::WriteCode\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  pcb = (CodeBuffer *) pv;
  fTag = fTagStructure;
  fPrecompute = fPreUDFunc;
  WriteCode();
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 MathNode::WriteCode
// Scope:	 Public
// Description:	 This function writes C++ source code for the math node
//		 expression.
void MathNode::WriteCode ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("gencode.cpp MathNode::WriteCode\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  static int iPD, i1, i2;
  UINT i;

  switch (uNode)
  { // buffer index math node
    case TYPE_INDEX:
      pcb->AddIndex(data.index.i, FALSE);
      break;

    // buffer variable math node
    case TYPE_VAR:
      pcb->AddVariable(data.var.i, fTag, FALSE);
      break;

    // buffer array math node
    case TYPE_ARRAY:
      pcb->AddArray(data.array.i, fTag, FALSE);
      for(i = 0; i < c; ++i)
      { pcb->AddChar('[');
	apmn[i]->WriteCode();
	pcb->AddChar(']');
      }
      break;

    // buffer real constant
    case TYPE_RCONST:
      pcb->AddReal(data.r, FALSE);
      break;

    // buffer integer constant
    case TYPE_ICONST:
      pcb->AddInteger(data.n, FALSE);
      break;

    // buffer internal function
    case TYPE_FUNC:
      pcb->AddFunction(data.func.iFunc, data.func.iPD, FALSE);
      pcb->AddChar('(');
      apmn[0]->WriteCode();
      for(i = 1; i < c; ++i)
      { pcb->AddChar(',');
	apmn[i]->WriteCode();
      }
      pcb->AddChar(')');
      break;

    // buffer argument math node
    case TYPE_ARG:
      pcb->AddArgument(data.arg.i, FALSE);
      break;

    // buffer local variable math node
    case TYPE_LVAR:
      if (uExpr & FLAG_DIFF) pcb->AddLocalDerivative(data.lvar.i, FALSE);
      else		     pcb->AddLocalVariable(data.lvar.i, FALSE);
      break;

    // precomputed user-defined function
    case TYPE_PRECOMP:
      iPD = apmn[0]->data.udfunc.iPD;
      i1  = (int) data.precomp.au1[iPD-1];
      i2  = (int) data.precomp.au2[iPD-1];
      if (!data.precomp.fNode && iPD)
      { if (i1 == -i2)
	  pcb->AddText((data.precomp.fOdd) ? "_pcfeven(" : "_pcfodd(");
	else
          pcb->AddText((data.precomp.fOdd) ? "_pcfodd(" : "_pcfeven(");
	for(i = 1; i < 4; ++i)
	{ apmn[i]->WriteCode();
	  pcb->AddChar(',');
	}
	pcb->AddPrecUDFunc(apmn[0]->data.udfunc.iFunc, FALSE);
	pcb->AddChar(',');
	pcb->AddInteger(i1, FALSE);
	pcb->AddChar(',');
	pcb->AddInteger((i2 < 0) ? -i2 : i2, FALSE);
	pcb->AddChar(')');
      }
      else
      { if (data.precomp.fOdd)
	{ pcb->AddText("_scheck(");
	  apmn[2]->WriteCode();
	  pcb->AddChar(',');
	  apmn[3]->WriteCode();
	  pcb->AddText(")*");
	}
	pcb->AddPrecUDFunc(apmn[0]->data.udfunc.iFunc, FALSE);
	pcb->AddChar('[');
	apmn[1]->WriteCode();
	pcb->AddChar(']');

	if (apmn[0]->data.udfunc.iPD)
	{ pcb->AddChar('[');
	  pcb->AddUInteger(i1, FALSE);
	  pcb->AddChar(']');
	}
	else
	{ for(i = 0; i < apmn[0]->c; ++i)
	    if (data.precomp.au1[i] != -1u) break;
	  if (i < apmn[0]->c) pcb->AddText("[0]");
	}
      }
      break;

    // buffer user-defined function
    case TYPE_UDFUNC:
      pcb->AddUDFunction(data.udfunc.iFunc, data.udfunc.iPD, FALSE);
      pcb->AddChar('(');
      apmn[0]->WriteCode();
      for(i = 1; i < c; ++i)
      { pcb->AddChar(',');
	apmn[i]->WriteCode();
      }
      pcb->AddChar(')');
      break;

    // buffer addition operator
    case TYPE_ADD:
      pcb->AddChar('(');
      apmn[0]->WriteCode();
      pcb->AddChar('+');
      apmn[1]->WriteCode();
      pcb->AddChar(')');
      break;

    // buffer subtraction operator
    case TYPE_SUB:
      pcb->AddChar('(');
      apmn[0]->WriteCode();
      pcb->AddChar('-');
      apmn[1]->WriteCode();
      pcb->AddChar(')');
      break;

    // buffer multiplication operator
    case TYPE_MUL:
      pcb->AddChar('(');
      apmn[0]->WriteCode();
      pcb->AddChar('*');
      apmn[1]->WriteCode();
      pcb->AddChar(')');
      break;

    // buffer division operator
    case TYPE_DIV:
      pcb->AddChar('(');
      apmn[0]->WriteCode();
      pcb->AddChar('/');
      apmn[1]->WriteCode();
      pcb->AddChar(')');
      break;

    // buffer and operator
    case TYPE_AND:
      pcb->AddText("(((");
      apmn[0]->WriteCode();
      pcb->AddText(")&&(");
      apmn[1]->WriteCode();
      if (uExpr & FLAG_REAL) pcb->AddText("))?1.0:0.0)");
      else		     pcb->AddText("))?1:0)");
      break;

    // buffer or operator
    case TYPE_OR:
      pcb->AddText("(((");
      apmn[0]->WriteCode();
      pcb->AddText(")||(");
      apmn[1]->WriteCode();
      if (uExpr & FLAG_REAL) pcb->AddText("))?1.0:0.0)");
      else		     pcb->AddText("))?1:0)");
      break;

    // buffer xor operator
    case TYPE_XOR:
      pcb->AddText("((((");
      apmn[0]->WriteCode();
      pcb->AddText(")&&!(");
      apmn[1]->WriteCode();
      pcb->AddText("))||(!(");
      apmn[0]->WriteCode();
      pcb->AddText(")&&(");
      apmn[1]->WriteCode();
      pcb->AddText(")))||(!(");
      if (uExpr & FLAG_REAL) pcb->AddText(")))?1.0:0.0)");
      else		     pcb->AddText(")))?1:0)");
      break;

    // buffer greater than operator
    case TYPE_GT:
      pcb->AddText("(((");
      apmn[0]->WriteCode();
      pcb->AddText(")>(");
      apmn[1]->WriteCode();
      if (uExpr & FLAG_REAL) pcb->AddText("))?1.0:0.0)");
      else		     pcb->AddText("))?1:0)");
      break;

    // buffer greater than or equal to operator
    case TYPE_GE:
      pcb->AddText("(((");
      apmn[0]->WriteCode();
      pcb->AddText(")>=(");
      apmn[1]->WriteCode();
      if (uExpr & FLAG_REAL) pcb->AddText("))?1.0:0.0)");
      else		     pcb->AddText("))?1:0)");
      break;

    // buffer less than operator
    case TYPE_LT:
      pcb->AddText("(((");
      apmn[0]->WriteCode();
      pcb->AddText(")<(");
      apmn[1]->WriteCode();
      if (uExpr & FLAG_REAL) pcb->AddText("))?1.0:0.0)");
      else		     pcb->AddText("))?1:0)");
      break;

    // buffer less than or equal to operator
    case TYPE_LE:
      pcb->AddText("(((");
      apmn[0]->WriteCode();
      pcb->AddText(")<=(");
      apmn[1]->WriteCode();
      if (uExpr & FLAG_REAL) pcb->AddText("))?1.0:0.0)");
      else		     pcb->AddText("))?1:0)");
      break;

    // buffer equal to operator
    case TYPE_EQ:
      pcb->AddText("(((");
      apmn[0]->WriteCode();
      pcb->AddText(")==(");
      apmn[1]->WriteCode();
      if (uExpr & FLAG_REAL) pcb->AddText("))?1.0:0.0)");
      else		     pcb->AddText("))?1:0)");
      break;

    // buffer not equal to operator
    case TYPE_NE:
      pcb->AddText("(((");
      apmn[0]->WriteCode();
      pcb->AddText(")!=(");
      apmn[1]->WriteCode();
      if (uExpr & FLAG_REAL) pcb->AddText("))?1.0:0.0)");
      else		     pcb->AddText("))?1:0)");
      break;

    // buffer negation operator
    case TYPE_NEG:
      pcb->AddText("-(");
      apmn[0]->WriteCode();
      pcb->AddChar(')');
      break;

    // buffer not operator
    case TYPE_NOT:
      pcb->AddText("((");
      apmn[0]->WriteCode();
      if (uExpr & FLAG_REAL) pcb->AddText(")?1.0:0.0)");
      else		     pcb->AddText(")?1:0)");
      break;

    // buffer real operator
    case TYPE_REAL:
      pcb->AddText("((REAL)(");
      apmn[0]->WriteCode();
      pcb->AddText("))");
      break;

    // buffer area operator
    case TYPE_AREA:
      pcb->AddText("_area(");
      if (data.n >= 0)
      { pcb->AddText("_bnode_map(");
	apmn[0]->WriteCode();
	pcb->AddChar(',');
	pcb->AddInteger(data.n, FALSE);
	pcb->AddChar(')');
      }
      else apmn[0]->WriteCode();
      pcb->AddChar(')');
      break;

    // buffer area 1 operator
    case TYPE_AREA1:
      pcb->AddText("_area1(");
      if (data.n >= 0)
      { pcb->AddText("_bnode_map(");
	apmn[0]->WriteCode();
	pcb->AddChar(',');
	pcb->AddInteger(data.n, FALSE);
	pcb->AddChar(')');
      }
      else apmn[0]->WriteCode();
      pcb->AddChar(',');
      apmn[1]->WriteCode();
      pcb->AddChar(')');
      break;

    // buffer area 2 operator
    case TYPE_AREA2:
      pcb->AddText("_area2(");
      if (data.n >= 0)
      { pcb->AddText("_bnode_map(");
	apmn[0]->WriteCode();
	pcb->AddChar(',');
	pcb->AddInteger(data.n, FALSE);
	pcb->AddChar(')');
      }
      else apmn[0]->WriteCode();
      pcb->AddChar(',');
      apmn[1]->WriteCode();
      pcb->AddChar(')');
      break;

    // buffer node operator
    case TYPE_NODE:
      pcb->AddText("_node(");
      if (data.n >= 0)
      { pcb->AddText("_bnode_map(");
	apmn[0]->WriteCode();
	pcb->AddChar(',');
	pcb->AddInteger(data.n, FALSE);
	pcb->AddChar(')');
      }
      else apmn[0]->WriteCode();
      pcb->AddChar(',');
      apmn[1]->WriteCode();
      pcb->AddChar(')');
      break;

    // buffer edge operator
    case TYPE_EDGE:
      pcb->AddText("_edge(");
      if (data.n >= 0)
      { pcb->AddText("_bnode_map(");
	apmn[0]->WriteCode();
	pcb->AddChar(',');
	pcb->AddInteger(data.n, FALSE);
	pcb->AddChar(')');
      }
      else apmn[0]->WriteCode();
      pcb->AddChar(',');
      apmn[1]->WriteCode();
      pcb->AddChar(')');
      break;

    // buffer edge length operator
    case TYPE_ELEN:
      pcb->AddText("_elen(");
      if (data.n >= 0)
      { pcb->AddText("_bnode_map(");
	apmn[0]->WriteCode();
	pcb->AddChar(',');
	pcb->AddInteger(data.n, FALSE);
	pcb->AddChar(')');
      }
      else apmn[0]->WriteCode();
      pcb->AddChar(',');
      apmn[1]->WriteCode();
      pcb->AddChar(')');
      break;

    // buffer edge length operator
    case TYPE_ELENABS:
      if (fPrecompute)
//    { pcb->AddText("_elen_abs(");
//	pcb->AddIndex(1, FALSE);
//	pcb->AddChar(',');
//	pcb->AddIndex(2, FALSE);
//	pcb->AddChar(')');
//    }
      { pcb->AddText("_elen(");
	pcb->AddIndex(0, FALSE);
	pcb->AddChar(')');
      }
      else
//    { pcb->AddText("_elen_abs(");
//	if (data.n >= 0)
//	{ pcb->AddText("_bnode_map(");
//	  apmn[0]->WriteCode();
//	  pcb->AddChar(',');
//	  pcb->AddInteger(data.n, FALSE);
//	  pcb->AddChar(')');
//	}
//	else apmn[0]->WriteCode();
//	pcb->AddChar(',');
//	apmn[1]->WriteCode();
//	pcb->AddChar(')');
//    }
      { pcb->AddText("_elen(");
	apmn[0]->WriteCode();
	pcb->AddChar(')');
      }
      break;

    // buffer integration length operator
    case TYPE_ILEN:
      pcb->AddText("_ilen(");
      if (data.n >= 0)
      { pcb->AddText("_bnode_map(");
	apmn[0]->WriteCode();
	pcb->AddChar(',');
	pcb->AddInteger(data.n, FALSE);
	pcb->AddChar(')');
      }
      else apmn[0]->WriteCode();
      pcb->AddChar(',');
      apmn[1]->WriteCode();
      pcb->AddChar(')');
      break;

    // buffer node summation
    case TYPE_NSUM:
      pcb->AddSum(data.nsum.iExpr);
      break;

    // buffer label summation
    case TYPE_LSUM:
      pcb->AddSum(data.lsum.iExpr);
      break;
  }
}



///////////////////////////////////////////////////////////////////////////////
// Function:	 Statements::CodeProcedure
// Scope:	 Public
// Description:	 This function codes the procedure starting at the specified
//		 statement.
void Statements::CodeProcedure(char *pszResFile, void *pv, UINT i)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("gencode.cpp Statements::CodeProcedure\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  // add the end token to the main procedure
  static BOOL fAddToken = TRUE;
  if (fAddToken)
  { AddMain(new Statement(TYPE_RETURN, -1u));
    fAddToken = FALSE;
  }

  // declare the procedure summations
  DeclareProcedureSums(pv, i);

  // initialize local variables
  UINT cStatement = DAPMainProc.GetElements();
  CodeBuffer *pcb = (CodeBuffer *) pv;
  BOOL fReturn = FALSE;

  do
  { pcb->SetLeftMargin(2);
    pcb->AddLabel(i, TRUE, FALSE);
    pcb->AddChar('\n');
    Statement *ps = DAPMainProc[i++];


    VarAssign	*pva;
    ArrayAssign *paa;
    Goto	*pgoto;
    Call	*pcall;
    FileOpen	*pfopen;
    FileGoto	*pfgoto;
    FileLabel	*pflabel;
    FileComment *pfcomment;
    FileRead	*pfread;
    FileWrite	*pfwrite;
    Indices	*pindices;
    UINT	i, c;

    switch (ps->uType)
    { // variable assignment statement
      case TYPE_VASSIGN:
	pva = TYPESAFE_DOWNCAST(ps, VarAssign);
	pva->pmn->WriteSums();
	pcb->AddVariable(pva->i, FALSE, TRUE);
	pcb->AddText("= ");
	pcb->SetLeftMargin(4);
	pcb->AddMathNode(pva->pmn);
	pcb->AddText(";\n\n");
	break;

      // array assignment statement
      case TYPE_AASSIGN:
	paa = TYPESAFE_DOWNCAST(ps, ArrayAssign);
	pindices = paa->pindices;
	c = pindices->GetDim();
	for(i = 0; i < c; ++i)
	{ BOOL fLabelSum = pindices->GetType(i) == INDEX_LABEL;
	  pcb->SetLeftMargin((i + 1) * 2);
	  pcb->AddText("for(");
	  pcb->AddIndex(i,TRUE);
	  pcb->AddText("= ");
	  if (fLabelSum)
	  { pcb->AddText("_label_init(");
	    pcb->AddUInteger(i, FALSE);
	    pcb->AddChar(',');
	    pcb->AddUInteger(pindices->GetLabelIndex(i), FALSE);
	    pcb->AddChar(')');
	  }
	  else pcb->AddInteger(pindices->GetBegin(i), FALSE);
	  pcb->AddText("; ");
	  pcb->AddIndex(i,TRUE);
	  if (fLabelSum) pcb->AddText("!= -1u");
	  else
	  { pcb->AddText("<= ");
	    pcb->AddInteger(pindices->GetEnd(i), FALSE);
	  }
	  pcb->AddText("; ");
	  pcb->AddIndex(i,TRUE);
	  if (fLabelSum)
	  { pcb->AddText("= _label_next(");
	    pcb->AddUInteger(i, FALSE);
	    pcb->AddChar(')');
	  }
	  else
	  { pcb->AddText("+= ");
	    pcb->AddInteger(pindices->GetStep(i), FALSE);
	  }
	  pcb->AddText(")\n");
	}
	pcb->AddText("{ ");
	pcb->SetLeftMargin((c + 1) * 2);
	paa->pmn->WriteSums();
	pcb->AddArray(paa->i, FALSE, FALSE);
	for(i = 0; i < c; ++i)
	{ pcb->AddChar('[');
	  pcb->AddIndex(i, FALSE);
	  pcb->AddChar(']');
	}
	pcb->AddText(" = ");
	pcb->SetLeftMargin((c + 2) * 2);
	pcb->AddMathNode(paa->pmn);
	pcb->AddText(";\n");
	pcb->SetLeftMargin(c*2);
	pcb->AddText("}\n\n");
	break;

      // conditional goto statement
      case TYPE_GOTO:
	pgoto = TYPESAFE_DOWNCAST(ps, Goto);
	if (pgoto->pmn)
	{ pgoto->pmn->WriteSums();
	  pcb->AddText("if (");
	  pcb->AddMathNode(pgoto->pmn);
	  pcb->AddText(") ");
	}
	pcb->AddText("goto ");
	pcb->AddLabel(pgoto->i, FALSE, FALSE);
	pcb->AddText(";\n\n");
	break;

      // file open statement
      case TYPE_FILEOPEN:
	pfopen = TYPESAFE_DOWNCAST(ps, FileOpen);
	pcb->AddText("FileIO_Open(");
	pcb->AddString(pfopen->psz);
	pcb->AddText((pfopen->u == IOMODE_READ)   ? ", IOMODE_READ"  :
		     ((pfopen->u == IOMODE_WRITE) ? ", IOMODE_WRITE"
		 			          : ", IOMODE_APPEND"));
	pcb->AddText(");\n\n");
	break;

      // file close statement
      case TYPE_FILECLOSE:
	pcb->AddText("FileIO_Close();\n\n");
	break;

      // file goto statement
      case TYPE_FILEGOTO:
	pfgoto = TYPESAFE_DOWNCAST(ps, FileGoto);
	pcb->AddText("FileIO_Goto(");
	pcb->AddString(pfgoto->psz);
	pcb->AddText(");\n\n");
	break;

      // file label statement
      case TYPE_FILELABEL:
	pflabel = TYPESAFE_DOWNCAST(ps, FileLabel);
	pcb->AddText("FileIO_Label(");
	pcb->AddString(pflabel->psz);
	pcb->AddText(");\n\n");
	break;

      // file comment statement
      case TYPE_FILECOMM:
	pfcomment = TYPESAFE_DOWNCAST(ps, FileComment);
	pcb->AddText("FileIO_Comment(");
	pcb->AddString(pfcomment->psz);
	pcb->AddText(");\n\n");
	break;

      // file read statement
      case TYPE_FILEREAD:
	pfread = TYPESAFE_DOWNCAST(ps, FileRead);
	pcb->AddText("FileIO_Read(");
	pcb->AddUInteger(pfread->i);
	pcb->AddText((pfread->c) ? ", TRUE" : ", FALSE");
	pcb->AddText(", axlatvar@L@, axlatarray@L@, (REAL *) &var@L@);\n\n");
	break;

      // file write statement
      case TYPE_FILEWRITE:
	pfwrite = TYPESAFE_DOWNCAST(ps, FileWrite);
	pcb->AddText("FileIO_Write(");
	pcb->AddUInteger(pfwrite->i);
	pcb->AddText((pfwrite->c) ? ", TRUE" : ", FALSE");
	pcb->AddText(", axlatvar@L@, axlatarray@L@, (REAL *) &var@L@);\n\n");
	break;

      // call statement
      case TYPE_CALL:
	pcall = TYPESAFE_DOWNCAST(ps, Call);
	pcb->AddProcedure(pcall->iProc, FALSE);
	pcb->AddText(";\n\n");
	break;

      // solve, write, end, fileclose and return statements
      case TYPE_SOLVE:
	pcb->SetRightMargin(250);
	pcb->AddText("SolveEquations(plinsolve_@L@, pmtrx_@L@, pvectDX_@L@, "
		     "pvectRHS_@L@, aiPermRow_@L@, aiPermCol_@L@, atag_@L@, "
		     "aanEI_@L@, (REAL *) &var@L@, apfequ_@L@, apfpd_@L@, "
		     "PCF_@U@);\n\n");
	break;

      case TYPE_WRITE:
	pcb->AddText("WriteElements(");
	pcb->AddString(pszResFile);
	pcb->AddText(", &reshead@L@, (REAL *) &var@L@);\n\n");
	break;

      case TYPE_END:
	pcb->AddText("exit(0);\n\n");
	break;

      case TYPE_RETURN:
	pcb->AddText("return;\n");
	fReturn = TRUE;
	break;
    }
  } while ((i < cStatement) && !fReturn);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Statements::CodeEquations
// Scope:	 Public
// Description:	 This function codes the equations and their derivatives.
void Statements::CodeEquations (void *pv)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("gencode.cpp Statements::CodeEquations\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i, j, k;

  CodeBuffer *pcb = (CodeBuffer *) pv;
  UINT c = DAPEquation.GetElements();

  // code the precompute macros
  pcb->SetLeftMargin(0);
  pcb->AddText("#define _pcfodd(edge, nodeA, nodeB, func, iPD1, iPD2) \\\n"
	       "  (((nodeA)<(nodeB))?(func[(edge)][iPD1]):(-func[(edge)][iPD2]))\n\n"
	       "#define _pcfeven(edge, nodeA, nodeB, func, iPD1, iPD2) \\\n"
	       "  (((nodeA)<(nodeB))?(func[(edge)][iPD1]):(func[(edge)][iPD2]))\n\n\n");

  // for each equation
  cSum = 0;
  for(i = 0; i < c; ++i)
  { Statement *ps = DAPEquation[i];
    MathNode *pmnEqu;
    MathNode **apmnElem;
    MathNode **apmnDiff;
    UINT cContrib;
    UINT cDim;
    if (ps->uType == TYPE_VEQU)
    { VarEquation *pve = TYPESAFE_DOWNCAST(ps, VarEquation);
      pmnEqu   = pve->pmn;
      apmnElem = pve->apmnElem;
      apmnDiff = pve->apmnDiff;
      cContrib = pve->cContrib;
      cDim     = 0;
    }
    else
    { ArrayEquation *pae = TYPESAFE_DOWNCAST(ps, ArrayEquation);
      pmnEqu   = pae->pmn;
      apmnElem = pae->apmnElem;
      apmnDiff = pae->apmnDiff;
      cContrib = pae->cContrib;
      cDim     = pae->pindices->GetDim();
    }

    // write equation function
    cSum = 0;
    pcb->SetLeftMargin(0);
    pcb->AddText("REAL ");
    pcb->AddEquation(i, TRUE);
    pcb->AddText("(I3ARRAY an)\n{\n");
    pcb->SetLeftMargin(2);
    for(k = 0; k < cDim; ++k)
    { pcb->AddText("UINT ");
      pcb->AddIndex(k, TRUE);
      pcb->AddText("= an[");
      pcb->AddUInteger(k, FALSE);
      pcb->AddText("];\n");
    }
    if (pszMeshFileName)
    { pcb->AddText("UINT ");
      pcb->AddIndex(k, FALSE);
      pcb->AddText(";\n");
      pcb->AddText("UINT ");
      pcb->AddIndex(k+1, FALSE);
      pcb->AddText(";\n");
    }
    pcb->AddChar('\n');
    pmnEqu->WriteSums((void *) pcb, FALSE);
    pcb->AddText("return ");
    pcb->SetLeftMargin(4);
    pcb->AddMathNode(pmnEqu);
    pcb->AddText(";\n");
    pcb->SetLeftMargin(0);
    pcb->AddText("}\n\n");

    // code partial derivative function
    pcb->SetLeftMargin(0);
    pcb->AddText("void ");
    pcb->AddPartial(i, TRUE);
    pcb->AddText("(REAL *arVWR, UINT *aiPerm, I3ARRAY an)\n{\n");
    pcb->SetLeftMargin(2);
    for(k = 0; k < cDim; ++k)
    { pcb->AddText("UINT ");
      pcb->AddIndex(k, TRUE);
      pcb->AddText("= an[");
      pcb->AddUInteger(k, FALSE);
      pcb->AddText("];\n");
    }
    if (pszMeshFileName)
    { pcb->AddText("UINT ");
      pcb->AddIndex(k, FALSE);
      pcb->AddText(";\n");
      pcb->AddText("UINT ");
      pcb->AddIndex(k+1, FALSE);
      pcb->AddText(";\n");
    }
    pcb->AddText("UINT iUnknown;\n\n");
    WriteContribSums((void *) pcb, cContrib, apmnDiff);
    for(j = 0; j < cContrib; ++j)
    { pcb->SetLeftMargin(2);
      MathNode *pmn = apmnElem[j];
      MathNode *pmnLSUM = (MathNode *) ((pmn->uNode == TYPE_VAR) ? pmn->data.var.pmnLSUM : pmn->data.array.pmnLSUM);
      MathNode *pmnNSUM = (MathNode *) ((pmn->uNode == TYPE_VAR) ? pmn->data.var.pmnNSUM : pmn->data.array.pmnNSUM);
      UINT iLeftMargin = 4;
      if (pmnLSUM)
      {	 pcb->AddText("for(");
	 pcb->AddIndex(pmnLSUM->data.lsum.iIndex, TRUE);
	 pcb->AddText("= _label_init(0,");
	 pcb->AddInteger(pmnLSUM->data.lsum.iLabel);
	 pcb->AddText("); ");
	 pcb->AddIndex(pmnLSUM->data.lsum.iIndex, TRUE);
	 pcb->AddText("!= -1u; ");
	 pcb->AddIndex(pmnLSUM->data.lsum.iIndex, TRUE);
	 pcb->AddText("= _label_next(0))\n{ ");
	 pcb->SetLeftMargin(iLeftMargin);
	 iLeftMargin += 2;
      }
      if (pmnNSUM)
      { UINT iLabel = pmnNSUM->data.nsum.iLabel;
	pcb->AddText("for(_nsum_init(");
	if (iLabel < cRegLabel)
	{ pcb->AddText("_bnode_map(");
	  pcb->AddMathNode(pmnNSUM->apmn[0]);
	  pcb->AddChar(',');
	  pcb->AddUInteger(iLabel, FALSE);
	  pcb->AddChar(')');
	  iLabel = -1u;
	}
	else pcb->AddMathNode(pmnNSUM->apmn[0]);
	pcb->AddChar(',');
	if (iLabel == -1u) pcb->AddText("-1u");
	else pcb->AddUInteger(iLabel, FALSE);
	pcb->AddChar(',');
	pcb->AddText((pmnNSUM->data.nsum.fTriSum) ? "TRUE" : "FALSE");
	pcb->AddText("); !_nsum_loopover(); _nsum_next())\n{ ");
	pcb->AddIndex(pmnNSUM->data.nsum.iIndex, TRUE);
	pcb->AddText("= _nsum_index();\n");
	pcb->SetLeftMargin(iLeftMargin);
	iLeftMargin += 2;
      }

      pcb->AddText("iUnknown = ");
      pcb->AddMathNode(pmn, TRUE);
      pcb->AddText(";\nif (iUnknown != -1u) arVWR[aiPerm[iUnknown]] += ");
      pcb->SetLeftMargin(iLeftMargin);
      pcb->AddMathNode(apmnDiff[j]);
      pcb->AddText(";\n");

      if (pmnNSUM)
      { pcb->SetLeftMargin(iLeftMargin-4);
	pcb->AddText("}\n");
      }

      if (pmnLSUM)
      { pcb->SetLeftMargin(2);
	pcb->AddText("}\n");
      }

      pcb->AddChar('\n');
    }
    pcb->SetLeftMargin(0);
    pcb->AddText("}\n\n");
  }
}

///////////////////////////////////////////////////////////////////////////////
// Function:	 FileInputOutput::nWriteCode
// Scope:	 Public
// Description:	 This function writes C++ source code which when linked to
//		 the numerical analysis library creates an executable
//		 simulation.
int FileInputOutput::nWriteCode (char *pszSource, UINT iMainProc, void *pv)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("gencode.cpp FileInputOutput::nWriteCode\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  char *psz;
  UINT i, j, k, l, c;
  SymIConst *psymIC;
  SymRConst *psymRC;
  SymVar *psymV;
  SymArray *psymA;
  UINT cEquations = pstatements->GetNumOfEqu();
  XFERBUF04 *pxb = (XFERBUF04 *) pv;

  // form the name prefix
  if ((psz = strrchr(pszSource, '\\')) == NULL) psz = pszSource;
  else ++psz;
  strcpy(pszGlobalBuf, psz);
  if ((psz = strrchr(pszGlobalBuf, '.')) != NULL) *psz = '\0';
  for(i = strlen(pszGlobalBuf); i < LEN_PREFIX; ++i) pszGlobalBuf[i] = '_';
  pszGlobalBuf[LEN_PREFIX] = '\0';
  char szPrefix[LEN_PREFIX + 1];
  strcpy(szPrefix, pszGlobalBuf);

  // allocate the header and source instances of the code buffer class
  FormFileName(pszSource, FILE_HEADER);
  CodeBuffer *pcbHeader = new CodeBuffer(pszGlobalBuf, szPrefix, cmdargs.cPrecision);
  FormFileName(pszSource, FILE_CODE);
  CodeBuffer *pcbSource = new CodeBuffer(pszGlobalBuf, szPrefix, cmdargs.cPrecision);

  char szHeaderFile[LEN_FILENAME];
  FormFileName(pszSource, FILE_HEADER, TRUE);
  strcpy(szHeaderFile, pszGlobalBuf);
  /* strlwr(szHeaderFile); */ /* Switch to case-sensitive by Josh */

  char szResultFile[LEN_FILENAME];
  FormFileName(pszSource, FILE_RESULT, TRUE);
  strcpy(szResultFile, pszGlobalBuf);
  /* strlwr(szResultFile); */ /* Switch to case-sensitive by Josh */

  char szTopFile[LEN_FILENAME];
  FormFileName(pszSource, FILE_TOP, TRUE);
  strcpy(szTopFile, pszGlobalBuf);
  /* strlwr(szTopFile); */ /* Switch to case-sensitive by Josh */

  char szPermFile[LEN_FILENAME];
  FormFileName(pszSource, FILE_PERM, TRUE);
  strcpy(szPermFile, pszGlobalBuf);
  /* strlwr(szPermFile); */ /* Switch to case-sensitive by Josh */

  pcbHeader->SetLeftMargin(0);
  pcbHeader->AddText("/* This simulation was created by SGFramework version 1.0 */\n");
  pcbHeader->AddText("/* Copyright (c) 1997 K. Kramer */\n\n");
  pcbSource->SetLeftMargin(0);
  pcbSource->AddText("/* This simulation was created by SGFramework version 1.0 */\n");
  pcbSource->AddText("/* Copyright (c) 1997 K. Kramer */\n\n");
  pcbSource->AddText("#include <math.h>\n"
		     "#include <stdio.h>\n"
		     "#include <stdlib.h>\n"
		     "#include \"sgnam.h\"\n"
		     "#include \"");
  pcbSource->AddText(szHeaderFile);
  pcbSource->AddText("\"\n\n");

  // write variable and tag structures in the header file and declare it in
  // the source file
  for(l = (cEquations) ? 0 : 1; l < 2; ++l)
  { pcbHeader->AddText((l) ? "typedef struct tagVAR@U@\n{ " : "typedef struct tagTAG@U@\n{ ");
    pcbHeader->SetLeftMargin(2);
    UINT cVars = symTable->cSymbols(TYPE_VAR);
    UINT cArrays = symTable->cSymbols(TYPE_ARRAY);
    i = j = 0;
    psymV = (cVars)   ? TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_VAR, i),   SymVar)   : NULL;
    psymA = (cArrays) ? TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_ARRAY, j), SymArray) : NULL;
    while ((i < cVars) || (j < cArrays))
    { if (!psymA || (psymV && (psymV->u < psymA->u)))
      { pcbHeader->AddText((l) ? "REAL _" : "UINT _");
	pcbHeader->AddText(psymV->szName);
	pcbHeader->AddText(";\n");
	psymV = (++i < cVars) ?
	  TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_VAR, i), SymVar) :
	  NULL;
      }
      else
      { pcbHeader->AddText((l) ? "REAL _" : "UINT _");
	pcbHeader->AddText(psymA->szName);
	for(k = 0; k < psymA->c; ++k)
	{ pcbHeader->AddChar('[');
	  pcbHeader->AddUInteger(psymA->ac[k], FALSE);
	  pcbHeader->AddChar(']');
	}
	pcbHeader->AddText(";\n");
	psymA = (++j < cArrays) ?
	  TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_ARRAY, j), SymArray) :
	  NULL;
      }
    }
    pcbHeader->SetLeftMargin(0);
    pcbHeader->AddText((l) ? "} VAR@U@;\nextern VAR@U@ var@L@;\n\n" : "} TAG@U@;\nextern TAG@U@ tag@L@;\n\n");
    pcbSource->AddText((l) ? "VAR@U@ var@L@;\n\n" : "TAG@U@ tag@L@;\n");
  }

  // write the structure which stores the precomputed functions and derivatives
  UINT cUDFuncs = symTable->cSymbols(TYPE_UDFUNC);
  UINT cPrecomputedFunction = 0;
  UINT cMaxNumOfPrecFuncArg = 0;
  UINT cNodePrecomputeFunc  = 0;
  UINT cEdgePrecomputeFunc  = 0;
  for(i = 0; i < cUDFuncs; ++i)
  { SymUDFunc *psymUDFunc = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_UDFUNC, i), SymUDFunc);
    if (psymUDFunc->pmnTemplate)
    {
      // write structure intro if this is the first precomputed function
      if (!cPrecomputedFunction)
      { pcbHeader->SetLeftMargin(0);
	pcbHeader->AddText("typedef struct tagPRE@U@\n{ ");
	pcbHeader->SetLeftMargin(2);
      }

      // are there any derivatives?
      UINT cArg = psymUDFunc->c;
      UINT cDerivative = 0;
      for(j = 0; j < cArg; ++j)
	if (psymUDFunc->af[j]) ++cDerivative;

      // write the storage array
      pcbHeader->AddText("REAL _");
      pcbHeader->AddText(psymUDFunc->szName);
      pcbHeader->AddChar('[');
      pcbHeader->AddUInteger((psymUDFunc->fNode) ? cNode : cEdge, FALSE);
      pcbHeader->AddChar(']');
      if (cDerivative)
      { pcbHeader->AddChar('[');
	pcbHeader->AddUInteger(cDerivative + 1,FALSE);
	pcbHeader->AddChar(']');
      }
      pcbHeader->AddText(";\n");

      if (cArg > cMaxNumOfPrecFuncArg) cMaxNumOfPrecFuncArg = cArg;
      if (psymUDFunc->fNode) ++cNodePrecomputeFunc;
      else		     ++cEdgePrecomputeFunc;
      ++cPrecomputedFunction;
    }
  }
  pcbHeader->SetLeftMargin(0);
  if (cPrecomputedFunction)
  { pcbHeader->AddText("} PRE@U@;\nextern PRE@U@ pre@L@;\n\n");
    pcbSource->AddText("PRE@U@ pre@L@;\n\n");
  }

  if (cEquations)
  { pcbHeader->AddText("extern SparseMatrix *pmtrx_@L@;\n"
		       "extern Vector *pvectRHS_@L@;\n"
		       "extern Vector *pvectDX_@L@;\n"
		       "extern LinearSolve *plinsolve_@L@;\n"
		       "extern UINT *aiPermRow_@L@;\n"
		       "extern UINT *aiPermCol_@L@;\n"
		       "extern UINT *aiTop_@L@;\n"
		       "extern TAG *atag_@L@;\n"
		       "extern TOP *atop_@L@;\n"
		       "extern I3ARRAY *aanEI_@L@;\n"
		       "extern UINT cElem_@L@;\n"
		       "extern UINT cTop_@L@;\n"
		       "extern UINT cTag_@L@;\n"
		       "extern UINT cEqu_@L@;\n\n");
    pcbSource->AddText("SparseMatrix *pmtrx_@L@;\n"
		       "Vector *pvectRHS_@L@;\n"
		       "Vector *pvectDX_@L@;\n"
		       "LinearSolve *plinsolve_@L@;\n"
		       "UINT *aiPermRow_@L@;\n"
		       "UINT *aiPermCol_@L@;\n"
		       "UINT *aiTop_@L@;\n"
		       "TAG *atag_@L@;\n"
		       "TOP *atop_@L@;\n"
		       "I3ARRAY *aanEI_@L@;\n"
		       "UINT cElem_@L@;\n"
		       "UINT cTop_@L@;\n"
		       "UINT cTag_@L@;\n"
		       "UINT cEqu_@L@;\n\n");
  }

  // write the pointer to mesh variables array
  pcbSource->SetLeftMargin(0);
  pcbSource->AddText("REAL *apr@U@_MeshVar[] =\n{ ");
  pcbSource->SetLeftMargin(2);
  UINT cArrays = symTable->cSymbols(TYPE_ARRAY);
  char *pszX;
  char *pszY;
  for(i = 0, c = 0; i < cArrays; ++i)
  { psymA = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_ARRAY, i), SymArray);
    if (psymA->fMeshVar)
    { pcbSource->AddArray(psymA->szName, FALSE, FALSE);
      pcbSource->AddText(",\n");
      if      (c == 0) { pszX = psymA->szName; ++c; }
      else if (c == 1) { pszY = psymA->szName; ++c; }
    }
  }
  pcbSource->AddText("NULL\n");
  pcbSource->SetLeftMargin(0);
  pcbSource->AddText("};\n\n");

  // write the constant structure
  UINT cRConstant = symTable->cSymbols(TYPE_RCONST);
  UINT cIConstant = symTable->cSymbols(TYPE_ICONST);
  UINT cConstant = cRConstant + cIConstant;
  pcbHeader->SetLeftMargin(0);
  pcbSource->SetLeftMargin(0);
  if (cConstant)
  { pcbHeader->AddText("extern XLATCONST axlatconst@L@[];\n");
    pcbSource->AddText("XLATCONST axlatconst@L@[] =\n{ ");
    pcbSource->SetLeftMargin(2);
    for(i = 0; i < cIConstant; ++i)
    { psymIC = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_ICONST, i), SymIConst);
      if (i) pcbSource->AddText(",\n");
      pcbSource->AddText("{ ");
      pcbSource->AddString(psymIC->szName);
      pcbSource->AddText(", TYPE_ICONST, { ");
      pcbSource->AddReal((REAL) psymIC->n, TRUE);
      pcbSource->AddText("} }");
      pcbHeader->AddText("#define CONST_@U@_");
      pcbHeader->AddText(psymIC->szName);
      pcbHeader->AddChar(' ');
      pcbHeader->AddInteger(psymIC->n, FALSE);
      pcbHeader->AddChar('\n');
    }
    for(i = 0; i < cRConstant; ++i)
    { psymRC = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_RCONST, i), SymRConst);
      if (i || cIConstant) pcbSource->AddText(",\n");
      pcbSource->AddText("{ ");
      pcbSource->AddString(psymRC->szName);
      pcbSource->AddText(", TYPE_RCONST, { ");
      pcbSource->AddReal(psymRC->r, TRUE);
      pcbSource->AddText("} }");
      pcbHeader->AddText("#define CONST_@U@_");
      pcbHeader->AddText(psymRC->szName);
      pcbHeader->AddChar(' ');
      pcbHeader->AddReal(psymRC->r, FALSE);
      pcbHeader->AddChar('\n');
    }
    pcbSource->SetLeftMargin(0);
    pcbSource->AddText("\n};\n\n");
    pcbHeader->AddChar('\n');
  }
  else
  { pcbHeader->AddText("extern XLATCONST *axlatconst@L@;\n");
    pcbSource->AddText("XLATCONST *axlatconst@L@ = NULL;\n\n");
  }

  // write the variable structure
  UINT cVariable = symTable->cSymbols(TYPE_VAR);
  pcbHeader->SetLeftMargin(0);
  pcbSource->SetLeftMargin(0);
  if (cVariable)
  { pcbHeader->AddText("extern XLATVAR axlatvar@L@[];\n");
    pcbSource->AddText("XLATVAR axlatvar@L@[] =\n{ ");
    pcbSource->SetLeftMargin(2);
    for(i = 0; i < cVariable; ++i)
    { psymV = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_VAR, i), SymVar);
      if (i) pcbSource->AddText(",\n");
      pcbSource->AddText("{ ");
      pcbSource->AddString(psymV->szName);
      pcbSource->AddText(", ");
      pcbSource->AddUInteger(psymV->u, TRUE);
      pcbSource->AddText("}");
    }
    pcbSource->SetLeftMargin(0);
    pcbSource->AddText("\n};\n\n");
  }
  else
  { pcbHeader->AddText("extern XLATVAR *axlatvar@L@;\n");
    pcbSource->AddText("XLATVAR *axlatvar@L@ = NULL;\n\n");
  }

  // write the array structure
  UINT cArray = symTable->cSymbols(TYPE_ARRAY);
  UINT acArrayDimension[3] = { 0, 0, 0 };
  pcbHeader->SetLeftMargin(0);
  pcbSource->SetLeftMargin(0);
  if (cArray)
  { pcbHeader->AddText("extern XLATARRAY axlatarray@L@[];\n\n");
    pcbSource->AddText("XLATARRAY axlatarray@L@[] =\n{ ");
    pcbSource->SetLeftMargin(2);
    for(i = 0; i < cArray; ++i)
    { psymA = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_ARRAY, i), SymArray);
      ++acArrayDimension[psymA->c - 1];
      if (i) pcbSource->AddText(",\n");
      pcbSource->AddText("{ ");
      pcbSource->AddString(psymA->szName);
      pcbSource->AddText(", ");
      pcbSource->AddUInteger(psymA->u, FALSE);
      pcbSource->AddText(", ");
      pcbSource->AddUInteger(psymA->c, FALSE);
      pcbSource->AddText(", { ");
      pcbSource->AddUInteger(psymA->ac[0], FALSE);
      pcbSource->AddText(", ");
      pcbSource->AddUInteger(psymA->ac[1], FALSE);
      pcbSource->AddText(", ");
      pcbSource->AddUInteger(psymA->ac[2], TRUE);
      pcbSource->AddText("} }");
    }
    pcbSource->SetLeftMargin(0);
    pcbSource->AddText("\n};\n\n");
  }
  else
  { pcbHeader->AddText("extern XLATARRAY *axlatarray@L@;\n\n");
    pcbSource->AddText("XLATARRAY *axlatarray@L@ = NULL;\n\n");
  }

  // write the constant, variable and array structure references
  pcbHeader->SetLeftMargin(0);
  pcbHeader->AddText("#define NUM_@U@_ICONST ");
  pcbHeader->AddUInteger(cIConstant);
  pcbHeader->AddText("\n#define NUM_@U@_RCONST ");
  pcbHeader->AddUInteger(cRConstant);
  pcbHeader->AddText("\n#define NUM_@U@_CONST  ");
  pcbHeader->AddUInteger(cConstant);
  pcbHeader->AddText("\n#define NUM_@U@_VAR    ");
  pcbHeader->AddUInteger(cVariable);
  pcbHeader->AddText("\n#define NUM_@U@_ARRAY  ");
  pcbHeader->AddUInteger(cArray);
  pcbHeader->AddText("\n");

  // write result header structure
  UINT cElement = pstatements->GetNumOfElem();
  UINT cMaxArrayDim = 0;
  for(i = 0; i < 3; ++i) if (acArrayDimension[i]) cMaxArrayDim = i + 1;
  UINT cNumOfIndices = (pszMeshFileName) ? cMaxArrayDim + 2 : cMaxArrayDim;
  pcbHeader->SetLeftMargin(0);
  pcbHeader->AddText("extern RESHEAD reshead@L@;\n\n");
  pcbSource->SetLeftMargin(0);
  pcbSource->AddText("RESHEAD reshead@L@ = {");
  pcbSource->SetLeftMargin(2);
  pcbSource->AddText(" \"SGFramework Result File  Version 1.0\\n\\x1a\", ");
  pcbSource->AddString(SIGNATURE_RESULT);
  pcbSource->AddText(", ");
  if (pszMeshFileName) pcbSource->AddString(pszMeshFileName);
  else		       pcbSource->AddText("\"\"");
  pcbSource->AddText(", ");
  pcbSource->AddUInteger(cConstant, FALSE);
  pcbSource->AddText(", ");
  pcbSource->AddUInteger(cVariable, FALSE);
  pcbSource->AddText(", ");
  pcbSource->AddUInteger(cArray, FALSE);
  pcbSource->AddText(", ");
  pcbSource->AddUInteger(cElement, FALSE);
  pcbSource->AddText(", ");
  for(i = 0; i < 3; ++i)
  { pcbSource->AddUInteger(acArrayDimension[i], FALSE);
    pcbSource->AddText(", ");
  }
  pcbSource->AddText("0 };\n\n");

  // write the numerical algorithm structure
  pcbHeader->SetLeftMargin(0);
  pcbHeader->AddText("extern NUMALG numalg@L@;\n\n");
  pcbSource->SetLeftMargin(0);
  pcbSource->AddText("NUMALG numalg@L@ = {");
  pcbSource->SetLeftMargin(2);
  pcbSource->AddInteger(pxb->nEIiter, FALSE);
  pcbSource->AddText(", ");
  pcbSource->AddReal(pxb->rERaccu, FALSE);
  pcbSource->AddText(", ");
  pcbSource->AddInteger(pxb->nEIdamp, FALSE);
  pcbSource->AddText(", ");
  pcbSource->AddUInteger(pxb->uILsolu, FALSE);
  pcbSource->AddText(", ");
  pcbSource->AddUInteger(pxb->uILsort, FALSE);
  pcbSource->AddText(", ");
  pcbSource->AddUInteger(pxb->uILfill, FALSE);
  pcbSource->AddText(", ");
  pcbSource->AddUInteger(pxb->uILprec, FALSE);
  pcbSource->AddText(", ");
  pcbSource->AddInteger(pxb->nEIlite, FALSE);
  pcbSource->AddText(", ");
  pcbSource->AddReal(pxb->rERlacc, FALSE);
  pcbSource->AddText(", ");
  pcbSource->AddReal(pxb->rERover, FALSE);
  pcbSource->AddText(", ");
  pcbSource->AddUInteger(pxb->uCBzero, FALSE);
  pcbSource->AddText(", ");
  pcbSource->AddReal(pxb->rERdist, FALSE);
  pcbSource->SetLeftMargin(0);
  pcbSource->AddText(" };\n\n");

  // write user-defined functions
  pcbHeader->AddText("void PCF_@U@ (BOOL);\n\n");
  for(i = 0; i < cUDFuncs; ++i)
  { SymUDFunc *psymUDFunc = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_UDFUNC, i), SymUDFunc);
    UINT cArg = psymUDFunc->c;
    UINT cLocal = psymUDFunc->cAssign;
    UINT *afIsPresent;
    BOOL *afConst;
    REAL *arConst;
    MathNode **apmn;

    if (cLocal)
    {
      // allocate and initialize the local variable present flag array
      c = (cArg + 1) * cLocal;
      afIsPresent = new UINT[c];
      afConst	  = new BOOL[c];
      arConst	  = new REAL[c];
      apmn   = new MathNode*[c];
      for(j = 0; j < c; ++j)
      { afIsPresent[j] = 0;
	afConst[j] = FALSE;
      }

      // for each local assignment statement
      for(j = 0; j < cLocal; ++j)
      { MathNode *pmn = psymUDFunc->apsym[j]->pmn;
	if (pmn->uExpr & FLAG_CONST)
	{ afConst[j] = TRUE;
	  arConst[j] = (pmn->uExpr & FLAG_REAL) ? pmn->data.r : (REAL) pmn->data.n;
	}
	pmn->FindLocalVars(afIsPresent);
	apmn[j] = pmn;

	// differentiate the assignment statement with respect to all of the
	// arguments and write the result
	MathNode *pmnDer = pmn->Diff();
	for(k = 0; k < cArg; ++k)
	{ MathNode *pmnT;
	  if (psymUDFunc->af[k])
	  { pmnT = pmnDer->Copy();
	    pmnT->SimplifyDiff(k, cLocal, afConst, arConst);
	  }
	  else pmnT = new MathNode(0.0, -1u);
	  l = (k + 1) * cLocal + j;
	  if (pmnT->uExpr & FLAG_CONST)
	  { afConst[l] = TRUE;
	    arConst[l] = (pmnT->uExpr & FLAG_REAL) ? pmnT->data.r : (REAL) pmnT->data.n;
	  }
	  pmnT->FindLocalVars(afIsPresent + (k + 1) * cLocal);
	  apmn[l] = pmnT;
	}
	delete pmnDer;
      }

      // if a derivative refers to a local variable and this local
      // variable refers to another local variable, the latter local
      // local variable will not be tagged in the afIsPresent array.
      // we need to tag these local variables ...
      UINT *afTemp = new UINT[cLocal];
      for(k = 1; k <= cArg; ++k)
      { // initialize the afTemp array
	UINT *af = afIsPresent + k * cLocal;
	for(j = 0; j < cLocal; ++j)
	  afTemp[j] = af[j] & LVAR_VAR;

	// find and tag the local variables that were missed
	BOOL fDone;
	do
	{ for(j = 0; j < cLocal; ++j)
	    if (afTemp[j] & LVAR_VAR) apmn[j]->FindLocalVars(afTemp);
	  fDone = TRUE;
	  for(j = 0; j < cLocal; ++j)
	    if (afTemp[j] & LVAR_VAR)
	    { if (af[j] & LVAR_VAR) afTemp[j] &= ~LVAR_VAR;
	      else { af[j] |= LVAR_VAR; fDone = FALSE; }
	    }
	} while (!fDone);
      }
      delete [] afTemp;

    }

    // write the user-defined functions and derivatives
    for(j = 0; j <= cArg; ++j)
    { if (j && !psymUDFunc->af[j-1]) continue;

      pcbHeader->SetLeftMargin(0);
      pcbHeader->SetRightMargin(250);
      pcbHeader->AddText("REAL ");
      pcbHeader->AddUDFunction(psymUDFunc->szName, j, FALSE);
      if (cLocal)
      { pcbSource->SetLeftMargin(0);
	pcbSource->SetRightMargin(250);
	pcbSource->AddText("REAL ");
	pcbSource->AddUDFunction(psymUDFunc->szName, j, FALSE);
	pcbSource->SetLeftMargin(2);
      }

      for(k = 0; k < cArg; ++k)
      { pcbHeader->AddChar((k) ? ',' : '(');
	pcbHeader->AddText("REAL ");
	pcbHeader->AddArgument(k, FALSE);
	if (cLocal)
	{ pcbSource->AddChar((k) ? ',' : '(');
	  pcbSource->AddText("REAL ");
	  pcbSource->AddArgument(k, FALSE);
	}
      }

      pcbHeader->SetLeftMargin(0);
      pcbHeader->AddText(");\n");
      pcbHeader->SetRightMargin(78);

      if (cLocal)
      {
	pcbSource->SetLeftMargin(0);
	pcbSource->AddText(")\n{\n");
	pcbSource->SetRightMargin(78);

	for(k = 0; k < cLocal - 1; ++k)
	{ if (afIsPresent[j * cLocal + k] & LVAR_VAR)
	  { pcbSource->SetLeftMargin(2);
	    pcbSource->AddText("REAL ");
	    pcbSource->AddLocalVariable(k, TRUE);
	    pcbSource->AddText("= ");
	    pcbSource->SetLeftMargin(4);
	    pcbSource->AddMathNode(apmn[k]);
	    pcbSource->AddText(";\n\n");
	  }
	}

	for(k = 0; k < cLocal - 1; ++k)
	{ if (afIsPresent[j * cLocal + k] & LVAR_DER)
	  { pcbSource->SetLeftMargin(2);
	    pcbSource->AddText("REAL ");
	    pcbSource->AddLocalDerivative(k, TRUE);
	    pcbSource->AddText("= ");
	    pcbSource->SetLeftMargin(4);
	    pcbSource->AddMathNode(apmn[j * cLocal + k]);
	    pcbSource->AddText(";\n\n");
	  }
	}

	pcbSource->SetLeftMargin(2);
	pcbSource->AddText("return ");
	pcbSource->SetLeftMargin(4);
	pcbSource->AddMathNode(apmn[j * cLocal + cLocal - 1]);
	pcbSource->AddText(";\n");
	pcbSource->SetLeftMargin(0);
	pcbSource->AddText("}\n\n");

      }
    }

    if (cLocal)
    {
      // deallocate the local variable present flag array
      for(j = cLocal; j < c; ++j) delete apmn[j];
      delete [] apmn;
      delete [] afIsPresent;
      delete [] afConst;
      delete [] arConst;

    }
  }

  if (cUDFuncs) pcbHeader->AddChar('\n');

  // write equations and their derivatives
  if (cEquations)
  { pstatements->CodeEquations((void *) pcbSource);
    pcbHeader->SetLeftMargin(0);
    pcbHeader->AddText("extern PFEQU apfequ_@L@[];\n"
		       "extern PFPD apfpd_@L@[];\n\n");
    pcbSource->SetLeftMargin(0);
    pcbSource->AddText("PFEQU apfequ_@L@[] = { ");
    pcbSource->SetLeftMargin(2);
    for(i = 0; i < cEquations; ++i)
    { if (i) pcbSource->AddText(", ");
      pcbSource->AddEquation(i, FALSE);
    }
    pcbSource->AddText(" };\n\n");
    pcbSource->SetLeftMargin(0);
    pcbSource->AddText("PFPD apfpd_@L@[] = { ");
    pcbSource->SetLeftMargin(2);
    for(i = 0; i < cEquations; ++i)
    { if (i) pcbSource->AddText(", ");
      pcbSource->AddPartial(i, FALSE);
    }
    pcbSource->AddText(" };\n\n");
  }

  // write procedures
  FormFileName(pszSource, FILE_RESULT);
  UINT cProc = symTable->cSymbols(TYPE_PROC);
  for(i = 0; i < cProc; ++i)
  { SymProc *psymProc = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_PROC, i), SymProc);
    pcbHeader->SetLeftMargin(0);
    pcbHeader->AddText("void ");
    pcbHeader->AddProcedure(psymProc->szName, FALSE);
    pcbHeader->AddText(";\n");
    if (!psymProc->fExt)
    { pcbSource->SetLeftMargin(0);
      pcbSource->AddText("void ");
      pcbSource->AddProcedure(psymProc->szName, FALSE);
      pcbSource->AddText("\n{\n");
      pcbSource->SetLeftMargin(2);
      for(j = 0; j < cNumOfIndices; ++j)
      { pcbSource->AddText((j) ? ", " : "UINT ");
	pcbSource->AddIndex(j, FALSE);
      }
      pcbSource->AddText(";\n\n");
      pstatements->CodeProcedure(szResultFile, (void *) pcbSource, psymProc->i);
      pcbSource->SetLeftMargin(0);
      pcbSource->AddText("}\n\n");
    }
  }
  pcbHeader->SetLeftMargin(0);
  pcbHeader->AddText("void ");
  pcbHeader->AddProcedure("main", FALSE);
  pcbHeader->AddText(";\n");
  pcbSource->SetLeftMargin(0);
  pcbSource->AddText("void ");
  pcbSource->AddProcedure("main", FALSE);
  pcbSource->AddText("\n{\n");
  pcbSource->SetLeftMargin(2);
  for(j = 0; j < cNumOfIndices; ++j)
  { pcbSource->AddText((j) ? ", " : "UINT ");
    pcbSource->AddIndex(j, FALSE);
  }
  pcbSource->AddText(";\n\n");
  pstatements->CodeProcedure(szResultFile, (void *) pcbSource, iMainProc);
  pcbSource->SetLeftMargin(0);
  pcbSource->AddText("}\n\n");

  // write precompute function procedure
  pcbSource->SetLeftMargin(0);
  pcbSource->AddText("void PCF_@U@ (BOOL fDerivatives)\n{\n");
  pcbSource->SetLeftMargin(2);
  if (cMaxNumOfPrecFuncArg)
    for(i = 0; i < 3; ++i)
    { pcbSource->AddText("UINT ");
      pcbSource->AddIndex(i,FALSE);
      pcbSource->AddText(";\n");
    }
  for(i = 0; i < cMaxNumOfPrecFuncArg; ++i)
  { pcbSource->AddText("REAL ");
    pcbSource->AddArgument(i,FALSE);
    pcbSource->AddText(";\n");
  }
  pcbSource->AddChar('\n');
  if (cNodePrecomputeFunc)
  { pcbSource->SetLeftMargin(2);
    pcbSource->AddText("for(");
    pcbSource->AddIndex(0, FALSE);
    pcbSource->AddText(" = 0; ");
    pcbSource->AddIndex(0, FALSE);
    pcbSource->AddText(" < ");
    pcbSource->AddUInteger(cNode, FALSE);
    pcbSource->AddText("; ++");
    pcbSource->AddIndex(0, FALSE);
    pcbSource->AddText(")\n{ ");
    for(i = 0; i < cUDFuncs; ++i)
    { SymUDFunc *psymUDFunc = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_UDFUNC, i), SymUDFunc);
      MathNode *pmnTemplate = psymUDFunc->pmnTemplate;
      if (pmnTemplate && psymUDFunc->fNode)
      { UINT c = psymUDFunc->c;
	BOOL fDerivatives = FALSE;
	for(j = 0; j < c; ++j)
	{ pcbSource->SetLeftMargin(4);
	  pcbSource->AddArgument(j, FALSE);
	  pcbSource->AddText(" = ");
	  pcbSource->SetLeftMargin(6);
	  pcbSource->AddMathNode(pmnTemplate->apmn[j], FALSE, TRUE);
	  pcbSource->AddText(";\n");
	  if (psymUDFunc->auArg1[j] != -1u) fDerivatives = TRUE;
	}
	pcbSource->SetLeftMargin(4);
	pcbSource->AddPrecUDFunc(i, FALSE);
	pcbSource->AddChar('[');
	pcbSource->AddIndex(0, FALSE);
	pcbSource->AddChar(']');
	if (fDerivatives) pcbSource->AddText("[0]");
	pcbSource->AddText(" = ");
	pcbSource->SetLeftMargin(6);
	pcbSource->AddUDFunction(i, 0, FALSE);
	pcbSource->AddChar('(');
	for(k = 0; k < c; ++k)
	{ if (k) pcbSource->AddText(", ");
	  pcbSource->AddArgument(k, FALSE);
	}
	pcbSource->AddText(");\n");
	if (fDerivatives)
	{ pcbSource->SetLeftMargin(4);
	  pcbSource->AddText("if (fDerivatives)\n{ ");
	}
	for(j = 0; j < c; ++j)
	{ if (psymUDFunc->af[j])
	  { pcbSource->SetLeftMargin(6);
	    pcbSource->AddPrecUDFunc(i, FALSE);
	    pcbSource->AddChar('[');
	    pcbSource->AddIndex(0, FALSE);
	    pcbSource->AddChar(']');
	    pcbSource->AddChar('[');
	    pcbSource->AddUInteger(psymUDFunc->auArg1[j], FALSE);
	    pcbSource->AddChar(']');
	    pcbSource->AddText(" = ");
	    pcbSource->SetLeftMargin(8);
	    pcbSource->AddUDFunction(i, j+1, FALSE);
	    pcbSource->AddChar('(');
	    for(k = 0; k < c; ++k)
	    { if (k) pcbSource->AddText(", ");
	      pcbSource->AddArgument(k, FALSE);
	    }
	    pcbSource->AddText(");\n");
	  }
	}
	if (fDerivatives)
	{ pcbSource->SetLeftMargin(4);
	  pcbSource->AddText("}\n");
	}
	pcbSource->AddChar('\n');
      }
    }
    pcbSource->SetLeftMargin(2);
    pcbSource->AddText("}\n\n");
  }
  if (cEdgePrecomputeFunc)
  { pcbSource->SetLeftMargin(2);
    pcbSource->AddText("for(");
    pcbSource->AddIndex(0, FALSE);
    pcbSource->AddText(" = 0; ");
    pcbSource->AddIndex(0, FALSE);
    pcbSource->AddText(" < ");
    pcbSource->AddUInteger(cEdge, FALSE);
    pcbSource->AddText("; ++");
    pcbSource->AddIndex(0, FALSE);
    pcbSource->AddText(")\n{ _init_edge_nodes(");
    pcbSource->AddIndex(0, FALSE);
    pcbSource->AddText(", &");
    pcbSource->AddIndex(1, FALSE);
    pcbSource->AddText(", &");
    pcbSource->AddIndex(2, FALSE);
    pcbSource->AddText(");\n\n");

    for(i = 0; i < cUDFuncs; ++i)
    { SymUDFunc *psymUDFunc = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_UDFUNC, i), SymUDFunc);
      MathNode *pmnTemplate = psymUDFunc->pmnTemplate;
      if (pmnTemplate && !psymUDFunc->fNode)
      { UINT c = psymUDFunc->c;
	BOOL fDerivatives = FALSE;
	for(j = 0; j < c; ++j)
	{ pcbSource->SetLeftMargin(4);
	  pcbSource->AddArgument(j, FALSE);
	  pcbSource->AddText(" = ");
	  pcbSource->SetLeftMargin(6);
	  pcbSource->AddMathNode(pmnTemplate->apmn[j], FALSE, TRUE);
	  pcbSource->AddText(";\n");
	  if (psymUDFunc->auArg1[j] != -1u) fDerivatives = TRUE;
	}
	pcbSource->SetLeftMargin(4);
	pcbSource->AddPrecUDFunc(i, FALSE);
	pcbSource->AddChar('[');
	pcbSource->AddIndex(0, FALSE);
	pcbSource->AddChar(']');
	if (fDerivatives) pcbSource->AddText("[0]");
	pcbSource->AddText(" = ");
	pcbSource->SetLeftMargin(6);
	pcbSource->AddUDFunction(i, 0, FALSE);
	pcbSource->AddChar('(');
	for(k = 0; k < c; ++k)
	{ if (k) pcbSource->AddText(", ");
	  pcbSource->AddArgument(k, FALSE);
	}
	pcbSource->AddText(");\n");
	if (fDerivatives)
	{ pcbSource->SetLeftMargin(4);
	  pcbSource->AddText("if (fDerivatives)\n{ ");
	}
	for(j = 0; j < c; ++j)
	{ if (psymUDFunc->af[j])
	  { pcbSource->SetLeftMargin(6);
	    pcbSource->AddPrecUDFunc(i, FALSE);
	    pcbSource->AddChar('[');
	    pcbSource->AddIndex(0, FALSE);
	    pcbSource->AddChar(']');
	    pcbSource->AddChar('[');
	    pcbSource->AddUInteger(psymUDFunc->auArg1[j], FALSE);
	    pcbSource->AddChar(']');
	    pcbSource->AddText(" = ");
	    pcbSource->SetLeftMargin(8);
	    pcbSource->AddUDFunction(i, j+1, FALSE);
	    pcbSource->AddChar('(');
	    for(k = 0; k < c; ++k)
	    { if (k) pcbSource->AddText(", ");
	      pcbSource->AddArgument(k, FALSE);
	    }
	    pcbSource->AddText(");\n");
	  }
	}
	if (fDerivatives)
	{ pcbSource->SetLeftMargin(4);
	  pcbSource->AddText("}\n");
	}
	pcbSource->AddChar('\n');
      }
    }
    pcbSource->SetLeftMargin(2);
    pcbSource->AddText("}\n");
  }
  pcbSource->SetLeftMargin(0);
  pcbSource->AddText("}\n\n\n");

  // add default main procedure
  pcbSource->SetLeftMargin(0);
  pcbSource->AddText("#if !defined(EXTMAIN)\n"
		     "int main (UINT cArg, char **apszArg)\n"
		     "{\n");
  pcbSource->SetLeftMargin(2);
  pcbSource->AddText("error = new ErrorHandler();\n"
		     "pcmd = new CmdParse(cArg, apszArg);\n"
		     "if (!nParseCommandLine(&numalg@L@))\n"
		     "{ ");
  pcbSource->SetLeftMargin(4);
  pcbSource->SetRightMargin(250);
  if (cEquations)
  { pcbSource->AddText("ReadTopFile(");
    pcbSource->AddString(szTopFile);
    pcbSource->AddText(", &cTop_@L@, &cTag_@L@, &cEqu_@L@, &cElem_@L@, "
		       "(UINT *) &tag@L@, &aiTop_@L@,"
		       "&atag_@L@, &atop_@L@, &aanEI_@L@);\n");
    pcbSource->AddText("ReadPermFile(");
    pcbSource->AddString(szPermFile);
    pcbSource->AddText(", cTag_@L@, &aiPermRow_@L@, &aiPermCol_@L@);\n");
    pcbSource->AddText("pnzHeap = new NonZeroHeap();\n");
    pcbSource->AddText("InitNumAlg(&plinsolve_@L@, &pmtrx_@L@, "
		       "&pvectDX_@L@, &pvectRHS_@L@, cTag_@L@, "
		       "aiPermRow_@L@, aiPermCol_@L@, aiTop_@L@, atop_@L@);\n");
  }
  if (pszMeshFileName)
  { pcbSource->AddText("rDistScale = numalg@L@.rERdist;\n");
    pcbSource->AddText("InitCoordArrays(");
    pcbSource->AddArray(pszX, FALSE, FALSE);
    pcbSource->AddText(", ");
    pcbSource->AddArray(pszY, FALSE, FALSE);
    pcbSource->AddText(");\n");
    pcbSource->AddText("ReadMeshFile(");
    pcbSource->AddString(pszMeshFileName);
    pcbSource->AddText(", apr@U@_MeshVar);\n");
  }
  pcbSource->AddText("InitResultFile(");
  pcbSource->AddString(szResultFile);
  pcbSource->AddText(", &reshead@L@, NUM_@U@_ICONST, NUM_@U@_RCONST, "
		     "NUM_@U@_VAR, NUM_@U@_ARRAY, axlatconst@L@, axlatvar@L@, "
		     "axlatarray@L@);\n"
		     "PROC@U@_main();\n"
		     "if (!reshead@L@.cSet) WriteElements(");
  pcbSource->AddString(szResultFile);
  pcbSource->AddText(", &reshead@L@, (REAL *) &var@L@);\n");
  if (cEquations)
  { pcbSource->AddText("CleanUp(aiTop_@L@, atag_@L@, atop_@L@, aanEI_@L@, "
		       "aiPermRow_@L@, aiPermCol_@L@, pmtrx_@L@, "
		       "pvectRHS_@L@, pvectDX_@L@, plinsolve_@L@);\n");
    pcbSource->AddText("delete pnzHeap;\n");
  }
  pcbSource->SetLeftMargin(2);
  pcbSource->SetRightMargin(78);
  pcbSource->AddText("}\n"
		     "delete pcmd;\n"
		     "delete error;\n"
		     "return 0;\n");
  pcbSource->SetLeftMargin(0);
  pcbSource->AddText("}\n"
		     "#endif\n\n");

  // deallocate the header and source instances of the code generation class
  delete pcbHeader;
  delete pcbSource;
  return 0;
}
