/*-----------------------------------------------------------------------------

 PROJECT:    SIMULATION GENERATION FRAMEWORK (SGFramework)

 SUBSYSTEM:  sgxlat.exe
 FILE:	     index.cpp
 AUTHOR:     Kevin M. Kramer

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 04/15/94    KMK      initial implementation
 10/20/94    KMK      modified for use in command line version


 DESCRIPTION:

 This module contains the index and indices class definitions

-----------------------------------------------------------------------------*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "global.h"

#ifndef UNIX
#pragma hdrstop
#endif

#include <stdlib.h>
#include "index.h"
#include "statemnt.h"
#include "symbol.h"


///////////////////////////////////////////////////////////////////////////////
////////////////////////////////  INDEX CLASS  ////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Function:	 Index::Index
// Scope:	 Public, Constructor
Index::Index (UINT iI)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("index.cpp Index::Index\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  uType	    = INDEX_RANGE;
  fLoopOver = FALSE;
  fAll	    = TRUE;
  iIndex    = iI;
  nB	    = 0;
  nE	    = 0;
  nS	    = 0;
  n	    = 0;
  iList	    = -1u;
  c	    = 0;
  an	    = NULL;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Index::Index
// Scope:	 Public, Constructor
Index::Index (UINT iI, int nBegin)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("index.cpp Index::Index\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  uType	    = INDEX_RANGE;
  fLoopOver = FALSE;
  fAll	    = FALSE;
  iIndex    = iI;
  iLabel    = -1u;
  nB	    = nBegin;
  nE	    = nBegin;
  nS	    = 1;
  n	    = nBegin;
  iList	    = -1u;
  c	    = 0;
  an	    = NULL;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Index::Index
// Scope:	 Public, Constructor
Index::Index (UINT iI, int nBegin, int nEnd)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("index.cpp Index::Index\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  uType	    = INDEX_RANGE;
  fLoopOver = FALSE;
  fAll	    = FALSE;
  iIndex    = iI;
  iLabel    = -1u;
  nB	    = nBegin;
  nE	    = nEnd;
  nS	    = (nEnd - nBegin >= 0) ? 1 : -1;
  n	    = nBegin;
  iList	    = -1u;
  c	    = 0;
  an	    = NULL;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Index::Index
// Scope:	 Public, Constructor
Index::Index (UINT iI, int nBegin, int nEnd, int nStep)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("index.cpp Index::Index\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  uType	    = INDEX_RANGE;
  fLoopOver = FALSE;
  fAll	    = FALSE;
  iLabel    = -1u;
  iIndex    = iI;
  nB	    = nBegin;
  nE	    = nEnd;
  nS	    = nStep;
  n	    = nBegin;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Index::Index
// Scope:	 Public, Constructor
Index::Index (UINT iI, UINT iL, UINT cIndice, int *anIndice)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("index.cpp Index::Index\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  uType	    = INDEX_LABEL;
  fLoopOver = FALSE;
  fAll	    = FALSE;
  iLabel    = iL;
  iIndex    = iI;
  nB	    = anIndice[0];
  nE	    = anIndice[cIndice-1];
  nS	    = 0;
  n	    = nB;
  iList	    = 0;
  c	    = cIndice;
  an	    = anIndice;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Index::Index
// Scope:	 Public, Constructor
Index::Index (UINT iI, UINT iNode, UINT iL, BOOL fTriSum)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("index.cpp Index::Index\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  uType	    = INDEX_NODE;
  iIndex    = iI;
  iLabel    = iL;
  iList	    = -1u;
  an	    = (int *) aauNeighbors[iNode];
  c	    = (fTriSum) ? (UINT) an[1] : (UINT) an[0];
  nB	    = 0;
  nE	    = 0;
  nS	    = 0;
  n	    = 0;
  fLoopOver = FALSE;
  fAll	    = FALSE;

}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Index::SetAll
// Scope:	 Public
// Description:	 This function initializes the index for ALL
void Index::SetAll(UINT c)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("index.cpp Index::SetAll\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if (fAll)
  { nB = 0;
    nE = (int) c - 1;
    nS = 1;
  }
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Index::Init
// Scope:	 Public
// Description:	 This function initializes the index
void Index::Init()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("index.cpp Index::Init\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  fLoopOver = FALSE;

  switch (uType)
  { case INDEX_RANGE:
      n = nB;
      break;

    case INDEX_LABEL:
      n = nB;
      iList = 0;
      break;

    case INDEX_NODE:
      n = -1;
      while (++n < (int) c)
	if ((iLabel == -1u) || (iLabel == (UINT) an[2*n+3])) break;
      break;
  }
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Index::Next
// Scope:	 Public
// Description:	 This function increments the index to the next value
void Index::Next()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("index.cpp Index::Next\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  switch (uType)
  { case INDEX_RANGE:
      n += nS;
      fLoopOver = ((nS > 0) && (n > nE)) || ((nS < 0) && (n < nE));
      break;

    case INDEX_LABEL:
      ++iList;
      fLoopOver = (iList >= c);
      n = (fLoopOver) ? -1 : an[iList];
      break;

    case INDEX_NODE:
      while (++n < (int) c)
	if ((iLabel == -1u) || (iLabel == (UINT) an[2*n+3])) break;
      fLoopOver = (UINT) n >= c;
      break;
  }
}


///////////////////////////////////////////////////////////////////////////////
///////////////////////////////	 INDICES CLASS	///////////////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Function:	 Indices::Indices
// Scope:	 Public, Constructor
Indices::Indices ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("index.cpp Indices::Indices\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  i = -1u;
  c = 0;
  ac[0] = ac[1] = ac[2] = -1u;
  apindex = new Index*[5];
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Indices::~Indices
// Scope:	 Public, Constructor
Indices::~Indices ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("index.cpp Indices::~Indices\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;
  for(i = 0; i < c; ++i)
    if (apindex[i])
      delete apindex[i];
  delete [] apindex;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Indices::Element
// Scope:	 Public
// Description:	 This function returns the offset of the current array element
UINT Indices::Element ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("index.cpp Indices::Element\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;
  UINT uOffset = GetCurrent(0);
  for(i = 1; i < c; ++i)
  { uOffset *= ac[i];
    uOffset += GetCurrent(i);
  }
  return uOffset + u;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Indices::BeginOK
// Scope:	 Public
// Description:	 This function validates the begin value of index i
BOOL Indices::BeginOK (UINT i)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("index.cpp Indices::BeginOK\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if (GetType(i) != INDEX_RANGE) return TRUE;
  int n = apindex[i]->GetBegin();
  return (n >= 0) && (n < ac[i]);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Indices::EndOK
// Scope:	 Public
// Description:	 This function validates the end value of index i
BOOL Indices::EndOK (UINT i)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("index.cpp Indices::EndOK\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if (GetType(i) != INDEX_RANGE) return TRUE;
  int n = apindex[i]->GetEnd();
  return (n >= 0) && (n < ac[i]);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Indices::StepOK
// Scope:	 Public
// Description:	 This function validates the step value of index i
BOOL Indices::StepOK (UINT i)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("index.cpp Indices::StepOK\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if (GetType(i) != INDEX_RANGE) return TRUE;

  int n	 = apindex[i]->GetEnd() - apindex[i]->GetBegin();
  int nS = apindex[i]->GetStep();

  if ((n >= 0) && (nS > 0)) return TRUE;
  if ((n <= 0) && (nS < 0)) return TRUE;
  return FALSE;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Indices::Array
// Scope:	 Public
// Description:	 This function initializes the array fields.
void Indices::Array (UINT iArray, UINT uOffset, UINT *acDim)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("index.cpp Indices::Array\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT j;
  i = iArray;
  u = uOffset;
  ac[0] = acDim[0];
  ac[1] = acDim[1];
  ac[2] = acDim[2];
  for(j = 0; j < c; ++j)
    if (apindex[j]->IsAll()) apindex[j]->SetAll(ac[j]);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Indices::AddIndex
// Scope:	 Public
// Description:	 This function adds an index instance to the indices instance.
void Indices::AddIndex (Index * pindex)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("index.cpp Indices::AddIndex\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT j = c++;
//  allocation is now done in constructor
//  apindex = (Index **) realloc(apindex, c * sizeof(Index *));
  apindex[j] = pindex;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Indices::Init
// Scope:	 Public
// Description:	 This function initializes the index classes
void Indices::Init ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("index.cpp Indices::Init\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT j;
  for(j = 0; j < c; ++j) apindex[j]->Init();
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Indices::Next
// Scope:	 Public
// Description:	 This function increments the indices to their next values
void Indices::Next ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("index.cpp Indices::Next\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  switch (c)
  { case 3:
      apindex[2]->Next();
      if (apindex[2]->IsLoopOver()) apindex[2]->Init();
      else break;

    case 2:
      apindex[1]->Next();
      if (apindex[1]->IsLoopOver()) apindex[1]->Init();
      else break;

    case 1:
      apindex[0]->Next();
      break;
  }
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Indices::NSUMInit
// Scope:	 Public
// Description:	 This function initializes the node summation
void Indices::NSUMInit (void *pv)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("index.cpp Indices::NSUMInit\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if (!pv) { iNSUM = -1u; return; }

  MathNode *pmn = (MathNode *) pv;
  pmn->apmn[0]->Evaluate();
  UINT iNode   = (UINT) pstatements->PopNum();
  UINT iLabel  = pmn->data.nsum.iLabel;
  UINT fTriSum = pmn->data.nsum.fTriSum;

  if (iLabel < cRegLabel)
  { if (iNode < cBndryNode) iNode = cNode + iNode * cRegLabel + iLabel;
    iLabel = -1u;
  }

  iNSUM = (iLSUM == -1u) ? c : c + 1;
  apindex[iNSUM] = new Index(iNSUM, iNode, iLabel, fTriSum);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Indices::LSUMInit
// Scope:	 Public
// Description:	 This function initializes the label summation
void Indices::LSUMInit (void *pv)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("index.cpp Indices::LSUMInit\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if (!pv) { iLSUM = -1u; return; }

  MathNode *pmn = (MathNode *) pv;
  UINT iLabel = pmn->data.lsum.iLabel;

  iLSUM = c;
  SymLabel *psym = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_LABEL, iLabel), SymLabel);
  apindex[iLSUM] = new Index(iLSUM, psym->i, psym->c, (int *) psym->au);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Indices::SUMRemove
// Scope:	 Public
// Description:	 This function removes the label and node sum indices
void Indices::SUMRemove ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("index.cpp Indices::SUMRemove\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if (iNSUM != -1u) { delete apindex[iNSUM]; apindex[iNSUM] = NULL; }
  if (iLSUM != -1u) { delete apindex[iLSUM]; apindex[iLSUM] = NULL; }
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Indices::SUMInit
// Scope:	 Public
// Description:	 This function initializes the label and node sum indices
void Indices::SUMInit ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("index.cpp Indices::SUMInit\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  fSumOver = FALSE;
  if (iNSUM != -1u) apindex[iNSUM]->Init();
  if (iLSUM != -1u) apindex[iLSUM]->Init();
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Indices::SUMNext
// Scope:	 Public
// Description:	 This function loops to the next set of label and node sum
//		 indices
void Indices::SUMNext ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("index.cpp Indices::SUMNext\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if	  ((iNSUM == -1u) && (iLSUM == -1u)) fSumOver = TRUE;
  else if ((iNSUM != -1u) && (iLSUM == -1u)) apindex[iNSUM]->Next();
  else if ((iNSUM == -1u) && (iLSUM != -1u)) apindex[iLSUM]->Next();
  else
  { apindex[iNSUM]->Next();
    if (apindex[iNSUM]->IsLoopOver())
    { apindex[iNSUM]->Init();
      apindex[iLSUM]->Next();
    }
  }
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Indices::SUMOver
// Scope:	 Public
// Description:	 This function returns TRUE if the node and label sum loops are
//		 over; otherwise FALSE is returned.
BOOL Indices::SUMOver ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("index.cpp Indices::SUMOver\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if	  ((iNSUM == -1u) && (iLSUM == -1u)) return fSumOver;
  else if ((iNSUM != -1u) && (iLSUM == -1u)) return apindex[iNSUM]->IsLoopOver();
  else					     return apindex[iLSUM]->IsLoopOver();
}



