/*-----------------------------------------------------------------------------

 PROJECT:    SIMULATION GENERATION FRAMEWORK (SGFramework)

 SUBSYSTEM:  sgxlat.exe
 FILE:	     index.h
 AUTHOR:     Kevin M. Kramer

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 04/15/94    KMK      initial implementation
 10/20/94    KMK      modified for use in command line version

 DESCRIPTION:

 This module contains the index and indices class definitions

-----------------------------------------------------------------------------*/


#ifndef __index_h
#define __index_h

#define INDEX_RANGE 0
#define INDEX_LABEL 1
#define INDEX_NODE  2

class Index
{ private:
    UINT uType;				// type of index
    UINT iIndex;			// index's index
    UINT iPtr;				// buffer index
    UINT iLabel;			// label index
    UINT iList;				// list index
    UINT c;				// number of indices in the list
    BOOL fLoopOver;			// loop over flag
    BOOL fAll;				// ALL keyword specified
    int	 nB;				// begin index value
    int	 nE;				// end index value
    int	 nS;				// step value
    int	 n;				// current index value
    int *an;				// list of indices

  public:
    inline UINT GetIndex      ()       { return iIndex;	   }
    inline UINT GetLabelIndex ()       { return iLabel;	   }
    inline UINT GetBufPtr     ()       { return iPtr;	   }
    inline int	GetBegin      ()       { return nB;	   }
    inline int	GetEnd	      ()       { return nE;	   }
    inline int	GetStep	      ()       { return nS;	   }
    inline int	GetCurrent    ()       { return n;	   }
    inline BOOL IsLoopOver    ()       { return fLoopOver; }
    inline BOOL IsAll	      ()       { return fAll;	   }
    inline UINT GetType	      ()       { return uType;	   }
    inline void SetBufPtr     (UINT u) { iPtr = u;	   }

    void SetAll (UINT);			// initialize index for ALL
    void Init ();			// initialize index
    void Next ();			// set index to next value

  public:				// constructors
    Index (UINT);			// range index
    Index (UINT, int);			// range index
    Index (UINT, int, int);		// range index
    Index (UINT, int, int, int);	// range index
    Index (UINT, UINT, UINT, int *);	// label index
    Index (UINT, UINT, UINT, BOOL);	// node index
};


class Indices
{ private:
    UINT i;				// array index
    UINT u;				// array offset
    UINT c;				// number of dimensions
    UINT iPtr;				// buffer pointer
    UINT ac[5];				// elements per dimension
    Index **apindex;			// array of index pointers

  private:
    UINT iNSUM;				// node sum index
    UINT iLSUM;				// label sum index
    BOOL fSumOver;			// sum over flag

  public:
    void NSUMInit  (void *);		// initialize node summation
    void LSUMInit  (void *);		// initialize label summation
    void SUMRemove ();			// remove the summation indices
    void SUMInit   ();			// initialize summation loop(s)
    void SUMNext   ();			// goto next set of summation indices
    BOOL SUMOver   ();			// summation(s) over

  public:
    inline UINT GetArray      ()	{ return i; }
    inline UINT GetDim	      ()	{ return c; }
    inline UINT GetIndex      (UINT i)	{ return apindex[i]->GetIndex(); }
    inline UINT GetLabelIndex (UINT i)	{ return apindex[i]->GetLabelIndex(); }
    inline int	GetBegin      (UINT i)	{ return apindex[i]->GetBegin();   }
    inline int	GetEnd	      (UINT i)	{ return apindex[i]->GetEnd();	   }
    inline int	GetStep	      (UINT i)	{ return apindex[i]->GetStep();	   }
    inline int	GetCurrent    (UINT i)	{ return apindex[i]->GetCurrent(); }
    inline BOOL IsDefined     (UINT i)	{ return apindex[i] != NULL;	   }
    inline BOOL IsAll	      (UINT i)	{ return apindex[i]->IsAll();	   }
    inline UINT GetType	      (UINT i)	{ return apindex[i]->GetType();	   }
    inline BOOL IsLoopOver    ()	{ return apindex[0]->IsLoopOver(); }
    inline void SetAll	      (UINT i)	{ apindex[i]->SetAll(ac[i]);	   }
    inline void SetBufPtr     (UINT u)	{ iPtr = u;			   }
    inline UINT GetBufPtr     (UINT i)	{ return (i == -1u) ? iPtr : apindex[i]->GetBufPtr(); }

    UINT Element ();			// return current array element
    BOOL BeginOK (UINT);		// is begin value of spec. index ok?
    BOOL EndOK (UINT);			// is end value of spec. index ok?
    BOOL StepOK (UINT);			// is step value of spec. index ok?
    void Array (UINT, UINT, UINT *);	// initialize the array information
    void AddIndex (Index *);		// add specified index
    void Init ();			// initialize index
    void Next ();			// set index to next value

  public:				// constructor and destructor
     Indices();
    ~Indices();
};


#endif
