/*-----------------------------------------------------------------------------

 PROJECT:  SIMULATION GENERATION FRAMEWORK (SGFramework)

 SUBSYSTEM:  sgxlat.exe
 FILE:       main.cpp
 AUTHOR:     Kevin M. Kramer

 REVISION HISTORY
 DATE     INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 10/20/94    KMK      initial implementation
 09/25/99    JJF      line by line auditing and debug

 DESCRIPTION:

 This module contains the main procedure

-----------------------------------------------------------------------------*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "global.h"

#ifndef UNIX
#pragma hdrstop
#endif

#include "source.h"
#include "cmdparse.h"
#include "error.h"

#define RETCODE_OK     0
#define RETCODE_HELP   1
#define RETCODE_ERROR  2
#define RETCODE_FATAL  3

// global definitions
char *pszGlobalBuf;
CMDARGS cmdargs;

char *pszMeshFileName;
UINT **aauNeighbors;
UINT cNode;
UINT cEdge;
UINT cAdjInfo;
UINT cRegLabel;
UINT cBndryNode;


// help information
char *apszHelp[] =
{ "syntax:  sgxlat [options] file",
  " -h        display help screen on the standard output device",
  " -p##      floating point precision where ## = 1 to 16 (default = 6)",
  " -nc       no coupling between equations",
  NULL
};



int main ( UINT cArg, char **apszArg )
{
  int ReturnCode;

  // allocate an instance of the command line parser
  CmdParse *pcmd = new CmdParse(cArg, apszArg);

  // parse the command line arguments
  fprintf(stdout, "SG Translator       Version 1.0  (c) 1997  K. Kramer\n");
  if (pcmd->fArgExist("h") || (pcmd->GetNumOfPosArgs() != 1))
  { for(char **ppsz = apszHelp; *ppsz; ++ppsz)
      fprintf(stdout, "%s\n", *ppsz);
    ReturnCode = RETCODE_HELP;
  }
  else
  { pszGlobalBuf = new char[GLOBAL_BUFSIZE];
    error = new ErrorHandler();

    // translate the source file
    cmdargs.cPrecision  = (UINT) pcmd->nGetSwitchArg ("p", 16, 1, 16);
    cmdargs.fNoCoupling = pcmd->fArgExist("nc");
    Source *psource = new Source(pcmd->pszGetPosArg(0, NULL));
    psource->Translate();
    delete psource;
    ReturnCode = (error->fFatal()) ? RETCODE_FATAL :
                ((error->fError()) ? RETCODE_ERROR :
                                     RETCODE_OK);
    delete error;
    delete [] pszGlobalBuf;
  }

  // clean up and return
  delete pcmd;
  return ReturnCode;
}


// functions not defined in UNIX C++
#if defined(UNIX)

// convert string to lower case
char *strlwr ( char *pszString )
{
  char ch;
  char *psz = pszString;
  while ((ch = *psz) != '\0')
  { if ((ch >= 'A') && (ch <= 'Z')) *psz += 'a' - 'A';
    ++psz;
  }
  return pszString;
}

// convert string to uppercase
char *strupr ( char *pszString )
{
  char ch;
  char *psz = pszString;
  while ((ch = *psz) != '\0')
  { if ((ch >= 'a') && (ch <= 'z')) *psz -= 'a' - 'A';
    ++psz;
  }
  return pszString;
}

// compare two string w/o regard for case
int stricmp ( char *psz1, char *psz2 )
{
  char ch1, ch2;
  do
  { ch1 = *psz1;
    ch2 = *psz2;
    if ((ch1 >= 'a') && (ch1 <= 'z')) ch1 -= 'a' - 'A';
    if ((ch2 >= 'a') && (ch2 <= 'z')) ch2 -= 'a' - 'A';
    if      (ch1 > ch2) return +1;
    else if (ch1 < ch2) return -1;
    ++psz1;
    ++psz2;
  } while (ch1 && ch2);

  return 0;
}

#endif
