/*-----------------------------------------------------------------------------

 PROJECT:  SIMULATION GENERATION FRAMEWORK (SGFramework)

 SUBSYSTEM:  sgxlat.exe
 FILE:	     numalg.h
 AUTHOR:     Kevin M. Kramer

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 07/28/94    KMK      initial implementation
 10/20/94    KMK      modified for use in command line version

 DESCRIPTION:

 This module contains the numerical algorithm transfer buffer class.

-----------------------------------------------------------------------------*/


#ifndef __numalg_h
#define __numalg_h


#define LINSOLVE_SOR	    0
#define LINSOLVE_PCG	    1
#define LINSOLVE_GE	    2

#define SORT_NONE	    0

#define FILL_ZERO	    0
#define FILL_ONE	    1
#define FILL_TWO	    2
#define FILL_THREE	    3
#define FILL_FOUR	    4
#define FILL_FIVE	    5
#define FILL_INFINITY	    6

#define PRECOND_NONE	    0
#define PRECOND_DIAG	    1
#define PRECOND_ILU	    2

#define DEF_04_EI_ITER	    100
#define DEF_04_ER_ACCU	    0.01
#define DEF_04_EI_DAMP	    3
#define DEF_04_IL_SOLU	    LINSOLVE_GE
#define DEF_04_IL_SORT	    SORT_NONE
#define DEF_04_IL_FILL	    FILL_INFINITY
#define DEF_04_IL_PREC	    PRECOND_ILU
#define DEF_04_EI_LITE	    1000
#define DEF_04_ER_LACC	    1e-6
#define DEF_04_ER_OVER	    1.0
#define DEF_04_CB_ZERO	    FALSE
#define DEF_04_ER_DIST	    1.0


// Numerical Algorithm Transfer Buffer
typedef struct tagXFERBUF04		// xb04
{ int  nEIiter;				// maximum number of Newton iterations
  REAL rERaccu;				// minimum Newton accuracy
  int  nEIdamp;				// damping factor
  UINT uILsolu;				// linear solution algorithm
  UINT uILsort;				// sorting algorithm
  UINT uILfill;				// maximum fill level
  UINT uILprec;				// preconditioner
  int  nEIlite;				// maximum number of linear iterations
  REAL rERlacc;				// minimum linear accuracy
  REAL rERover;				// overrelaxation parameter
  UINT uCBzero;				// zero initial guess
  REAL rERdist;				// distance scaling factor
} XFERBUF04;


// NumAlgXferBuf Class ... encapsulates the XFERBUF04 structure
class NumAlgXferBuf
{ protected:
    XFERBUF04 xb;

  public:
    void SetNewtonIterations  (int  n)	{ xb.nEIiter = n; }
    void SetNewtonAccuracy    (REAL r)	{ xb.rERaccu = r; }
    void SetNewtonDamping     (int  n)	{ xb.nEIdamp = n; }
    void SetLinearSolution    (UINT u)	{ xb.uILsolu = u; }
    void SetLinearSort	      (UINT u)	{ xb.uILsort = u; }
    void SetLinearFill	      (UINT u)	{ xb.uILfill = u; }
    void SetLinearPrecond     (UINT u)	{ xb.uILprec = u; }
    void SetLinearIterations  (int  n)	{ xb.nEIlite = n; }
    void SetLinearAccuracy    (REAL r)	{ xb.rERlacc = r; }
    void SetLinearOver	      (REAL r)	{ xb.rERover = r; }
    void SetLinearZero	      (UINT u)	{ xb.uCBzero = u; }
    void SetDistScale	      (REAL r)	{ xb.rERdist = r; }
    XFERBUF04 * pxbGetXferBuf ()	{ return &xb;	  }

  public:
    NumAlgXferBuf ();
};


extern NumAlgXferBuf *pNumAlgXferBuf;


#endif
