/*-----------------------------------------------------------------------------

 PROJECT:  SIMULATION GENERATION FRAMEWORK (SGFramework)

 SUBSYSTEM:  sgxlat.exe
 FILE:	     statemnt.cpp
 AUTHOR:     Kevin M. Kramer

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 04/18/94    KMK      initial implementation
 10/20/94    KMK      modified for use in command line version

 DESCRIPTION:

 This module implements the classes that store and process the equation
 and main procedure statements.

-----------------------------------------------------------------------------*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "global.h"

#ifndef UNIX
#pragma hdrstop
#endif

#include <limits.h>

#if defined(UNIX)
#include <unistd.h>
#else
#include <io.h>
#endif


#include "types.h"
#include "symbol.h"
#include "mathnode.h"
#include "index.h"
#include "error.h"
#include "statemnt.h"
#include "xlatinfo.h"
#include "codebuf.h"


// global declaration
Statements *pstatements = NULL;


///////////////////////////////////////////////////////////////////////////////
//////////////////  MEMBER FUNCTIONS OF THE STATEMENT CLASS  //////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Function:	 Statement::Statement
// Scope:	 Public, Constructor
Statement::Statement (UINT u, UINT i)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp Statement::Statement\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  uType = u;
  iPtr	= i;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Statement::Statement
// Scope:	 Public, Destructor
Statement::~Statement ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp Statement::Statement\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

}


///////////////////////////////////////////////////////////////////////////////
//////////////////  MEMBER FUNCTIONS OF THE VARASSIGN CLASS  //////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Function:	 VarAssign::VarAssign
// Scope:	 Public, Constructor
VarAssign::VarAssign (UINT iVar, UINT uOff, MathNode *pmn1, UINT iPtr)
	 : Statement (TYPE_VASSIGN, iPtr)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp VarAssign::VarAssign\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  i = iVar;
  u = uOff;
  pmn = pmn1;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 VarAssign::~VarAssign
// Scope:	 Public, Destructor
VarAssign::~VarAssign ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp VarAssign::~VarAssign\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  delete pmn;
}


///////////////////////////////////////////////////////////////////////////////
/////////////////  MEMBER FUNCTIONS OF THE ARRAYASSIGN CLASS  /////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Function:	 ArrayAssign::ArrayAssign
// Scope:	 Public, Constructor
ArrayAssign::ArrayAssign (UINT iArray, UINT uOff, Indices *pindices1, MathNode *pmn1, UINT iPtr)
	   : Statement	 (TYPE_AASSIGN, iPtr)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp ArrayAssign::ArrayAssign\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  i = iArray;
  u = uOff;
  pindices = pindices1;
  pmn = pmn1;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 ArrayAssign::~ArrayAssign
// Scope:	 Public, Destructor
ArrayAssign::~ArrayAssign ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp ArrayAssign::~ArrayAssign\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  delete pindices;
  delete pmn;
}


///////////////////////////////////////////////////////////////////////////////
////////////////////  MEMBER FUNCTIONS OF THE SOLVE CLASS  ////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Function:	 Solve::Solve
// Scope:	 Public, Constructor
Solve::Solve	 (UINT iPtr)
     : Statement (TYPE_SOLVE, iPtr)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp Solve::Solve\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Solve::~Solve
// Scope:	 Public, Destructor
Solve::~Solve ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp Solve::~Solve\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

}


///////////////////////////////////////////////////////////////////////////////
////////////////////  MEMBER FUNCTIONS OF THE WRITE CLASS  ////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Function:	 Write::Write
// Scope:	 Public, Constructor
Write::Write	 (UINT iPtr)
     : Statement (TYPE_WRITE, iPtr)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp Write::Write\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Write::~Write
// Scope:	 Public, Destructor
Write::~Write ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp Write::~Write\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

}


///////////////////////////////////////////////////////////////////////////////
////////////////////  MEMBER FUNCTIONS OF THE GOTO CLASS  /////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Function:	 Goto::Goto
// Scope:	 Public, Constructor
Goto::Goto	(MathNode *pmnCondition, UINT iStatement, UINT iPtr)
    : Statement (TYPE_GOTO, iPtr)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp Goto::Goto\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  pmn = pmnCondition;
  i = iStatement;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Goto::~Goto
// Scope:	 Public, Destructor
Goto::~Goto ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp Goto::~Goto\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if (pmn) delete pmn;
}


///////////////////////////////////////////////////////////////////////////////
////////////////////  MEMBER FUNCTIONS OF THE CALL CLASS  /////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Function:	 Call::Call
// Scope:	 Public, Constructor
Call::Call	(UINT iProcedure, UINT iStatement, UINT iPtr)
    : Statement (TYPE_CALL, iPtr)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp Call::Call\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  iProc = iProcedure;
  i = iStatement;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Call::~Call
// Scope:	 Public, Destructor
Call::~Call ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp Call::~Call\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

}


///////////////////////////////////////////////////////////////////////////////
//////////////////  MEMBER FUNCTIONS OF THE FILEOPEN CLASS  ///////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Function:	 FileOpen::FileOpen
// Scope:	 Public, Constructor
FileOpen::FileOpen  (char *pszFileName, UINT uMode, UINT iPtr)
	: Statement (TYPE_FILEOPEN, iPtr)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp FileOpen::FileOpen\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  psz = pszFileName;
  u = uMode;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 FileOpen::~FileOpen
// Scope:	 Public, Destructor
FileOpen::~FileOpen ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp FileOpen::~FileOpen\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if (psz) delete [] psz;
}


///////////////////////////////////////////////////////////////////////////////
/////////////////  MEMBER FUNCTIONS OF THE FILECLOSE CLASS  ///////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Function:	 FileClose::FileClose
// Scope:	 Public, Constructor
FileClose::FileClose (UINT iPtr)
	: Statement  (TYPE_FILECLOSE, iPtr)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp FileClose::FileClose\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

}


///////////////////////////////////////////////////////////////////////////////
// Function:	 FileClose::~FileClose
// Scope:	 Public, Destructor
FileClose::~FileClose ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp FileClose::~FileClose\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

}


///////////////////////////////////////////////////////////////////////////////
//////////////////  MEMBER FUNCTIONS OF THE FILEGOTO CLASS  ///////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Function:	 FileGoto::FileGoto
// Scope:	 Public, Constructor
FileGoto::FileGoto  (char *pszLabel, UINT iPtr)
	: Statement (TYPE_FILEGOTO, iPtr)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp FileGoto::FileGoto\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  psz = pszLabel;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 FileGoto::~FileGoto
// Scope:	 Public, Destructor
FileGoto::~FileGoto ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp FileGoto::~FileGoto\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if (psz) delete [] psz;
}


///////////////////////////////////////////////////////////////////////////////
/////////////////  MEMBER FUNCTIONS OF THE FILELABEL CLASS  ///////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Function:	 FileLabel::FileLabel
// Scope:	 Public, Constructor
FileLabel::FileLabel (char *pszLabel, UINT iPtr)
	: Statement  (TYPE_FILELABEL, iPtr)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp FileLabel::FileLabel\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  psz = pszLabel;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 FileLabel::~FileLabel
// Scope:	 Public, Destructor
FileLabel::~FileLabel ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp FileLabel::~FileLabel\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if (psz) delete [] psz;
}


///////////////////////////////////////////////////////////////////////////////
////////////////  MEMBER FUNCTIONS OF THE FILECOMMENT CLASS  //////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Function:	 FileComment::FileComment
// Scope:	 Public, Constructor
FileComment::FileComment (char *pszComment, UINT iPtr)
	: Statement	 (TYPE_FILECOMM, iPtr)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp FileComment::FileComment\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  psz = pszComment;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 FileComment::~FileComment
// Scope:	 Public, Destructor
FileComment::~FileComment ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp FileComment::~FileComment\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if (psz) delete [] psz;
}


///////////////////////////////////////////////////////////////////////////////
//////////////////  MEMBER FUNCTIONS OF THE FILEREAD CLASS  ///////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Function:	 FileRead::FileRead
// Scope:	 Public, Constructor
FileRead::FileRead  (UINT iVar, UINT uOffset, UINT iPtr)
	: Statement (TYPE_FILEREAD, iPtr)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp FileRead::FileRead\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  i = iVar;
  u = uOffset;
  c = 0;
  ac[0] = -1u;
  ac[1] = -1u;
  ac[2] = -1u;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 FileRead::FileRead
// Scope:	 Public, Constructor
FileRead::FileRead  (UINT iVar, UINT uOffset, UINT cDim, UINT *acDim, UINT iPtr)
	: Statement (TYPE_FILEREAD, iPtr)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp FileRead::FileRead\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  i = iVar;
  u = uOffset;
  c = cDim;
  ac[0] = acDim[0];
  ac[1] = acDim[1];
  ac[2] = acDim[2];
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 FileRead::~FileRead
// Scope:	 Public, Destructor
FileRead::~FileRead ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp FileRead::~FileRead\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

}


///////////////////////////////////////////////////////////////////////////////
/////////////////  MEMBER FUNCTIONS OF THE FILEWRITE CLASS  ///////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Function:	 FileWrite::FileWrite
// Scope:	 Public, Constructor
FileWrite::FileWrite (UINT iVar, UINT uOffset, UINT iPtr)
	: Statement  (TYPE_FILEWRITE, iPtr)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp FileWrite::FileWrite\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  i = iVar;
  u = uOffset;
  c = 0;
  ac[0] = -1u;
  ac[1] = -1u;
  ac[2] = -1u;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 FileWrite::FileWrite
// Scope:	 Public, Constructor
FileWrite::FileWrite (UINT iVar, UINT uOffset, UINT cDim, UINT *acDim, UINT iPtr)
	: Statement  (TYPE_FILEWRITE, iPtr)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp FileWrite::FileWrite\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  i = iVar;
  u = uOffset;
  c = cDim;
  ac[0] = acDim[0];
  ac[1] = acDim[1];
  ac[2] = acDim[2];
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 FileWrite::~FileWrite
// Scope:	 Public, Destructor
FileWrite::~FileWrite ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp FileWrite::~FileWrite\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

}


///////////////////////////////////////////////////////////////////////////////
/////////////////  MEMBER FUNCTIONS OF THE VAREQUATION CLASS  /////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Function:	 VarEquation::VarEquation
// Scope:	 Public, Constructor
VarEquation::VarEquation (UINT iVar, UINT uOff, MathNode *pmn1, UINT iPtr)
	   : Statement	 (TYPE_VEQU, iPtr)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp VarEquation::VarEquation\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  i = iVar;
  u = uOff;
  pmn = pmn1;
  cContrib = 0;
  apmnElem = NULL;
  apmnDiff = NULL;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 VarEquation::~VarEquation
// Scope:	 Public, Destructor
VarEquation::~VarEquation ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp VarEquation::~VarEquation\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;

  delete pmn;
  for(i = 0; i < cContrib; ++i)
  { delete apmnElem[i];
    delete apmnDiff[i];
  }
  if (apmnElem) delete [] apmnElem;
  if (apmnDiff) delete [] apmnDiff;
}


///////////////////////////////////////////////////////////////////////////////
////////////////  MEMBER FUNCTIONS OF THE ARRAYEQUATION CLASS  ////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Function:	 ArrayEquation::ArrayEquation
// Scope:	 Public, Constructor
ArrayEquation::ArrayEquation (UINT iArray, UINT uOff, Indices *pindices1, MathNode *pmn1, UINT iPtr)
	     : Statement     (TYPE_AEQU, iPtr)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp ArrayEquation::ArrayEquation\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  i = iArray;
  u = uOff;
  pindices = pindices1;
  pmn = pmn1;
  cContrib = 0;
  apmnElem = NULL;
  apmnDiff = NULL;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 ArrayEquation::~ArrayEquation
// Scope:	 Public, Destructor
ArrayEquation::~ArrayEquation ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp ArrayEquation::~ArrayEquation\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;

  delete pindices;
  delete pmn;
  for(i = 0; i < cContrib; ++i)
  { delete apmnElem[i];
    delete apmnDiff[i];
  }
  if (apmnElem) delete [] apmnElem;
  if (apmnDiff) delete [] apmnDiff;
}


///////////////////////////////////////////////////////////////////////////////
//////////////////  MEMBER FUNCTIONS OF THE STATEMENTS CLASS  /////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Function:	 Statements::Statements
// Scope:	 Public, Constructor
Statements::Statements ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp Statements::Statements\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  aiTag	 = NULL;
#if defined(P16BIT)
  aiTop	 = NULL;
  abfTag = NULL;
#endif
  atag	 = NULL;
  atop	 = NULL;
  aanEI	 = NULL;
  cTag	 = 0;
  cTop	 = 0;
  pindicesNULL = new Indices();
  DAPMainProc.SetManaged(TRUE);
  DAPEquation.SetManaged(TRUE);
  DAPMathExpr.SetManaged(FALSE);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Statements::~Statements
// Scope:	 Public, Destructor
Statements::~Statements ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp Statements::~Statements\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if (aiTag)  delete [] aiTag;
#if defined(P16BIT)
  if (aiTop)  delete [] aiTop;
  if (abfTag) delete [] abfTag;
#endif
  if (atag)   delete [] atag;
  if (atop)   delete [] atop;
  if (aanEI)  delete [] aanEI;
  delete pindicesNULL;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Statements::SetGotoIndex
// Scope:	 Public
// Description:	 This function sets the index of the specified goto statement.
void Statements::SetGotoIndex (UINT iStatement, UINT iIndex)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp Statements::SetGotoIndex\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  Statement *ps = DAPMainProc[iStatement];
  Goto *pgoto = TYPESAFE_DOWNCAST(ps, Goto);
  if (pgoto) pgoto->i = iIndex;
}

#if defined(P16BIT)
///////////////////////////////////////////////////////////////////////////////
// Function:	 Statements::AllocBitFlags
// Scope:	 Public
// Description:	 This function allocates the bit flags if they have not already
//		 been allocated.
void Statements::AllocBitFlags ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp Statements::AllocBitFlags\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;

  if (!abfTag)
  { UINT c = (MAX_VARIABLES >> 4) + 1;
    abfTag = new UINT[c];
    UINT *pu = abfTag;
    for(i = 0; i < c; ++i, ++pu) *pu = 0;
  }
}
#endif


#if !defined(P16BIT)
///////////////////////////////////////////////////////////////////////////////
// Function:	 Statements::AllocBitFlags
// Scope:	 Public
// Description:	 This function allocates the bit flags if they have not already
//		 been allocated.
void Statements::AllocBitFlags ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp Statements::AllocBitFlags\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;

  if (DAbfTag.GetElements() == 0)
    for(i = 0; i < DAUINT_INCSIZE; ++i)
      DAbfTag.Add(0);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Statements::SetBitFlag
// Scope:	 Public
// Description:	 This function sets a bit flag.
void Statements::SetBitFlag (UINT i)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp Statements::SetBitFlag\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT k;
  UINT j = i >> 4;
  while (j >= DAbfTag.GetElements())
    for(k = 0; k < DAUINT_INCSIZE; ++k) DAbfTag.Add(0);
  *(DAbfTag.GetPointer(j)) |= (1 << (i & 0x000F));
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Statements::ClearBitFlag
// Scope:	 Public
// Description:	 This function clear a bit flag.
void Statements::ClearBitFlag (UINT i)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp Statements::ClearBitFlag\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT k;
  UINT j = i >> 4;
  while (j >= DAbfTag.GetElements())
    for(k = 0; k < DAUINT_INCSIZE; ++k) DAbfTag.Add(0);
  *(DAbfTag.GetPointer(j)) &= ~(1 << (i & 0x000F));
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Statements::IsBitFlagSet
// Scope:	 Public
// Description:	 This function clear a bit flag.
BOOL Statements::IsBitFlagSet (UINT i)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp Statements::IsBitFlagSet\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT j = i >> 4;
  if (j >= DAbfTag.GetElements()) return FALSE;
  return DAbfTag[j] & (1 << (i & 0x000F));
}
#endif


///////////////////////////////////////////////////////////////////////////////
// Function:	 Statements::Operator
// Scope:	 Public
// Description:	 This function performs the specified operation.  Arguments
//		 are taken from the stack.
void Statements::Operator (UINT uOp, UINT iPtr)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp Statements::Operator\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  // pop arguments from the stack
  int n1, n2, n3;
  if	  ((uOp == TYPE_NEG) || (uOp == TYPE_NOT))
    n1 = stack.pop();
  else if ((uOp == TYPE_NODE) || (uOp == TYPE_EDGE))
  { n3 = stack.pop();
    n2 = stack.pop();
    n1 = stack.pop();
  }
  else
  { n2 = stack.pop();
    n1 = stack.pop();
  }

  // if the error flag is set, then push -1 on the stack
  if (fError) { stack.push(-1); return; }

  // check for division by zero
  if ((uOp == TYPE_DIV) && (n2 == 0))
  { error->AddMessage(EHM_DIVBYZERO, "/ operator", iPtr);
    fError = TRUE;
    stack.push(-1);
    return;
  }

  // compute the result
  long l;
  switch (uOp)
  { case TYPE_ADD:  l = (long) n1 + (long) n2;	break;
    case TYPE_SUB:  l = (long) n1 - (long) n2;	break;
    case TYPE_MUL:  l = (long) n1 * (long) n2;	break;
    case TYPE_DIV:  l = (long) n1 / (long) n2;	break;
    case TYPE_NEG:  l =		  - (long) n1;	break;
    case TYPE_GT:   l = (n1 >  n2) ? 1l : 0l;	break;
    case TYPE_GE:   l = (n1 >= n2) ? 1l : 0l;	break;
    case TYPE_LT:   l = (n1 <  n2) ? 1l : 0l;	break;
    case TYPE_LE:   l = (n1 <= n2) ? 1l : 0l;	break;
    case TYPE_EQ:   l = (n1 == n2) ? 1l : 0l;	break;
    case TYPE_NE:   l = (n1 != n2) ? 1l : 0l;	break;
    case TYPE_NOT:  l = (!n1	 ) ? 1l : 0l;	break;
    case TYPE_AND:  l = (n1 && n2) ? 1l : 0l;	break;
    case TYPE_OR:   l = (n1 || n2) ? 1l : 0l;	break;
    case TYPE_XOR:  l = ((!n1 && n2) || (n1 && !n2)) ? 1l : 0l;	 break;
    case TYPE_NODE:
      if ((n3 != -1) && (n1 < cBndryNode))
	n1 = cNode + cRegLabel * n1 + n3;
      n2 %= aauNeighbors[n1][0];
      l = (long) aauNeighbors[n1][2 * n2 + 2];
      break;
  }

  // check for overflow
  if ((l < (long) INT_MIN) || (l > (long) INT_MAX))
  { fError = TRUE;
    l = -1u;
  }

  // store the result
  stack.push((int) l);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Statements::VarOK
// Scope:	 Private
// Description:	 This function verifies the variable or array math node pmn.
BOOL Statements::VarOK (MathNode *pmn)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp Statements::VarOK\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;

  // variable math nodes do not need to be verified
  if (pmn->uNode != TYPE_ARRAY) return TRUE;

  // if the array math node has no children, then it refers to an array element
  // and has already been verified
  UINT c = pmn->c;
  if (c == 0) return TRUE;

  // check the array bounds
  BOOL fError = FALSE;
  int n[3];
  for (i = 0; i < c; ++i)
  { if (pmn->apmn[i]->uNode == TYPE_ICONST) n[i] = pmn->apmn[i]->data.n;
    else
    { ClearFlag();
      pmn->apmn[i]->Evaluate();
      n[i] = stack.pop();
      if ((n[i] < 0) || (n[i] >= pmn->data.array.ac[i]))
      { char *psz = symTable->Retrieve(TYPE_ARRAY, pmn->data.array.i)->szName;
	sprintf(pszGlobalBuf, "%d", i + 1);
	error->AddMessage(EHM_ARRAYDIM5, pszGlobalBuf, psz, pmn->apmn[i]->iPtrExpr);
	fError = TRUE;
      }
    }
  }

  // compute element offset
  UINT u = (fError) ? -1u : (UINT) n[0];
  if (!fError)
  { for(i = 1; i < c; ++i)
    { u *= pmn->data.array.ac[i];
      u += (UINT) n[i];
    }
    u += pmn->data.array.uA;
  }

  // store the element offset and return the result
  pmn->data.array.uE = u;
  return !fError;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Statements::ProcessMain
// Scope:	 Public
// Description:	 This function processes the statements of the main procedure
BOOL Statements::ProcessMain ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp Statements::ProcessMain\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i, j;

  // initialize local flag variables
  BOOL fSolve  = FALSE;
  BOOL fResult = TRUE;

  // loop through each statement
  UINT cStatements = DAPMainProc.GetElements();
  for(i = 0; i < cStatements; ++i)
  { Statement *ps = DAPMainProc[i];

    // set the solve flag if a solve statement was declared
    if (ps->uType == TYPE_SOLVE) fSolve = TRUE;

    // check the variable assigment statements
    else if (ps->uType == TYPE_VASSIGN)
    { VarAssign *pva = TYPESAFE_DOWNCAST(ps, VarAssign);
      pva->pmn->TagSumVariables();
      pva->pmn->Variables();
      pindices = pindicesNULL;

      // loop through each variable in the assignment expression
      UINT cVariables = DAPMathExpr.GetElements();
      for(j = 0; j < cVariables; ++j)
      { BOOL fError = FALSE;

	// consider only array math nodes w/ children
	MathNode *pmn = DAPMathExpr[j];
	if ((pmn->uNode != TYPE_ARRAY) || (pmn->c == 0)) continue;
	pindices->LSUMInit(pmn->data.array.pmnLSUM);
	pindices->NSUMInit(pmn->data.array.pmnNSUM);
	for(pindices->SUMInit(); !pindices->SUMOver(); pindices->SUMNext())
	  fError = fError || !VarOK(pmn);
	pindices->SUMRemove();
	if (fError) fResult = FALSE;
      }

      // remove the variables from the dynamic array
      DAPMathExpr.Flush();
    }

    // check the array assignment statements
    else if (ps->uType == TYPE_AASSIGN)
    { ArrayAssign *paa = TYPESAFE_DOWNCAST(ps, ArrayAssign);
      paa->pmn->TagSumVariables();
      paa->pmn->Variables();
      pindices = paa->pindices;

      // loop through each variable in the assignment expression
      UINT cVariables = DAPMathExpr.GetElements();
      for(j = 0; j < cVariables; ++j)
      { BOOL fError = FALSE;

	// consider only array math nodes w/ children
	MathNode *pmn = DAPMathExpr[j];
	if ((pmn->uNode != TYPE_ARRAY) || (pmn->c == 0)) continue;

	// loop through each set of indices and verify the array
	for (pindices->Init(); !fError && !pindices->IsLoopOver(); pindices->Next())
	{ pindices->LSUMInit(pmn->data.array.pmnLSUM);
	  pindices->NSUMInit(pmn->data.array.pmnNSUM);
	  for(pindices->SUMInit(); !pindices->SUMOver(); pindices->SUMNext())
	    fError = fError || !VarOK(pmn);
	  pindices->SUMRemove();
	}
	if (fError) fResult = FALSE;
      }

      // remove the variables from the dynamic array
      DAPMathExpr.Flush();
    }

  }

  // if no main procedure statements were declared then report an error
  if (!cStatements)
  { error->AddMessage(EHM_NOMAINPROC);
    fResult = FALSE;
  }

  // if the solve statement is present and not equations are declared, then
  // report an error
  if (fSolve && !DAPEquation.GetElements())
  { error->AddMessage(EHM_NOEQUATION);
    fResult = FALSE;
  }

  return fResult;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Statements::Allocate
// Scope:	 Private
// Description:	 This function allocates the array data structures.
void Statements::Allocate ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp Statements::Allocate\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;

  aiTag = new UINT[cElem];
#if defined(P16BIT)
  aiTop = new UINT[MAX_DENSITY];
#endif
  atag	= new TAG[cElem];
  atop	= new TOP[cElem];
  aanEI = new I3ARRAY[cElem];
  for(i = 0; i < cElem; ++i) aiTag[i] = -1u;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 pindicesMake
// Scope:	 Private
// Description:	 This function makes an instance of the indices class that
//		 spans the entire range of the passed array.
static Indices * pindicesMake (SymArray *psym)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp pindicesMake\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;

  Indices *pindices = new Indices();
  UINT c = psym->c;
  for(i = 0; i < c; ++i) pindices->AddIndex(new Index(i));
  pindices->Array(psym->i, psym->u, psym->ac);
  return pindices;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Statements::nTag1
// Scope:	 Private
// Description:	 This function tags unknown variables and array elements
//		 based upon the bit flags.  This function returns 0 if
//		 successful and nonzero is an error occurs.
int Statements::nTag1 ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp Statements::nTag1\n", fileDEBUG);
  fclose(fileDEBUG);
#endif


  UINT i, j;

  // tag unknown variables and array elements
  UINT c = symTable->cSymbols(TYPE_VAR);
  for (i = 0; i < c; ++i)
  { SymVar *psymV = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_VAR, i), SymVar);
    UINT iVar = psymV->u;
    if (IsBitFlagSet(iVar))
    { aiTag[iVar] = cTag;
      atag[cTag].iVar = iVar;
      atag[cTag].iEqu = -1u;
      atag[cTag].iArray = -1u;
      aanEI[cTag][0] = -1;
      aanEI[cTag][1] = -1;
      aanEI[cTag][2] = -1;
      ++cTag;
    }
  }

  c = symTable->cSymbols(TYPE_ARRAY);
  for (i = 0; i < c; ++i)
  { SymArray *psymA = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_ARRAY, i), SymArray);
    Indices *pindices = pindicesMake(psymA);
    for (pindices->Init(); !pindices->IsLoopOver(); pindices->Next())
    { UINT iVar = pindices->Element();
      if (IsBitFlagSet(iVar))
      { aiTag[iVar] = cTag;
	atag[cTag].iVar = iVar;
	atag[cTag].iEqu = -1u;
	atag[cTag].iArray = psymA->i;
	UINT cDim = pindices->GetDim();
	aanEI[cTag][0] =	      pindices->GetCurrent(0);
	aanEI[cTag][1] = (cDim > 1) ? pindices->GetCurrent(1) : -1;
	aanEI[cTag][2] = (cDim > 2) ? pindices->GetCurrent(2) : -1;
	++cTag;
      }
    }
    delete pindices;
  }

  // loop through the equation and clear the corresponding bit flags
  UINT cStatements = DAPEquation.GetElements();
  for(i = 0; i < cStatements; ++i)
  { Statement *ps = DAPEquation[i];

    // variable equation
    if (ps->uType == TYPE_VEQU)
    { VarEquation *pve = TYPESAFE_DOWNCAST(ps, VarEquation);
      ClearBitFlag(pve->u);
    }

    // array equation
    else
    { ArrayEquation *pae = TYPESAFE_DOWNCAST(ps, ArrayEquation);
      pindices = pae->pindices;
      for(pindices->Init(); !pindices->IsLoopOver(); pindices->Next())
	ClearBitFlag(pindices->Element());
    }
  }

  // make sure each unknown has a corresponding equation
  BOOL fUnderSpecified = FALSE;
  c = symTable->cSymbols(TYPE_VAR);
  for (i = 0; i < c; ++i)
  { SymVar *psymV = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_VAR, i), SymVar);
    UINT iVar = psymV->u;

    // if an unknown variable does not have a corresponding equation, then
    // report a warning
    if (IsBitFlagSet(iVar))
    { error->AddMessage(EHM_NOEQU, psymV->szName);
      fUnderSpecified = TRUE;
    }
  }

  c = symTable->cSymbols(TYPE_ARRAY);
  for (i = 0; i < c; ++i)
  { SymArray *psymA = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_ARRAY, i), SymArray);
    Indices *pindices = pindicesMake(psymA);
    for (pindices->Init(); !pindices->IsLoopOver(); pindices->Next())
    { UINT iVar = pindices->Element();
      if (IsBitFlagSet(iVar))
      { strcpy(pszGlobalBuf, psymA->szName);
	for(j = 0; j < psymA->c; ++j)
	{ char sz[16];
	  sprintf(sz, "[%d]", pindices->GetCurrent(j));
	  strcat(pszGlobalBuf, sz);
	}
	error->AddMessage(EHM_NOEQU, pszGlobalBuf);
	fUnderSpecified = TRUE;
      }
    }
    delete pindices;
  }

  // report an error if the system of equation is underspecified and return
  if (fUnderSpecified) error->AddMessage(EHM_UNDERSPEC);
  return (fUnderSpecified) ? -1 : 0;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Statements::nTag2
// Scope:	 Private
// Description:	 This function tags unknown variables based upon the equations.
//		 This function returns zero if successful and nonzero if an
//		 error occurs.
int Statements::nTag2 ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp Statements::nTag2\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;

  // known/unknown statements not used ... determine unknown variables and
  // array elements from equation statements
  UINT cStatements = DAPEquation.GetElements();
  for(i = 0; i < cStatements; ++i)
  { Statement *ps = DAPEquation[i];

    // variable equation
    if (ps->uType == TYPE_VEQU)
    { VarEquation *pve = TYPESAFE_DOWNCAST(ps, VarEquation);
      UINT iVar = pve->u;
      if (aiTag[iVar] == -1u)
      { aiTag[iVar] = cTag;
	atag[cTag].iVar = iVar;
	atag[cTag].iEqu = -1u;
	atag[cTag].iArray = -1u;
	aanEI[cTag][0] = -1;
	aanEI[cTag][1] = -1;
	aanEI[cTag][2] = -1;
	++cTag;
      }
    }

    // array equation
    else
    { ArrayEquation *pae = TYPESAFE_DOWNCAST(ps, ArrayEquation);
      pindices = pae->pindices;

      // loop through each set of indices
      for (pindices->Init(); !pindices->IsLoopOver(); pindices->Next())
      { UINT iVar = pindices->Element();
	if (aiTag[iVar] == -1u)
	{ aiTag[iVar] = cTag;
	  atag[cTag].iVar = iVar;
	  atag[cTag].iEqu = -1u;
	  atag[cTag].iArray = pae->i;
	  UINT cDim = pindices->GetDim();
	  aanEI[cTag][0] =		pindices->GetCurrent(0);
	  aanEI[cTag][1] = (cDim > 1) ? pindices->GetCurrent(1) : -1;
	  aanEI[cTag][2] = (cDim > 2) ? pindices->GetCurrent(2) : -1;
	  ++cTag;
	}
      }
    }
  }

  return 0;
}


#define DIFFINFO_SUCCESS      0
#define DIFFINFO_NOINFO	     -1
#define DIFFINFO_OUTOFBOUNDS -2
#define DIFFINFO_NOPIVOT     -3


///////////////////////////////////////////////////////////////////////////////
// Function:	 Statements::nDiffInfo
// Scope:	 Private
// Description:	 This function verifies the array expressions in an equation
//		 and computes the topology information if the fInfo flag is
//		 set.  The following values are returned.
//
//		 DIFFINFO_SUCCESS	equation verified and tolopogy computed
//		 DIFFINFO_NOINFO	fInfo flag was not set
//		 DIFFINFO_OUTOFBOUNDS	out-of-bound error occurred
//		 DIFFINFO_NOPIVOT	equation does not have a pivot
//
int Statements::nDiffInfo (UINT iVar, BOOL fInfo)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp Statements::nDiffInfo\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;

  // initialize local variables
  UINT iTop = cTop;
  BOOL fPivot = FALSE;

  // loop through each variable or array in the expression
  UINT c = DAPMathExpr.GetElements();
  for(i = 0; i < c; ++i)
  { MathNode *pmn = DAPMathExpr[i];
    if ((pmn->uNode == TYPE_VAR) && (pmn->data.var.iT != -1u))
    { if (fInfo) fPivot |= AddTopologyIndex(iVar, pmn->data.var.u, iTop);
    }
    else if ((pmn->uNode == TYPE_ARRAY) && (pmn->data.array.uA != -1u))
    { pindices->LSUMInit(pmn->data.array.pmnLSUM);
      pindices->NSUMInit(pmn->data.array.pmnNSUM);
      for(pindices->SUMInit(); !pindices->SUMOver(); pindices->SUMNext())
      { if (!VarOK(pmn))
	{ pmn->data.array.uA = -1u;
	  return DIFFINFO_OUTOFBOUNDS;
	}
	else if (fInfo && (pmn->data.array.iT != -1u))
	  fPivot |= AddTopologyIndex(iVar, pmn->data.array.uE, iTop);
      }
      pindices->SUMRemove();
    }
  }

  // update the topology index array if appropriate
  if (fInfo)
  { atop[aiTag[iVar]].i = cTop;
    atop[aiTag[iVar]].c = iTop - cTop;
    cTop = iTop;
    return (fPivot) ? DIFFINFO_SUCCESS :  DIFFINFO_NOPIVOT;
  }

  return DIFFINFO_NOINFO;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Statements::AddTopologyIndex
// Scope:	 Private
// Description:	 This function adds the index to the topology array if the
//		 index is not duplicated in the current row.  TRUE is
//		 returned if the index is the pivot.
BOOL Statements::AddTopologyIndex (UINT iVar, UINT iNZ, UINT &iTop)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp Statements::AddTopologyIndex\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;

  // check if variable is an unknown
  UINT iUnknown = aiTag[iNZ];
  if (iUnknown == -1u) return FALSE;

  // check if the variable or array element is the pivot
  if (iNZ == iVar) return TRUE;

  // check if the variable or array element is already present in the
  // topology array
  for(i = cTop; i < iTop; ++i)
#if defined(P16BIT)
    if (aiTop[i] == iUnknown) return FALSE;
#else
    if (DAiTop[i] == iUnknown) return FALSE;
#endif

#if defined(P16BIT)
  // store the variable or array element in the topology array if it is
  // not already present
  if (iTop == MAX_DENSITY)
  { error->AddMessage(EHM_MAXDENSITY);
     return DIFFINFO_NOINFO;
  }

  aiTop[iTop++] = iUnknown;
#else
  DAiTop.Add(iUnknown);
  ++iTop;
#endif
  return FALSE;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Statements::ComputeContribs
// Scope:	 Public
// Description:	 This function computes the contributions to the elements of
//		 Jacobian matrix.  The math node dynamic pointer array must
//		 be initialized passed equation's variables and arrays prior
//		 to calling this function.
void Statements::ComputeContribs (Statement *ps, UINT iArray)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp Statements::ComputeContribs\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;

  // loop through each variable or array in the expression
  UINT c = DAPMathExpr.GetElements();
  UINT cContrib = 0;
  for(i = 0; i < c; ++i)
  { MathNode *pmn = DAPMathExpr[i];
    if	    ((pmn->uNode == TYPE_ARRAY) && (pmn->data.array.iT != -1u))
    { BOOL f = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_ARRAY, pmn->data.array.i), SymArray)->f;
      if (cmdargs.fNoCoupling && (pmn->data.array.i != iArray)) f = FALSE;
      pmn->data.array.iT = (f) ? cContrib++ : -1u;
    }
    else if ((pmn->uNode == TYPE_VAR) && (pmn->data.var.iT != -1u))
    { BOOL f = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_VAR, pmn->data.var.i), SymVar)->f;
      pmn->data.var.iT = (f) ? cContrib++ : -1u;
    }
  }

  MathNode **apmnElem = new MathNode*[cContrib];
  MathNode **apmnDiff = new MathNode*[cContrib];
  MathNode  *pmnD;
  VarEquation	*pve;
  ArrayEquation *pae;

  if (ps->uType == TYPE_VEQU)
  { pve	 = TYPESAFE_DOWNCAST(ps, VarEquation);
    pmnD = pve->pmn->Diff();
  }
  else
  { pae	 = TYPESAFE_DOWNCAST(ps, ArrayEquation);
    pmnD = pae->pmn->Diff();
  }

  for(i = 0; i < c; ++i)
  { MathNode *pmnE = DAPMathExpr[i];
    UINT iDiff = (pmnE->uNode == TYPE_ARRAY) ? pmnE->data.array.iT : pmnE->data.var.iT;
    if (iDiff != -1u)
    { apmnElem[iDiff] = pmnE->Copy();
      apmnDiff[iDiff] = pmnD->Copy();
      apmnDiff[iDiff]->SimplifyDiff(iDiff);
    }
  }

  if (ps->uType == TYPE_VEQU)
  { pve->cContrib = cContrib;
    pve->apmnElem = apmnElem;
    pve->apmnDiff = apmnDiff;
  }
  else
  { pae->cContrib = cContrib;
    pae->apmnElem = apmnElem;
    pae->apmnDiff = apmnDiff;
  }

  delete pmnD;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Statements::ProcessEquations
// Scope:	 Public
// Description:	 This function processes the equation statements.
BOOL Statements::ProcessEquations (UINT cVars)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp Statements::ProcessEquations\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;

  // allocate the topology and unknown arrays
  cElem = cVars;
  Allocate();

  // tag the unknown variables
#if defined(P16BIT)
  int nResult = (abfTag) ? nTag1() : nTag2();
#else
  int nResult = (DAbfTag.GetElements()) ? nTag1() : nTag2();
#endif
  if (nResult) return FALSE;

  // loop through each equation statement
  BOOL fResult = TRUE;
  UINT cStatements = DAPEquation.GetElements();
  for(i = 0; i < cStatements; ++i)
  { Statement *ps = DAPEquation[i];

    // variable equation
    if (ps->uType == TYPE_VEQU)
    { VarEquation *pve = TYPESAFE_DOWNCAST(ps, VarEquation);
      pindices = pindicesNULL;

      // if the variable already has an associated equation, then report a
      // warning and consider the next equation
      UINT iVar = pve->u;
      UINT iUnknown = aiTag[iVar];
      if (atag[iUnknown].iEqu != -1u)
      { char *psz = symTable->Retrieve(TYPE_VAR, pve->i)->szName;
	error->AddMessage(EHM_USELESS, "variable", psz, pve->iPtr);
	continue;
      }

      // locate the equation's variables and array elements and compute the
      // differentiation information
      pve->pmn->TagSumVariables();
      pve->pmn->Variables();
      ComputeContribs(ps, -1u);
      int nResult = nDiffInfo(iVar, fResult);
      if (nResult == DIFFINFO_NOPIVOT)
      { char *psz = symTable->Retrieve(TYPE_VAR, pve->i)->szName;
	error->AddMessage(EHM_NOPIVOT, "variable", psz, pve->iPtr);
      }
      fResult = nResult == DIFFINFO_SUCCESS;
      DAPMathExpr.Flush();
      atag[iUnknown].iEqu = i;
      continue;
    }

    // array equation ... locate the equations's variables and array elements
    // and set up the indices loop
    BOOL fUseless = TRUE;
    ArrayEquation *pae = TYPESAFE_DOWNCAST(ps, ArrayEquation);
    pindices = pae->pindices;
    pae->pmn->TagSumVariables();
    pae->pmn->Variables();
    ComputeContribs(ps, pae->i);

    // loop through each set of indices
    for (pindices->Init(); !pindices->IsLoopOver(); pindices->Next())
    { UINT iVar = pindices->Element();
      UINT iUnknown = aiTag[iVar];
      if (iUnknown != -1U)	  // @@@ FIXED BUG
      { if (atag[iUnknown].iEqu == -1u)
	{ // the array element does not have an associated equation
	  fUseless = FALSE;
	  int nResult = nDiffInfo(iVar, fResult);
	  if (nResult == DIFFINFO_NOPIVOT)
	  { char *psz = symTable->Retrieve(TYPE_ARRAY, pae->i)->szName;
	    error->AddMessage(EHM_NOPIVOT, "array", psz, pae->iPtr);
	  }
	  fResult = nResult == DIFFINFO_SUCCESS;
	  atag[iUnknown].iEqu = i;
	}
      }
    }
    DAPMathExpr.Flush();

    // if the equation was not used, then report a warning
    if (fUseless)
    { char *psz = symTable->Retrieve(TYPE_ARRAY, pae->i)->szName;
      error->AddMessage(EHM_USELESS, "array", psz, pae->iPtr);
    }
  }
  return fResult;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Statements::WriteTopFile
// Scope:	 Public
// Description:	 This function writes the topology file
void Statements::WriteTopFile (int nFile, TOPHEAD *phead)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp Statements::WriteTopFile\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  phead->cElement  = cElem;
  phead->cTag	   = cTag;
  phead->cTopology = cTop;
  phead->cEquation = GetNumOfEqu();

  write(nFile, phead, sizeof(TOPHEAD));
  write(nFile, aiTag,  cElem * sizeof(UINT));
#if defined(P16BIT)
  write(nFile, aiTop,  cTop  * sizeof(UINT));
#else
  write(nFile, DAiTop.GetPointer(0), cTop * sizeof(UINT));
#endif
  write(nFile, atag,   cTag  * sizeof(TAG));
  write(nFile, atop,   cTag  * sizeof(TOP));
  write(nFile, aanEI,  cTag  * sizeof(I3ARRAY));
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 Statements::TestForUserDefProc
// Scope:	 Public
// Description:	 This function initializes the user-defined flag of the
//		 last procedure symbol.
void Statements::TestForUserDefProc ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("statemnt.cpp Statements::TestForUserDefProc\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT iProc = symTable->cSymbols(TYPE_PROC) - 1;
  SymProc *psym = TYPESAFE_DOWNCAST(symTable->Retrieve(TYPE_PROC, iProc), SymProc);
  psym->fExt = (DAPMainProc[psym->i]->uType == TYPE_RETURN);
}


///////////////////////////////////////////////////////////////////////////////
///////////  DYNAMIC ARRAY OF POINTERS TO INSTANCES OF STATEMENTS  ////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Function:	 DAPStatement::DAPStatement
// Scope:	 Public, Constructor
DAPStatement::DAPStatement (UINT c)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("dynarray.cpp DAPStatement::DAPStatement\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  cElements = 0;
  fManaged = FALSE;
  dynarray = (Statement**) calloc((cSize = uInc = c), sizeof(Statement*));
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 DAPStatement::~DAPStatement
// Scope:	 Public, Destructor
DAPStatement::~DAPStatement ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("dynarray.cpp DAPStatement::~DAPStatement\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;

  // if the array is managed then deallocate the stored objects
  if (fManaged)
    for(i = 0; i < cElements; ++i)
      delete dynarray[i];

  free((void *) dynarray);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 DAPStatement::Add
// Scope:	 Public
// Description:	 This function adds an element to the dynamic array increases
//		 the array size if necessary.
UINT DAPStatement::Add (Statement *t)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("dynarray.cpp DAPStatement::Add\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  // increase the array size if necessary
  if (cElements == cSize)
  { cSize += uInc;
    dynarray = (Statement **) realloc(dynarray, cSize * sizeof(MathNode *));
  }

  dynarray[cElements] = t;
  return cElements++;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 DAPStatement::RemoveLast
// Scope:	 Public
// Description:	 This function removes the last element from the dynarmic array
void DAPStatement::RemoveLast ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("dynarray.cpp DAPStatement::RemoveLast\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  --cElements;

  // if the array is managed then deallocate the stored object
  if (fManaged) delete dynarray[cElements];
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 DAPStatement::Remove
// Scope:	 Public
// Description:	 This function removes an element from the dynarmic array
void DAPStatement::Remove (UINT iRemove)
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("dynarray.cpp DAPStatement::Remove\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;

  --cElements;
  if (fManaged) delete dynarray[iRemove];
  for(i = iRemove; i < cElements; ++i)
    dynarray[i] = dynarray[i+1];
  dynarray[cElements] = NULL;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 DAPStatement::Flush
// Scope:	 Public
// Description:	 This function adds an element to the dynamic array increases
//		 the array size if necessary.
void DAPStatement::Flush ()
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("dynarray.cpp DAPStatement::Flush\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;

  // if the array is managed then deallocate the stored objects
  if (fManaged)
    for(i = 0; i < cElements; ++i)
      delete dynarray[i];

  // reset the element count
  cElements = 0;
}


