#include <stdio.h>
#include <fcntl.h>

#if defined(UNIX)
#include <unistd.h>
#include <sys/stat.h>
#define	 O_BINARY 0x0000
#else
#include <io.h>
#include <sys\stat.h>
#endif


typedef unsigned int UINT;
typedef int	     I3ARRAY[3];

typedef struct tagTOPHEAD
{ char szLogo[64];			// text logo
  char szSign[16];			// signature
  UINT cElement;			// number of variables & array elements
  UINT cTag;				// number of unknown elements
  UINT cTopology;			// number of nonzeroes excluding pivots
  UINT cEquation;			// number of equations
} TOPHEAD;

typedef struct tagTAG			// tag
{ UINT iVar;				// element array index (i.e. offset)
  UINT iEqu;				// equation index
  UINT iArray;				// array index ... -1u == variable
} TAG;

typedef struct tagTOP			// top
{ UINT i;				// index in topology array
  UINT c;				// number of nonzeroes in equation
} TOP;



int main(int cArg, char *apszArg[])
{
  fprintf(stderr, "SGFramework Topology - SMMS	Version 1.0  Copyright (c) 1995 Kevin M. Kramer\n");
  if (cArg != 3)
  { fprintf(stderr, "usage:  top2smms top_file smms_matrix_file\n");
    return 1;
  }

  int nFile = open(apszArg[1], O_BINARY | O_RDONLY);
  if (nFile == -1)
  { fprintf(stderr, "cannot open file \"%s\"\n", apszArg[1]);
    return 2;
  }

  FILE *file = fopen(apszArg[2], "wt");
  if (file == NULL)
  { fprintf(stderr, "cannot create file \"%s\"\n", apszArg[2]);
    return 2;
  }

  TOPHEAD  head;
  read(nFile, &head, sizeof(TOPHEAD));
  UINT *aiTop = new UINT[head.cTopology];
  UINT *aiTag = new UINT[head.cElement];
  TAG *atag = new TAG[head.cTag];
  TOP *atop = new TOP[head.cTag];
  I3ARRAY *aanEI = new I3ARRAY[head.cTag];
  read(nFile, aiTag, head.cElement  * sizeof(UINT));
  read(nFile, aiTop, head.cTopology * sizeof(UINT));
  read(nFile, atag,  head.cTag * sizeof(TAG));
  read(nFile, atop,  head.cTag * sizeof(TOP));
  read(nFile, aanEI, head.cTag * sizeof(I3ARRAY));

  close(nFile);

  UINT i, j;
  fprintf(file, "%u\n", head.cTag);
  for(i = 0; i < head.cTag; ++i)
  { fprintf(file, "%u %u 1\n", i+1, i+1);
    UINT iB = atop[i].i;
    UINT iE = atop[i].c + iB;
    for(j = iB; j < iE; ++j)
      fprintf(file, "%u %u 1\n", i+1, aiTop[j]+1);
  }
  fprintf(file, "0 0 0\n");
  fclose(file);

  delete [] aiTop;
  delete [] aiTag;
  delete [] atag;
  delete [] atop;
  delete [] aanEI;
  return 0;
}

