/*-----------------------------------------------------------------------------

 PROJECT:  SIMULATION GENERATION FRAMEWORK (SGFramework)

 SUBSYSTEM:  sgxlat.exe
 FILE:	     types.h
 AUTHOR:     Kevin M. Kramer

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 02/18/94    KMK      initial implementation
 10/20/94    KMK      modified for use in command line version

 DESCRIPTION:

 This module contains the symbol, math node and operator type defintions.

-----------------------------------------------------------------------------*/


#ifndef __types_h
#define __types_h


#define TYPE_INDEX	    0		// index
#define TYPE_VAR	    1		// variable
#define TYPE_ARRAY	    2		// array
#define TYPE_RCONST	    3		// real constant
#define TYPE_ICONST	    4		// integer constant
#define TYPE_FUNC	    5		// function
#define TYPE_ARG	    6		// function argument
#define TYPE_LVAR	    7		// local variable
#define TYPE_LRCONST	    8		// local real constant
#define TYPE_LICONST	    9		// local integer constant
#define TYPE_UDFUNC	    10		// user-defined function
#define TYPE_PROC	    11		// procedure
#define TYPE_LABEL	    12		// label
#define TYPE_NSUM	    13		// node sum expression
#define TYPE_LSUM	    14		// label sum expression
#define SYMBOLS		    15		// number of symbol types
#define TYPE_ADD	    16		// addition math node
#define TYPE_SUB	    17		// subtraction math node
#define TYPE_MUL	    18		// multiplication math node
#define TYPE_DIV	    19		// division math node
#define TYPE_AND	    20		// logical and math node
#define TYPE_OR		    21		// logical or math node
#define TYPE_XOR	    22		// logical exclusive or math node
#define TYPE_GT		    23		// greater than math node
#define TYPE_GE		    24		// greater than and equal to math node
#define TYPE_LT		    25		// less than math node
#define TYPE_LE		    26		// less than and equal to math node
#define TYPE_EQ		    27		// equal to math node
#define TYPE_NE		    28		// not equal to math node
#define TYPE_NEG	    29		// negation math node
#define TYPE_NOT	    30		// logical not math node
#define TYPE_REAL	    31		// convert to real math node
#define TYPE_AREA	    32		// total area
#define TYPE_AREA1	    33		// partial area 1
#define TYPE_AREA2	    34		// partial area 2
#define TYPE_NODE	    35		// node index
#define TYPE_EDGE	    36		// edge index
#define TYPE_ELEN	    37		// edge length
#define TYPE_ELENABS	    38		// edge length ... 2 nodes given
#define TYPE_ILEN	    39		// integration segment length
#define TYPE_PRECOMP	    40		// precomputed information
#define TYPE_VASSIGN	    41		// variable assignment statement
#define TYPE_AASSIGN	    42		// array assignment statement
#define TYPE_SOLVE	    43		// solve statement
#define TYPE_WRITE	    44		// write statement
#define TYPE_VEQU	    45		// variable equation statement
#define TYPE_AEQU	    46		// array equation statement
#define TYPE_FILEOPEN	    47		// file open statement
#define TYPE_FILECLOSE	    48		// file close statement
#define TYPE_FILEGOTO	    49		// file goto statement
#define TYPE_FILELABEL	    50		// file label statement
#define TYPE_FILECOMM	    51		// file comment statement
#define TYPE_FILEREAD	    52		// file read statement
#define TYPE_FILEWRITE	    53		// file write statement
#define TYPE_GOTO	    54		// conditional goto statement
#define TYPE_CALL	    55		// call statement
#define TYPE_RETURN	    56		// return statement
#define TYPE_END	    57		// end of program

#define FLAG_REAL	    0x0001	// real flag
#define FLAG_INT	    0x0002	// integer flag
#define FLAG_CONST	    0x0004	// constant flag
#define FLAG_VAR	    0x0008	// variable flag
#define FLAG_ODD	    0x0200	// odd flag
#define FLAG_EVEN	    0x0400	// even flag
#define FLAG_TEST	    0x0800	// cannot determine odd or even
#define FLAG_LSUM	    0x1000	// label sum expression
#define FLAG_NSUM	    0x2000	// node sum expression
#define FLAG_ZEROCHECK	    0x4000	// "0.0 *" check
#define FLAG_DIFF	    0x8000	// differentiation flag

#define FLAG_SUM	    (FLAG_LSUM | FLAG_NSUM)
#define MASK_UNKNOWN	    (FLAG_ODD | FLAG_EVEN | FLAG_TEST)


#endif
