/*-----------------------------------------------------------------------------

 PROJECT:  SIMULATION GENERATION FRAMEWORK (SGFramework)

 SUBSYSTEM:  sgxlat.exe
 FILE:	     xlatinfo.h
 AUTHOR:     Kevin M. Kramer

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 04/22/94    KMK      initial implementation

 DESCRIPTION:

 This module contains the xlat structures that are shared among header files

-----------------------------------------------------------------------------*/


#ifndef __xlatinfo_h
#define __xlatinfo_h


#ifndef __numalg_h
#include "numalg.h"
#endif


// integer array
typedef int I3ARRAY[3];			// i3a


// translate header structure definition
typedef struct tagXLATHEAD
{ char szLogo[64];			// text logo
  char szSign[16];			// signature
  long lconstants;			// constant list file index
  long lvariables;			// variable list file index
  long larrays;				// array list file index
  long ludfuncs;			// user-defined function list file index
  long linfo;				// unknown variables and topology info
  long lequations;			// equation statements file index
  long lmainproc;			// main procedure statements file index
  UINT iMainProc;			// statement index of main procedure
  UINT cConstant;			// number of constants
  UINT cVariable;			// number of variables
  UINT cArray;				// number of arrays
  UINT cElement;			// number of variables & array elements
  UINT cUDFunc;				// number of user-defined functions
  UINT cTag;				// number of unknown elements
  UINT cTopology;			// number of nonzeroes excluding pivots
  UINT cEquation;			// number of equations
  UINT cMainProc;			// number of main procedure statements
  XFERBUF04 xbNumAlg;			// numerical algorithms xfer buffer
} XLATHEAD;


// result header structure definition
typedef struct tagRESHEAD
{ char szLogo[64];			// text logo
  char szSign[16];			// signature
  char szMeshFile[128];			// name of mesh file
  UINT cConstant;			// number of constants
  UINT cVariable;			// number of variables
  UINT cArray;				// number of arrays
  UINT cElement;			// number of elements
  UINT c1DArray;			// number of 1D arrays
  UINT c2DArray;			// number of 2D arrays
  UINT c3DArray;			// number of 3D arrays
  UINT cSet;				// number of data sets
} RESHEAD;


// tag (unknown) and topology file
typedef struct tagTOPHEAD
{ char szLogo[64];			// text logo
  char szSign[16];			// signature
  UINT cElement;			// number of variables & array elements
  UINT cTag;				// number of unknown elements
  UINT cTopology;			// number of nonzeroes excluding pivots
  UINT cEquation;			// number of equations
} TOPHEAD;


// mesh header
typedef struct tagMESHHEAD
{ char szLogo[64];			// text logo
  char szSign[16];			// signature
  UINT cConstant;			// number of constants
  UINT cLabel;				// number of labels
  UINT cArray;				// number of 1D arrays
  UINT cRegLabel;			// number of region labels
  UINT cBndryNode;			// number of boundary nodes
  UINT cNode;				// number of nodes
  UINT cEdge;				// number of edges
  UINT cTriangle;			// number of triangles
  BOOL fCylGeom;			// cylindrical geometry flag
} MESHHEAD;


// constant structure definition
typedef struct tagXLATCONST
{ char szName[LEN_IDENT+1];		// constant name
  UINT uType;				// constant type
  union
  { REAL r;				// value of real constant
    int	 n;				// value of integer constant
  } data;
} XLATCONST;


// variable structure definition
typedef struct tagXLATVAR
{ char szName[LEN_IDENT+1];		// variable name
  UINT uOffset;				// variable offset
} XLATVAR;


// array structure definition
typedef struct tagXLATARRAY
{ char szName[LEN_IDENT+1];		// array name
  UINT uOffset;				// array offset
  UINT cDim;				// number of dimensions
  UINT acElements[3];			// elements per dimensions
} XLATARRAY;


// unknown index definition
typedef struct tagTAG			// tag
{ UINT iVar;				// element array index (i.e. offset)
  UINT iEqu;				// equation index
  UINT iArray;				// array index ... -1u == variable
} TAG;


// user-defined structure definition
typedef struct tagXLATUDFUNC
{ char szName[LEN_IDENT+1];		// function name
  UINT iPtr;				// function location in buffer
  UINT cArg;				// number of arguments
  UINT cLocal;				// number of local assignments
} XLATUDFUNC;


// topology index definition
typedef struct tagTOP			// top
{ UINT i;				// index in topology array
  UINT c;				// number of nonzeroes in equation
} TOP;


// equation index definition
typedef struct tagXLATEQUINDEX
{ UINT iPtr;				// equation location in buffer
  UINT cContrib;			// number of contributions to Jacobian
  long lEquation;			// equation file index
  long lDerivative;			// derivative list file index
} XLATEQUINDEX;


// main procedure statement index definition
typedef struct tagXLATMPINDEX
{ UINT iPtr;				// statement location in buffer
  UINT uType;				// type of statement
  long lStatement;			// statement file index
} XLATMPINDEX;


// label structure definition
typedef struct tagXLATLABEL
{ char szName[LEN_IDENT+1];		// label name
  UINT iIndex;				// label index
  UINT uType;				// label type
  UINT cNode;				// number of nodes
} XLATLABEL;


typedef UINT INODE;			// inode
typedef UINT IEDGE;			// iedge
typedef UINT ITRI;			// itri


// node structure definition
typedef struct tagNODE			// node
{ REAL	x;				// x coordinate
  REAL	y;				// y coordinate
} NODE;


// edge structure definition
typedef struct tagEDGE			// edge
{ UINT	uLabel;				// edge label
  INODE inodeA;				// index of node A
  INODE inodeB;				// index of node B
} EDGE;


// triangle structure definition
typedef struct tagTRI			// tri
{ UINT	uLabel;				// region label
  IEDGE iedgeAB;			// index of edge AB
  IEDGE iedgeBC;			// index of edge BC
  IEDGE iedgeAC;			// index of edge AC or CD
  IEDGE iedgeAD;			// index of edge AD
  ITRI	itriAB;				// index of tri. adj. to edge AB
  ITRI	itriBC;				// index of tri. adj. to edge BC
  ITRI	itriAC;				// index of tri. adj. to edge AC or CD
  ITRI	itriAD;				// index of tri. adj. to edge AD
} TRI;


// edge information structure
typedef struct tagEDGEINFO		// edgeinfo
{ REAL	ilen;				// integration length
  REAL	elen;				// edge length
  REAL	Area1;				// partial area 1
  REAL	Area2;				// partial area 2
  IEDGE iedge;				// edge index
  INODE inode;				// node index
  ITRI	ielem;				// element index
} EDGEINFO;


// node information structure
typedef struct tagNODEINFO		// nodeinfo
{ REAL Area;				// area of integration box
  UINT cNeighbor;			// number of neighbors
  UINT cTriangle;			// number of triangles
  EDGEINFO *aedgeinfo;			// edge information array
} NODEINFO;


#endif
