/*-----------------------------------------------------------------------------

 FILE:	   ENHANCE.C

 PROJECT:  TRIANGULAR GRID GENERATOR

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 11/03/93    KMK      initial implementation

 DESCRIPTION:

 This module contains functions which enhance the quality of the mesh.

-----------------------------------------------------------------------------*/


#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if 0
#include <conio.h>
#endif

#include "typedefs.h"
#include "literals.h"
#include "error.h"
#include "mesh.h"
#include "enhance.h"
#include "geometry.h"
#include "graphics.h"


#if defined(ANSIPROTO)
int   nAdjacentEdge	  (PMESH,ITRI,IEDGE,int);
BOOL  fSwapEdges	  (PMESH);
int   nGetTriangleNodes	  (PEDGE,PTRI,INODE *);
int   nGetRectangleNodes  (PEDGE,PTRI,INODE *);
BOOL  fCCW		  (PNODE,INODE,INODE,INODE);
int   nFindNeighbors	  (PMESH,PTRI,INODE *,int);
int   nNudgeNode	  (PMESH);
#else
int   nAdjacentEdge	  ();
BOOL  fSwapEdges	  ();
int   nGetTriangleNodes	  ();
int   nGetRectangleNodes  ();
BOOL  fCCW		  ();
int   nFindNeighbors	  ();
int   nNudgeNode	  ();
#endif


#define VERTEX_A  0
#define VERTEX_B  1
#define VERTEX_C  2
#define VERTEX_D  3
#define EDGE_AB	  0
#define EDGE_BC	  1
#define EDGE_AC	  2
#define SCALE	  1.732050807568877/2.0
#define TOLERANCE 1.0e-3


#define NEIGHBORS 16
typedef struct tagNADJ			/* nadj */
{ WORD	cnode;				/* number of adjacent nodes */
  WORD	inode;				/* index of node under analysis */
  BOOL	fBndry;				/* set if node is a boundary node */
  INODE ainode[NEIGHBORS];		/* indices of neighboring nodes */
} NADJ;

typedef struct EADJ			/* eadj */
{ ITRI	itri1;				/* index of triangle 1 */
  ITRI	itri2;				/* index of triangle 2 */
  ITRI	itriBC1;			/* index of adjacent triangle */
  ITRI	itriBC2;			/* index of adjacent triangle */
  ITRI	itriAC1;			/* index of adjacent triangle */
  ITRI	itriAC2;			/* index of adjacent triangle */
  IEDGE iedgeAB;			/* index of adjacent edge */
  IEDGE iedgeBC1;			/* edge index */
  IEDGE iedgeBC2;			/* edge index */
  IEDGE iedgeAC1;			/* edge index */
  IEDGE iedgeAC2;			/* edge index */
  INODE inodeA;				/* index of adjacent edge node A */
  INODE inodeB;				/* index of adjacent edge node B */
  INODE inodeC1;			/* index of node C of triangle 1 */
  INODE inodeC2;			/* index of node C of triangle 2 */
  BOOL	fBndryAB;			/* edge connects bndy nodes? */
  BOOL	fBndryC;			/* swapped edge connects bndy nodes? */
} EADJ;

static NADJ nadj;
static EADJ eadj;


/******************************************************************************/
/**************************  PUBLIC FUNCTIONS  ********************************/
/******************************************************************************/


/* ----------------------------------------------------------------------------
 * nEnhMeshEdge:  This function enhances the quality of the mesh by swapping
 * adjacent edges between triangles.
 */
#if defined(ANSIPROTO)
int
nEnhMeshEdge(
PMESH pmesh)				/* mesh structure */
#else
int
nEnhMeshEdge(pmesh)
PMESH pmesh;
#endif
{
//PNODE	 anode = pmesh->anode;
  PEDGE	 aedge = pmesh->aedge;
  PTRI	 atri  = pmesh->atri;
  WORD	 cedge = pmesh->cedge;
  WORD	 ctri  = pmesh->ctri;

  PBOOL	 afVisited;			/* flag set if node has been visited */
  BOOL	 fSwapped;			/* set if edges are swapped */
  BOOL	 fDone;				/* inner loop done flag */
  PTRI	 ptri;				/* triangle pointer */
//INODE	 ainode[4];			/* array of node indices */
  IEDGE	 aiedge[3];			/* indices of triangle's nodes */
  IEDGE *piedge;			/* pointer to edge index */
  int	 i,j;				/* loop indices */

#if defined(DEBUG_FUNC)
DEBUG_INFO("enhance.nEnhMeshEdge\n");
#endif

  /* allocate and initialize the visit flag array */
  afVisited = (PBOOL) calloc(cedge, sizeof(BOOL));
  MEMCHECK(afVisited, LIT_ENHANCE, LIT_ENHANCE1);

  /* mark the edges of the rectangular elements as visited so that */
  /* they are not swapped.					   */
  for(i = 0, ptri = atri; i < ctri; ++i, ++ptri)
    if (ptri->iedgeAD != -1u)
    { afVisited[ptri->iedgeAB] = TRUE;
      afVisited[ptri->iedgeBC] = TRUE;
      afVisited[ptri->iedgeCD] = TRUE;
      afVisited[ptri->iedgeAD] = TRUE;
    }

  /* loop through each triangle */
  for(i = 0, ptri = atri; i < ctri; ++i, ++ptri)
  { if (ptri->iedgeAD != -1u) continue;
    do
    { aiedge[0] = ptri->iedgeAB;
      aiedge[1] = ptri->iedgeBC;
      aiedge[2] = ptri->iedgeAC;
      fDone = TRUE;
      for(j = 0, piedge = aiedge; j < 3; ++j, ++piedge)
      { if (!afVisited[*piedge])
	{ afVisited[*piedge] = TRUE;
	  if (aedge[*piedge].wLabel) continue;
	  nAdjacentEdge(pmesh,i,*piedge,j);
#if 0
	  if (cmdargs.wVisual == VISUAL_CONSTRUCT)
	  { ainode[0] = eadj.inodeA;
	    ainode[1] = eadj.inodeC1;
	    ainode[2] = eadj.inodeB;
	    ainode[3] = eadj.inodeC2;
	    nFillPoly(anode, ainode, 4);
	    if (cmdargs.cNoPress != -1)
	    { if (cmdargs.cNoPress) --cmdargs.cNoPress;
	      else { if (getch() == 'q') exit(20); }
	    }
	  }
#endif
	  fSwapped = fSwapEdges(pmesh);
#if 0
	  if (cmdargs.wVisual == VISUAL_CONSTRUCT)
	  { if (!fSwapped)
	    { nDrawTriangle(anode,eadj.inodeA,eadj.inodeB,eadj.inodeC1);
	      nDrawTriangle(anode,eadj.inodeA,eadj.inodeB,eadj.inodeC2);
	    }
	    else
	    { nDrawTriangle(anode,eadj.inodeA,eadj.inodeC1,eadj.inodeC2);
	      nDrawTriangle(anode,eadj.inodeB,eadj.inodeC1,eadj.inodeC2);
	    }
	    if (cmdargs.cNoPress != -1)
	    { if (cmdargs.cNoPress) --cmdargs.cNoPress;
	      else { if (getch() == 'q') exit(20); }
	    }
	  }
#endif
	  if (fSwapped) { fDone = FALSE; break; }
	}
      }
    } while (!fDone);
  }

  /* deallocate flag array and return */
  free(afVisited);
  return(0);

} /* nEnhMeshEdge */


/* ----------------------------------------------------------------------------
 * nEnhMeshNode:  This function enhances the quality of the mesh by reposition-
 * ing inner mesh points.
 */
#if defined(ANSIPROTO)
int
nEnhMeshNode(
PMESH pmesh)				/* mesh structure */
#else
int
nEnhMeshNode(pmesh)
PMESH pmesh;
#endif
{
//PNODE	 anode = pmesh->anode;
  PEDGE	 aedge = pmesh->aedge;
  PTRI	 atri  = pmesh->atri;
  WORD	 cnode = pmesh->cnode;
  WORD	 ctri  = pmesh->ctri;

  PBOOL	 afVisited;			/* flag set if node has been visited */
  PTRI	 ptri;				/* triangle pointer */
  INODE	 ainode[3];			/* indices of triangle's nodes */
  INODE *pinode;			/* pointer to node index */
  IEDGE	 iedge;				/* edge index */
  int	 i,j;				/* loop indices */

#if defined(DEBUG_FUNC)
DEBUG_INFO("enhance.nEnhMeshNode\n");
#endif

  /* allocate and initialize the visit flag array */
  afVisited = (PBOOL) calloc(cnode, sizeof(BOOL));
  MEMCHECK(afVisited, LIT_ENHANCE, LIT_ENHANCE2);

  /* tag the nodes of rectangular elements as visited so that */
  /* they are not enhanced */
  for(i = 0, ptri = atri; i < ctri; ++i, ++ptri)
    if (ptri->iedgeAD != -1u)
    { iedge = ptri->iedgeAB;
      afVisited[aedge[iedge].inodeA] = TRUE;
      afVisited[aedge[iedge].inodeB] = TRUE;
      iedge = ptri->iedgeBC;
      afVisited[aedge[iedge].inodeA] = TRUE;
      afVisited[aedge[iedge].inodeB] = TRUE;
      iedge = ptri->iedgeCD;
      afVisited[aedge[iedge].inodeA] = TRUE;
      afVisited[aedge[iedge].inodeB] = TRUE;
      iedge = ptri->iedgeAD;
      afVisited[aedge[iedge].inodeA] = TRUE;
      afVisited[aedge[iedge].inodeB] = TRUE;
    }

  /* loop through each triangle */
  for(i = 0, ptri = atri; i < ctri; ++i, ++ptri)
  {
    if (ptri->iedgeAD != -1u) continue;
    nGetTriangleNodes(aedge,ptri,ainode);
    for(j = 0, pinode = ainode; j < 3; ++j, ++pinode)
      if (!afVisited[*pinode])
      { afVisited[*pinode] = TRUE;
	nFindNeighbors(pmesh,ptri,ainode,j);
	if (!nadj.fBndry)
	{
#if 0
	  if (cmdargs.wVisual == VISUAL_CONSTRUCT)
	  { nFillPoly(anode,nadj.ainode,nadj.cnode);
	    if (cmdargs.cNoPress != -1)
	    { if (cmdargs.cNoPress) --cmdargs.cNoPress;
	      else { if (getch() == 'q') exit(20); }
	    }
	  }
#endif
	  nNudgeNode(pmesh);
#if 0
	  if (cmdargs.wVisual == VISUAL_CONSTRUCT)
	  { nTriangles(anode,nadj.ainode,nadj.cnode,nadj.inode);
	    if (cmdargs.cNoPress != -1)
	    { if (cmdargs.cNoPress) --cmdargs.cNoPress;
	      else { if (getch() == 'q') exit(20); }
	    }
	  }
#endif
	}
      }
  }


  /* deallocate flag array and return */
  free(afVisited);
  return(0);

} /* nEnhMeshNode */


/* ----------------------------------------------------------------------------
 *  nRateMesh:	This function computes the quality of each triangle in the
 *  mesh in order to determine the minimum and average quality factor.
 */
#if defined(ANSIPROTO)
int
nRateMesh(
PMESH pmesh)
#else
int
nRateMesh(pmesh)
PMESH pmesh;
#endif
{
  PNODE	 anode = pmesh->anode;
  PEDGE	 aedge = pmesh->aedge;
  PTRI	 atri  = pmesh->atri;
  WORD	 ctri  = pmesh->ctri;
  WORD	 cedge = pmesh->cedge;
  WORD	 cnode = pmesh->cnode;
  WORD	 cte   = 0;
  WORD	 cre   = 0;
  WORD	 cObtuse = 0;

  PTRI	 ptri;				/* triangle pointer */
  INODE	 ainode[4];			/* indices of element's nodes */
  INODE	 inodeA,inodeB,inodeC;		/* node indices */
  REAL	 angle1,angle2,angle3;		/* angles */
  int	 nQual;				/* triangle quality */
  int	 i;				/* loop indices */



#if defined(DEBUG_FUNC)
DEBUG_INFO("enhance.nRateMesh\n");
#endif

  for(i = 0, ptri = atri; i < ctri; ++i, ++ptri)
  {
    if (ptri->iedgeAD != -1u)
    {
      nGetRectangleNodes(aedge,ptri,ainode);
      nFillRectangle(anode,ainode[0],ainode[1],ainode[2],ainode[3],QUALITY_GOOD);
      ++cre;
    }
    else
    {
      nGetTriangleNodes(aedge,ptri,ainode);
      inodeA = ainode[0];
      inodeB = ainode[1];
      inodeC = ainode[2];
      nQual = QUALITY_GOOD;
      if (!cmdargs.fNoQualReport)
      { angle1 = CompAngleI(anode,inodeA,inodeB,inodeC);
	angle2 = CompAngleI(anode,inodeB,inodeC,inodeA);
	angle3 = CompAngleI(anode,inodeC,inodeA,inodeB);
	if ((angle1 > PI/2) || (angle2 > PI/2) || (angle3 > PI/2))
	{ ++cObtuse;
	  nQual = QUALITY_FAIR;
	}
      }
      nFillTriangle(anode,inodeA,inodeB,inodeC,nQual);
      ++cte;
    }
  }
  if (!cmdargs.fNoQualReport)
    nQualReport(cnode,cedge,cte,cre,cObtuse);
  return(0);

} /* nRateMesh */


/******************************************************************************/
/**************************  PRIVATE FUNCTIONS	*******************************/
/******************************************************************************/


/* ----------------------------------------------------------------------------
 * nAdjacentEdge:  This function initializes the static adjacency structure
 * with the triangle that shares the specifed edge.  TRUE is returned if a
 * triangle shares the specified edge; otherwise, FALSE is returned.
 */
#if defined(ANSIPROTO)
int
nAdjacentEdge(
PMESH  pmesh,				/* mesh structure */
ITRI   itri1,				/* index of adjacent triangle */
IEDGE  iedgeAB,				/* index of adjacent edge */
int    iEdge)				/* edge AB = 0, BC = 1 & AC = 2 */
#else
int
nAdjacentEdge(pmesh,itri1,iedgeAB,iEdge)
PMESH  pmesh;
ITRI   itri1;
IEDGE  iedgeAB;
int    iEdge;
#endif
{
  PEDGE aedge	= pmesh->aedge;
  PTRI	atri	= pmesh->atri;

  PTRI	ptri;				/* pointer to triangle structure */
  PEDGE pedge;				/* pointer to an edge */
  PEDGE pedgeT;				/* pointer to an edge */
  INODE inode1;				/* node index */
  INODE inode2;				/* node index */
  BOOL	fBndryA;			/* boundary node flag */
  BOOL	fBndryB;			/* boundary node flag */
  BOOL	fBndryC1;			/* boundary node flag */
  BOOL	fBndryC2;			/* boundary node flag */

#if defined(DEBUG_FUNC)
DEBUG_INFO("enhance.nAdjacentEdge\n");
#endif

  pedge = aedge + iedgeAB;
  ptri	= atri	+ itri1;
  eadj.itri1   = itri1;
  eadj.iedgeAB = iedgeAB;
  eadj.inodeA  = pedge->inodeA;
  eadj.inodeB  = pedge->inodeB;
  switch (iEdge)
  { case EDGE_AB:
      eadj.itri2 = ptri->itriAB;
      pedgeT = aedge + ptri->iedgeBC;
      inode1 = pedgeT->inodeA;
      inode2 = pedgeT->inodeB;
      if ((eadj.inodeA == inode1) || (eadj.inodeA == inode2))
      { eadj.inodeC1  = (eadj.inodeA == inode1) ? inode2 : inode1;
	eadj.iedgeBC1 = ptri->iedgeAC;
	eadj.iedgeAC1 = ptri->iedgeBC;
	eadj.itriBC1  = ptri->itriAC;
	eadj.itriAC1  = ptri->itriBC;
      }
      else
      { eadj.inodeC1  = (eadj.inodeB == inode1) ? inode2 : inode1;
	eadj.iedgeBC1 = ptri->iedgeBC;
	eadj.iedgeAC1 = ptri->iedgeAC;
	eadj.itriBC1  = ptri->itriBC;
	eadj.itriAC1  = ptri->itriAC;
      }
      break;

    case EDGE_BC:
      eadj.itri2 = ptri->itriBC;
      pedgeT = aedge + ptri->iedgeAC;
      inode1 = pedgeT->inodeA;
      inode2 = pedgeT->inodeB;
      if ((eadj.inodeA == inode1) || (eadj.inodeA == inode2))
      { eadj.inodeC1  = (eadj.inodeA == inode1) ? inode2 : inode1;
	eadj.iedgeBC1 = ptri->iedgeAB;
	eadj.iedgeAC1 = ptri->iedgeAC;
	eadj.itriBC1  = ptri->itriAB;
	eadj.itriAC1  = ptri->itriAC;
      }
      else
      { eadj.inodeC1  = (eadj.inodeB == inode1) ? inode2 : inode1;
	eadj.iedgeBC1 = ptri->iedgeAC;
	eadj.iedgeAC1 = ptri->iedgeAB;
	eadj.itriBC1  = ptri->itriAC;
	eadj.itriAC1  = ptri->itriAB;
      }
      break;

    case EDGE_AC:
      eadj.itri2 = ptri->itriAC;
      pedgeT = aedge + ptri->iedgeAB;
      inode1 = pedgeT->inodeA;
      inode2 = pedgeT->inodeB;
      if ((eadj.inodeA == inode1) || (eadj.inodeA == inode2))
      { eadj.inodeC1  = (eadj.inodeA == inode1) ? inode2 : inode1;
	eadj.iedgeBC1 = ptri->iedgeBC;
	eadj.iedgeAC1 = ptri->iedgeAB;
	eadj.itriBC1  = ptri->itriBC;
	eadj.itriAC1  = ptri->itriAB;
      }
      else
      { eadj.inodeC1  = (eadj.inodeB == inode1) ? inode2 : inode1;
	eadj.iedgeBC1 = ptri->iedgeAB;
	eadj.iedgeAC1 = ptri->iedgeBC;
	eadj.itriBC1  = ptri->itriAB;
	eadj.itriAC1  = ptri->itriBC;
      }
      break;
  }

  ptri = atri + eadj.itri2;
  if (eadj.iedgeAB == ptri->iedgeAB)
  { pedgeT = aedge + ptri->iedgeBC;
    inode1 = pedgeT->inodeA;
    inode2 = pedgeT->inodeB;
    if ((eadj.inodeA == inode1) || (eadj.inodeA == inode2))
    { eadj.inodeC2  = (eadj.inodeA == inode1) ? inode2 : inode1;
      eadj.iedgeBC2 = ptri->iedgeAC;
      eadj.iedgeAC2 = ptri->iedgeBC;
      eadj.itriBC2  = ptri->itriAC;
      eadj.itriAC2  = ptri->itriBC;
    }
    else
    { eadj.inodeC2  = (eadj.inodeB == inode1) ? inode2 : inode1;
      eadj.iedgeBC2 = ptri->iedgeBC;
      eadj.iedgeAC2 = ptri->iedgeAC;
      eadj.itriBC2  = ptri->itriBC;
      eadj.itriAC2  = ptri->itriAC;
    }
  }
  else if (eadj.iedgeAB == ptri->iedgeBC)
  { pedgeT = aedge + ptri->iedgeAC;
    inode1 = pedgeT->inodeA;
    inode2 = pedgeT->inodeB;
    if ((eadj.inodeA == inode1) || (eadj.inodeA == inode2))
    { eadj.inodeC2  = (eadj.inodeA == inode1) ? inode2 : inode1;
      eadj.iedgeBC2 = ptri->iedgeAB;
      eadj.iedgeAC2 = ptri->iedgeAC;
      eadj.itriBC2  = ptri->itriAB;
      eadj.itriAC2  = ptri->itriAC;
    }
    else
    { eadj.inodeC2  = (eadj.inodeB == inode1) ? inode2 : inode1;
      eadj.iedgeBC2 = ptri->iedgeAC;
      eadj.iedgeAC2 = ptri->iedgeAB;
      eadj.itriBC2  = ptri->itriAC;
      eadj.itriAC2  = ptri->itriAB;
    }
  }
  else if (eadj.iedgeAB == ptri->iedgeAC)
  { pedgeT = aedge + ptri->iedgeAB;
    inode1 = pedgeT->inodeA;
    inode2 = pedgeT->inodeB;
    if ((eadj.inodeA == inode1) || (eadj.inodeA == inode2))
    { eadj.inodeC2  = (eadj.inodeA == inode1) ? inode2 : inode1;
      eadj.iedgeBC2 = ptri->iedgeBC;
      eadj.iedgeAC2 = ptri->iedgeAB;
      eadj.itriBC2  = ptri->itriBC;
      eadj.itriAC2  = ptri->itriAB;
    }
    else
    { eadj.inodeC2  = (eadj.inodeB == inode1) ? inode2 : inode1;
      eadj.iedgeBC2 = ptri->iedgeAB;
      eadj.iedgeAC2 = ptri->iedgeBC;
      eadj.itriBC2  = ptri->itriAB;
      eadj.itriAC2  = ptri->itriBC;
    }
  }

  fBndryA = fBndryB = fBndryC1 = fBndryC2 = FALSE;
  if (aedge[eadj.iedgeBC1].wLabel) { fBndryB = TRUE; fBndryC1 = TRUE; }
  if (aedge[eadj.iedgeAC1].wLabel) { fBndryA = TRUE; fBndryC1 = TRUE; }
  if (aedge[eadj.iedgeBC2].wLabel) { fBndryB = TRUE; fBndryC2 = TRUE; }
  if (aedge[eadj.iedgeAC2].wLabel) { fBndryA = TRUE; fBndryC2 = TRUE; }
  eadj.fBndryAB = fBndryA  && fBndryB;
  eadj.fBndryC	= fBndryC1 && fBndryC2;

  return(TRUE);

} /* nAdjacentEdge */


/* ----------------------------------------------------------------------------
 * fSwapEdges:	This function swaps edges in attempts to enhance the quality
 * of the mesh.
 */
#if defined(ANSIPROTO)
BOOL
fSwapEdges(
PMESH pmesh)				/* mesh structure */
#else
BOOL
fSwapEdges(pmesh)
PMESH pmesh;
#endif
{
  PNODE anode = pmesh->anode;
  PEDGE aedge = pmesh->aedge;
  PTRI	atri  = pmesh->atri;

  PEDGE pedge;				/* edge pointer */
  INODE inodeA,inodeB;			/* node indices */
  PTRI	ptri1;				/* triangle pointer */
  PTRI	ptri2;				/* triangle pointer */
  PTRI	ptriT;				/* triangle pointer */
  REAL	QualA,Qual1A,Qual2A;		/* triangle quality factors */
  REAL	QualB,Qual1B,Qual2B;		/* triangle quality factors */

#if defined(DEBUG_FUNC)
DEBUG_INFO("enhance.fSwapEdges\n");
#endif

  /* compute triangle quality factors ... nodes must be specified in a */
  /* counter-clockwise order for this to work! */
  inodeA = eadj.inodeA;
  inodeB = eadj.inodeB;
  Qual1A = CompQualityI(anode,inodeA,eadj.inodeC1,inodeB);
  if (Qual1A < 0.0)
  { Qual1A = -Qual1A;
    inodeA = inodeB;
    inodeB = eadj.inodeA;
  }
  Qual2A = CompQualityI(anode,inodeA,inodeB,eadj.inodeC2);
  Qual1B = CompQualityI(anode,inodeB,eadj.inodeC2,eadj.inodeC1);
  Qual2B = CompQualityI(anode,inodeA,eadj.inodeC1,eadj.inodeC2);
  QualA	 = min(Qual1A,Qual2A);
  QualB	 = min(Qual1B,Qual2B);

  /* swapping edges does not enhance mesh quality */
  if (( eadj.fBndryAB &&  eadj.fBndryC && (QualB <= QualA))	  ||
      (!eadj.fBndryAB &&  eadj.fBndryC && (QualB <= QualA / 0.9)) ||
      ( eadj.fBndryAB && !eadj.fBndryC && (QualB <= QualA * 0.9)) ||
      (!eadj.fBndryAB && !eadj.fBndryC && (QualB <= QualA)))
    return(FALSE);

  /* swap edges */
  ptri1 = atri	 + eadj.itri1;
  ptri2 = atri	 + eadj.itri2;
  pedge = aedge	 + eadj.iedgeAB;
  pedge->inodeA	 = eadj.inodeC1;
  pedge->inodeB	 = eadj.inodeC2;
  ptri1->iedgeAB = eadj.iedgeAB;
  ptri1->iedgeBC = eadj.iedgeAC1;
  ptri1->iedgeAC = eadj.iedgeAC2;
  ptri2->iedgeAB = eadj.iedgeAB;
  ptri2->iedgeBC = eadj.iedgeBC1;
  ptri2->iedgeAC = eadj.iedgeBC2;
  ptri1->itriAB	 = eadj.itri2;
  ptri1->itriBC	 = eadj.itriAC1;
  ptri1->itriAC	 = eadj.itriAC2;
  ptri2->itriAB	 = eadj.itri1;
  ptri2->itriBC	 = eadj.itriBC1;
  ptri2->itriAC	 = eadj.itriBC2;
  ptriT = atri	 + eadj.itriAC2;
  if	  (ptriT->itriAB == eadj.itri2) ptriT->itriAB = eadj.itri1;
  else if (ptriT->itriBC == eadj.itri2) ptriT->itriBC = eadj.itri1;
  else if (ptriT->itriAC == eadj.itri2) ptriT->itriAC = eadj.itri1;
  ptriT = atri	 + eadj.itriBC1;
  if	  (ptriT->itriAB == eadj.itri1) ptriT->itriAB = eadj.itri2;
  else if (ptriT->itriBC == eadj.itri1) ptriT->itriBC = eadj.itri2;
  else if (ptriT->itriAC == eadj.itri1) ptriT->itriAC = eadj.itri2;
  return(TRUE);

} /* fSwapEdges */




/* ----------------------------------------------------------------------------
 * nGetTriangleNodes:  This function returns the node indices of the triangles
 * three vertices.
 */
#if defined(ANSIPROTO)
int
nGetTriangleNodes(
PEDGE  aedge,				/* edge array */
PTRI   ptri,				/* triangle structure */
INODE *ainode)				/* array of node indices */
#else
nGetTriangleNodes(aedge,ptri,ainode)
PEDGE  aedge;
PTRI   ptri;
INODE *ainode;
#endif
{
  PEDGE pedgeAB = aedge + ptri->iedgeAB;
  PEDGE pedgeAC = aedge + ptri->iedgeAC;

#if defined(DEBUG_FUNC)
DEBUG_INFO("nGetTriangleNodes\n");
#endif

  if	  (pedgeAB->inodeA == pedgeAC->inodeA)
  { ainode[VERTEX_A] = pedgeAB->inodeA;
    ainode[VERTEX_B] = pedgeAB->inodeB;
    ainode[VERTEX_C] = pedgeAC->inodeB;
  }
  else if (pedgeAB->inodeA == pedgeAC->inodeB)
  { ainode[VERTEX_A] = pedgeAB->inodeA;
    ainode[VERTEX_B] = pedgeAB->inodeB;
    ainode[VERTEX_C] = pedgeAC->inodeA;
  }
  else if (pedgeAB->inodeB == pedgeAC->inodeA)
  { ainode[VERTEX_A] = pedgeAB->inodeB;
    ainode[VERTEX_B] = pedgeAB->inodeA;
    ainode[VERTEX_C] = pedgeAC->inodeB;
  }
  else if (pedgeAB->inodeB == pedgeAC->inodeB)
  { ainode[VERTEX_A] = pedgeAB->inodeB;
    ainode[VERTEX_B] = pedgeAB->inodeA;
    ainode[VERTEX_C] = pedgeAC->inodeA;
  }
  else vErrorHandler(ERROR_PROGRAMMER, NULL, NULL, TRUE);

  return(0);

} /* nGetTriangleNodes */


/* ----------------------------------------------------------------------------
 * nGetRectangleNodes:	This function returns the node indices of the
 * rectangles four vertices.
 */
#if defined(ANSIPROTO)
int
nGetRectangleNodes(
PEDGE  aedge,				/* edge array */
PTRI   ptri,				/* triangle structure */
INODE *ainode)				/* array of node indices */
#else
nGetRectangleNodes(aedge,ptri,ainode)
PEDGE  aedge;
PTRI   ptri;
INODE *ainode;
#endif
{
  PEDGE pedgeAB = aedge + ptri->iedgeAB;
  PEDGE pedgeBC = aedge + ptri->iedgeBC;
  PEDGE pedgeCD = aedge + ptri->iedgeCD;

#if defined(DEBUG_FUNC)
DEBUG_INFO("nGetRectangleNodes\n");
#endif

  if	  ((pedgeAB->inodeA == pedgeBC->inodeA) ||
	   (pedgeAB->inodeA == pedgeBC->inodeB))
  { ainode[VERTEX_A] = pedgeAB->inodeB;
    ainode[VERTEX_B] = pedgeAB->inodeA;
  }
  else if ((pedgeAB->inodeB == pedgeBC->inodeA) ||
	   (pedgeAB->inodeB == pedgeBC->inodeB))
  { ainode[VERTEX_A] = pedgeAB->inodeA;
    ainode[VERTEX_B] = pedgeAB->inodeB;
  }
  else vErrorHandler(ERROR_PROGRAMMER, NULL, NULL, TRUE);

  if	  ((pedgeCD->inodeA == pedgeBC->inodeA) ||
	   (pedgeCD->inodeA == pedgeBC->inodeB))
  { ainode[VERTEX_C] = pedgeCD->inodeA;
    ainode[VERTEX_D] = pedgeCD->inodeB;
  }
  else if ((pedgeCD->inodeB == pedgeBC->inodeA) ||
	   (pedgeCD->inodeB == pedgeBC->inodeB))
  { ainode[VERTEX_C] = pedgeCD->inodeB;
    ainode[VERTEX_D] = pedgeCD->inodeA;
  }
  else vErrorHandler(ERROR_PROGRAMMER, NULL, NULL, TRUE);


  return(0);

} /* nGetRectangleNodes */


/* ----------------------------------------------------------------------------
 * fCCW:  This functions determines if rotating segment 12 to 13 is in the
 * counter-clockwise direction.
 */
#if defined(ANSIPROTO)
BOOL
fCCW(
PNODE anode,				/* node array */
INODE inode1,				/* index of node 1 */
INODE inode2,				/* index of node 2 */
INODE inode3)				/* index of node 3 */
#else
BOOL
fCCW(anode,inode1,inode2,inode3)
PNODE anode;
INODE inode1;
INODE inode2;
INODE inode3;
#endif
{
  PNODE pnode1 = anode + inode1;
  PNODE pnode2 = anode + inode2;
  PNODE pnode3 = anode + inode3;
  REAL	x1     = pnode1->x;
  REAL	y1     = pnode1->y;
  REAL	x2     = pnode2->x;
  REAL	y2     = pnode2->y;
  REAL	x3     = pnode3->x;
  REAL	y3     = pnode3->y;
  REAL	x12    = x2 - x1;
  REAL	y12    = y2 - y1;
  REAL	x13    = x3 - x1;
  REAL	y13    = y3 - y1;
  REAL	d12    = sqrt(x12*x12 + y12*y12);
  REAL	d13    = sqrt(x13*x13 + y13*y13);
  REAL	t12    = x12/d12;
  REAL	t13    = x13/d13;
  REAL	a12;
  REAL	a13;

#if defined(DEBUG_FUNC)
DEBUG_INFO("enhance.fCCW\n");
#endif

  if	  (t12 >  1.0) t12 =  1.0;
  else if (t12 < -1.0) t12 = -1.0;
  if	  (t13 >  1.0) t13 =  1.0;
  else if (t13 < -1.0) t13 = -1.0;
  a12 = acos(t12);
  a13 = acos(t13);
  if (y12 < 0) a12 = 2*PI - a12;
  if (y13 < 0) a13 = 2*PI - a13;
  return(a13 > a12);

} /* fCCW */


/* ----------------------------------------------------------------------------
 * nFindNeighbors:  This function initializes the static adjacency structure
 * with the neighbors of the specified node.  The number of neighbors is
 * returned.
 */
#if defined(ANSIPROTO)
int
nFindNeighbors(
PMESH  pmesh,				/* mesh structure */
PTRI   ptri,				/* triangle structure */
INODE *ainode,				/* array of node indices */
int    iVertex)				/* vertex A = 0, B = 1 & C = 2 */
#else
int
nFindNeighbors(pmesh,ptri,ainode,iVertex)
PMESH  pmesh;
PTRI   ptri;
INODE *ainode;
int    iVertex;
#endif
{
  PNODE anode = pmesh->anode;
  PEDGE aedge = pmesh->aedge;
  PTRI	atri  = pmesh->atri;

  INODE ainodeT[3];
  ITRI	itri;
  WORD	cnode;
  WORD	i,i1,i2;
  WORD	wBndry;

#if defined(DEBUG_FUNC)
DEBUG_INFO("enhance.nFindNeighbors\n");
#endif

  nadj.inode   = ainode[iVertex];
  nadj.fBndry = TRUE;
  nadj.cnode  = 0;
  cnode = wBndry = 0;

  switch (iVertex)
  { case VERTEX_A:
      if (fCCW(anode,ainode[VERTEX_A],ainode[VERTEX_B],ainode[VERTEX_C]))
      { nadj.ainode[0] = ainode[VERTEX_B];
	wBndry = aedge[ptri->iedgeAB].wLabel;
      }
      else
      { nadj.ainode[0] = ainode[VERTEX_C];
	wBndry = aedge[ptri->iedgeAC].wLabel;
      }
      break;
    case VERTEX_B:
      if (fCCW(anode,ainode[VERTEX_B],ainode[VERTEX_C],ainode[VERTEX_A]))
      { nadj.ainode[0] = ainode[VERTEX_C];
	wBndry = aedge[ptri->iedgeBC].wLabel;
      }
      else
      { nadj.ainode[0] = ainode[VERTEX_A];
	wBndry = aedge[ptri->iedgeAB].wLabel;
      }
      break;
    case VERTEX_C:
      if (fCCW(anode,ainode[VERTEX_C],ainode[VERTEX_A],ainode[VERTEX_B]))
      { nadj.ainode[0] = ainode[VERTEX_A];
	wBndry = aedge[ptri->iedgeAC].wLabel;
      }
      else
      { nadj.ainode[0] = ainode[VERTEX_B];
	wBndry = aedge[ptri->iedgeBC].wLabel;
      }
      break;
  }

  memcpy(ainodeT,ainode,sizeof(ainodeT));
  for(;;)
  { i = (iVertex + 1) % 3;
    if (ainodeT[i] == nadj.ainode[cnode]) i = (i + 1) % 3;
    if ((nadj.ainode[++cnode] = ainodeT[i]) == nadj.ainode[0]) break;

    if (i > iVertex) { i1 = iVertex; i2 = i; }
    else	     { i1 = i; i2 = iVertex; }

    if	    ((i1 == VERTEX_A) && (i2 == VERTEX_B))
    { wBndry |= aedge[ptri->iedgeAB].wLabel;
      if ((itri = ptri->itriAB) == (ITRI) -1) return(0);
      ptri = atri + itri;
    }
    else if ((i1 == VERTEX_B) && (i2 == VERTEX_C))
    { wBndry |= aedge[ptri->iedgeBC].wLabel;
      if ((itri = ptri->itriBC) == (ITRI) -1) return(0);
      ptri = atri + itri;
    }
    else if ((i1 == VERTEX_A) && (i2 == VERTEX_C))
    { wBndry |= aedge[ptri->iedgeAC].wLabel;
      if ((itri = ptri->itriAC) == (ITRI) -1) return(0);
      ptri = atri + itri;
    }

    nGetTriangleNodes(aedge,ptri,ainodeT);
    if	    (ainodeT[VERTEX_A] == nadj.inode) iVertex = VERTEX_A;
    else if (ainodeT[VERTEX_B] == nadj.inode) iVertex = VERTEX_B;
    else if (ainodeT[VERTEX_C] == nadj.inode) iVertex = VERTEX_C;
  }

  nadj.fBndry = (BOOL) wBndry;
  nadj.cnode  = cnode;
  return(cnode);

} /* nFindNeighbors */



/* ----------------------------------------------------------------------------
 * nNudgeNode:	This function adjusts the position of an inner node in order
 * to enhance the overall quality of the grid.	The adjacency structure is
 * assumed to be properly initialized.
 */
#if defined(ANSIPROTO)
int
nNudgeNode(
PMESH pmesh)
#else
int
nNudgeNode(pmesh)
PMESH pmesh;
#endif
{
  PNODE	 anode	= pmesh->anode;
  INODE *ainode = nadj.ainode;
  INODE	 inode	= nadj.inode;
  PNODE	 pnode	= anode + inode;
  WORD	 cnode	= nadj.cnode;

  INODE *pinode;
  INODE	 inode1M,inode1R;
  INODE	 inode2M,inode2R;
  REAL	 Qual,Qual1,Qual2,QualT;	/* triangle quality factors */
  REAL	 xmin,ymin,xmax,ymax;
  REAL	 dx1,dy1,dx2,dy2,dx,dy;
  REAL	 xm1,ym1,xm2,ym2,xm,ym;
  REAL	 r1,r2,r;
  REAL	 a,b,c,beta;
  REAL	 xc1,yc1,xc2,yc2;
  REAL	 zx,zy;
  WORD	 m1,m2;
  WORD	 i;

#if defined(DEBUG_FUNC)
DEBUG_INFO("enhance.nNudgeNode\n");
#endif

  /* find the triangles with the smallest interior angles */
  m1	= m2	= -1u;
  Qual1 = Qual2 =  1.0;
  for(i = 0, pinode = ainode; i < cnode; ++i, ++pinode)
  { Qual = fabs(CompQualityI(anode,inode,pinode[0],pinode[1]));
    if	    (Qual < Qual1) { Qual2 = Qual1; Qual1 = Qual; m2 = m1; m1 = i; }
    else if (Qual < Qual2) { Qual2 = Qual;		  m2 = i; }
  }

  /* determine (xmin,ymin) and (xmax,ymax) ... (xmax,ymax) is the point	    */
  /* where the level curves (circles) for the two worst triangles are equal */
  /* and tangent */
  xmin	  = anode[inode].x;
  ymin	  = anode[inode].y;
  inode1M = ainode[m1];
  inode1R = ainode[m1+1];
  inode2M = ainode[m2];
  inode2R = ainode[m2+1];
  dx1	  = (anode[inode1R].x - anode[inode1M].x) * SCALE;
  dy1	  = (anode[inode1R].y - anode[inode1M].y) * SCALE;
  dx2	  = (anode[inode2R].x - anode[inode2M].x) * SCALE;
  dy2	  = (anode[inode2R].y - anode[inode2M].y) * SCALE;
  dx	  = dx1 - dx2;
  dy	  = dy1 - dy2;
  xm1	  = (anode[inode1R].x + anode[inode1M].x) / 2.0;
  ym1	  = (anode[inode1R].y + anode[inode1M].y) / 2.0;
  xm2	  = (anode[inode2R].x + anode[inode2M].x) / 2.0;
  ym2	  = (anode[inode2R].y + anode[inode2M].y) / 2.0;
  xm	  = xm1 - xm2;
  ym	  = ym1 - ym2;
  r1	  = sqrt(dx1*dx1 + dy1*dy1);
  r2	  = sqrt(dx2*dx2 + dy2*dy2);
  r	  = r1 + r2;
  a	  = r*r - dx*dx - dy*dy;
  b	  = ym*dx - xm*dy;
  c	  = xm*xm + ym*ym + r*r;
  beta	  = (a > 0.0) ? (b + sqrt(b*b + a*c))/a : 1.0;
  xc1	  = xm1 - beta * dy1;
  yc1	  = ym1 + beta * dx1;
  xc2	  = xm2 - beta * dy2;
  yc2	  = ym2 + beta * dx2;
  xmax	  = (xc1*r2 + xc2*r1) / r;
  ymax	  = (yc1*r2 + yc2*r1) / r;

  /* bisection loop */
  for(;;)
  { zx = fabs(xmin - xmax) / (fabs(xmin) + fabs(xmax) + EPS);
    zy = fabs(ymin - ymax) / (fabs(ymin) + fabs(ymax) + EPS);
    if (max(zx,zy) < TOLERANCE) { pnode->x = xmin; pnode->y = ymin; break; }
    pnode->x = (xmin + xmax) / 2.0;
    pnode->y = (ymin + ymax) / 2.0;
    QualT = 1.0;
    for(i = 0, pinode = ainode; i < cnode; ++i, ++pinode)
    { Qual = fabs(CompQualityI(anode,inode,pinode[0],pinode[1]));
      if (Qual < Qual1) { xmax = pnode->x; ymax = pnode->y; break; }
      QualT = min(QualT,Qual);
    }
    if (i == cnode) { xmin = pnode->x; ymin = pnode->y; Qual1 = QualT; }
  }

  return(0);

} /* nNudgeNode */


