/*-----------------------------------------------------------------------------

 FILE:	   ERROR.C

 PROJECT:  TRIANGULAR GRID GENERATOR

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 10/19/93    KMK      initial implementation
 10/30/93    KMK      second rewrite

 DESCRIPTION:

 This module implements an error handling routine.  All error messages should
 come through this routine in order to assure uniformity.

-----------------------------------------------------------------------------*/


#include <stdio.h>
#include <stdlib.h>

#include "typedefs.h"
#include "literals.h"


/* definition of error types ... ERROR_#####1 have file and line number */
/* information while ERROR_#####2 do not have this information */
#define ERROR_SYSTEM1	    1
#define ERROR_SYSTEM2	    2
#define ERROR_PARSER	    3
#define ERROR_WARNING	    4

/* device in which to report errors */
static FILE *fileError;

/* definition of error structure */
typedef struct tagERROR		    /* error */
{ WORD wErrorType;		    /* type of error */
  PSTR psz;			    /* error message */
} ERROR;

/* definition of error strings */
static char error0[]  = LIT_ERROR0;
static char error1[]  = LIT_ERROR1;
static char error2[]  = LIT_ERROR2;
static char error3[]  = LIT_ERROR3;
static char error4[]  = LIT_ERROR4;
static char error5[]  = LIT_ERROR5;
static char error6[]  = LIT_ERROR6;
static char error7[]  = LIT_ERROR7;
static char error8[]  = LIT_ERROR8;


/* error structure table */
static ERROR error[] =
{ { ERROR_SYSTEM1,  error0  },
  { ERROR_SYSTEM1,  error1  },
  { ERROR_SYSTEM2,  error2  },
  { ERROR_SYSTEM2,  error3  },
  { ERROR_SYSTEM2,  error4  },
  { ERROR_PARSER,   error5  },
  { ERROR_SYSTEM2,  error6  },
  { ERROR_SYSTEM2,  error7  },
  { ERROR_SYSTEM2,  error8  }
};


/******************************************************************************/
/**************************  PUBLIC FUNCTIONS  ********************************/
/******************************************************************************/


/* ----------------------------------------------------------------------------
 * vErrorHandler:  This function implements a standard method of reporting
 * errors and warnings.
 */
#if defined(ANSIPROTO)
void
vErrorHandler(
WORD wError,			    /* error number */
PSTR pszArg1,			    /* first argument */
PSTR pszArg2,			    /* second argument */
BOOL fAbort)			    /* abort program if TRUE */
#else
void
vErrorHandler(wError, pszArg1, pszArg2, fAbort)
WORD wError;
PSTR pszArg1;
PSTR pszArg2;
BOOL fAbort;
#endif
{
#if defined(DEBUG_FUNC)
DEBUG_INFO("ERROR.vErrorHandler\n");
#endif

  switch (error[wError].wErrorType)
  {
    case ERROR_SYSTEM1:
      fprintf(fileError, LIT_ERRORa, error[wError].psz, pszArg1, pszArg2);
      break;

    case ERROR_SYSTEM2:
      fprintf(fileError, LIT_ERRORb);
      fprintf(fileError, error[wError].psz, pszArg1, pszArg2);
      break;

    case ERROR_PARSER:
      fprintf(fileError, LIT_ERRORc);
      fprintf(fileError, error[wError].psz, pszArg1, pszArg2);
      break;

    case ERROR_WARNING:
      fprintf(fileError, LIT_ERRORd);
      fprintf(fileError, error[wError].psz, pszArg1, pszArg2);
      break;
  }

  /* if the abort flag is set, exit the program */
  if (fAbort) exit(RETCODE_ERROR);
  return;

} /* vErrorHandler */


/* ----------------------------------------------------------------------------
 * vErrorInit:	This file initialize the error module by specifying which
 * device to report errors.  NOTE:  The file name variable is currently
 * ignored by this function and errors may only be sent to the standard error
 * or standard output device.  When error recovery is implemented and multiple
 * errors may be reported, error messages may be sent to a file.
 */
#if defined(ANSIPROTO)
void
vErrorInit(
WORD wErrorDevice,		    /* error device:  stderr stdout or file */
PSTR pszFileName)		    /* name of error file */
#else
void
vErrorInit(wErrorDevice, pszFileName)
WORD wErrorDevice;
PSTR pszFileName;
#endif
#ifdef MSDOS
#pragma argsused
#endif
{
#if defined(DEBUG_FUNC)
DEBUG_INFO("error.vErrorInit\n");
#endif

  fileError = (wErrorDevice == ERRDEV_STDOUT) ? stdout : stderr;
  return;

} /* vErrorInit */


/* ----------------------------------------------------------------------------
 * fileGetErrorDevice:	This function returns the error stream.
 */
FILE *
fileGetErrorDevice()
{
#if defined(DEBUG_FUNC)
DEBUG_INFO("error.fileGetErrorDevice\n");
#endif

  return(fileError);

} /* fileGetErrorDevice */

