/*-----------------------------------------------------------------------------

 FILE:	   GEOMETRY.H

 PROJECT:  TRIANGULAR GRID GENERATOR

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 10/25/93    KMK      initial implementation

 DESCRIPTION:

 This module contains geometric functions that compute the angle between three
 points, computes the local spacing between three points, computes the quality
 factor of a triangle, etc.

-----------------------------------------------------------------------------*/


#ifndef	 HEADER_GEOMETRY
#define	 HEADER_GEOMETRY


/******************************************************************************/
/***********************  PUBLIC TYPE DEFINITION  *****************************/
/******************************************************************************/


typedef struct tagCRDNT			/* crdnt */
{ REAL x;				/* x coordinate */
  REAL y;				/* y coordinate */
} CRDNT;
typedef CRDNT *PCRDNT;			/* pcrdnt */


/******************************************************************************/
/*********************	PUBLIC MACROS AND CONSTANTS  **************************/
/******************************************************************************/


#define	 GEOMETRY_COLINEAR    1
#define	 GEOMETRY_PLANER      0
#define	 GEOMETRY_PARALLEL    1
#define	 GEOMETRY_INTERSECT   0


/* ----------------------------------------------------------------------------
 * CompAngleI:	This macro computes the angle between three points given by
 * node indices 1, 2 and 3.  The angle is measured from vector 12 to 32.
 */
#define CompAngleI(a,i1,i2,i3) CompAngle(a[i1].x,a[i1].y, \
					 a[i2].x,a[i2].y, \
					 a[i3].x,a[i3].y)


/* ----------------------------------------------------------------------------
 * CompHlocalI:	 This macro computes the local spacing between three points
 * given by node indices 1, 2 and 3.
 */
#define CompHlocalI(a,i1,i2,i3) CompHlocal(a[i1].x,a[i1].y, \
					   a[i2].x,a[i2].y, \
					   a[i3].x,a[i3].y)


/* ----------------------------------------------------------------------------
 * CompQualityI:  This macro computes the quality factor of the triangle given
 * by node indices 1, 2 and 3.	The quality factor is normalized so that an
 * equilateral triangle has a quality factor of one and a triangle with a very
 * very small angle has a quality factor approaching zero.
 */
#define CompQualityI(a,i1,i2,i3) CompQuality(a[i1].x,a[i1].y, \
					     a[i2].x,a[i2].y, \
					     a[i3].x,a[i3].y)


/* ----------------------------------------------------------------------------
 * CompQualRectI:  This macro computes the quality factor of the rectangle
 * given by node indices 1, 2, 3 and 4.	 The quality factor is normalized so
 * that a square has a quality factor of one and rectangle with a poor aspect
 * ration approaches zero.
 */
#define CompQualRectI(a,i1,i2,i3,i4) CompRectQual(a[i1].x,a[i1].y, \
						  a[i2].x,a[i2].y, \
						  a[i3].x,a[i3].y)


/******************************************************************************/
/**************************  PUBLIC FUNCTIONS  ********************************/
/******************************************************************************/


#if defined(ANSIPROTO)
REAL  CompAngle	    (REAL,REAL,REAL,REAL,REAL,REAL);
REAL  CompHlocal    (REAL,REAL,REAL,REAL,REAL,REAL);
REAL  CompQuality   (REAL,REAL,REAL,REAL,REAL,REAL);
REAL  CompRectQual  (REAL,REAL,REAL,REAL,REAL,REAL);
int   nCircleCenter (REAL,REAL,REAL,REAL,REAL,REAL,PREAL,PREAL);
int   nParamArc	    (REAL,REAL,REAL,REAL,REAL,REAL,REAL,REAL,PREAL,PREAL,PREAL);
int   nIntersection (PCRDNT,PCRDNT,PCRDNT,PCRDNT,PREAL,PREAL);
int   cNumNewPoints (REAL, REAL, REAL, WORD, PREAL, PREAL);
int   cComputeGrade (REAL, PREAL, PREAL);
#else
REAL  CompAngle	    ();
REAL  CompHlocal    ();
REAL  CompQuality   ();
int   nCircleCenter ();
int   nParamArc	    ();
int   nIntersection ();
int   cNumNewPoints ();
int   cComputeGrade ();
#endif

#endif

