/*-----------------------------------------------------------------------------

 FILE:	   GRAPHICS.C

 PROJECT:  TRIANGULAR GRID GENERATOR

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 10/19/93    KMK      initial implementation

 DESCRIPTION:

 This module contains the graphic functions.

-----------------------------------------------------------------------------*/


#include <stdio.h>
#include <stdlib.h>
#include "typedefs.h"

#if defined(UNIX)
#include "xgraph.h"
#else
#include "wgraph.h"
#endif

#include "graphics.h"

int   nGetTriangleNodes   (PEDGE,PTRI,INODE *);
int   nGetRectangleNodes  (PEDGE,PTRI,INODE *);

static int    cxLeftBorder = 0;
static int    cxRightBorder = 0;
static int    cyTopBorder = 0;
static int    cyBottomBorder = 0;
static int    cxGraph = 0;
static int    cyGraph = 0;
static double Xmin = 0.0;
static double Xmax = 0.0;
static double Ymin = 0.0;
static double Ymax = 0.0;
static double dX   = 0.0;
static double dY   = 0.0;

#define MAPx(x) ((int) (cxGraph *	 ((x - Xmin) / dX)  + cxLeftBorder))
#define MAPy(y) ((int) (cyGraph * (1.0 - ((y - Ymin) / dY)) + cyTopBorder ))

int nCreateWindow(double xmin, double ymin, double xmax, double ymax, PFREDRAW pfredraw)
{
  int MAXx, MAXy;

  Xmin = xmin;
  Ymin = ymin;
  Xmax = xmax;
  Ymax = ymax;
  dX   = xmax - xmin;
  dY   = ymax - ymin;
  if (cmdargs.fKeepAspect) { if (dX > dY) dY = dX; else dX = dY; }

  cxLeftBorder = 10;
  cxRightBorder = 10;
  cyTopBorder = 10;
  cyBottomBorder = 10;

  switch (cmdargs.nResolution)
 { case RES_640x480:
      MAXx = 640;
      MAXy = 480;
      break;

    case RES_800x600:
      MAXx = 800;
      MAXy = 600;
      break;

    case RES_1024x768:
      MAXx = 1024;
      MAXy = 768;
      break;
  }

  cxGraph = MAXy - cxLeftBorder - cxRightBorder - 23;
  cyGraph = MAXy - cyTopBorder - cyBottomBorder - 23;

  GROpenGraphWin("Initial Mesh", "Mesh", -1, -1, MAXx, MAXy, (PFREDRAW) pfredraw);

  return(0);
}


int nPostScript (PMESH pmesh)
{
  PNODE	 anode = pmesh->anode;
  PEDGE	 aedge = pmesh->aedge;
  PTRI	 atri  = pmesh->atri;
  WORD	 ctri  = pmesh->ctri;

  PTRI	 ptri;				/* triangle pointer */
  INODE	 ainode[4];			/* indices of element's nodes */
  int	 i;				/* loop indices */

  cxGraph = 9360;
  cyGraph = 9360;
	cxLeftBorder = 1440;
	cyTopBorder = 3240;

  GRPrintOpen("grid");

  for(i = 0, ptri = atri; i < ctri; ++i, ++ptri)
  {
    if (ptri->iedgeAD != -1u)
    {
      nGetRectangleNodes(aedge,ptri,ainode);
      nFillRectangle(anode,ainode[0],ainode[1],ainode[2],ainode[3],0);
    }
    else
    {
      nGetTriangleNodes(aedge,ptri,ainode);
      nFillTriangle(anode,ainode[0],ainode[1],ainode[2],0);
    }
  }

  GRPrintClose();
  return(0);
}


int nGraphPolygon(PMESH pmesh)
{
  PPOLY ppoly = pmesh->ppoly;
  PVTX	avtx  = pmesh->avtx;
  PNODE anode = pmesh->anode;
  IVTX	ivtx  = ppoly->ivtxHead;
  WORD	cvtx  = ppoly->cvtx;

  INODE inode,inodeR;			/* point indices */
  IVTX	ivtxR;				/* vertex index */
  int	x,y,xr,yr;			/* coordinates */
  int	i;				/* loop index */

  inode = avtx[ivtx].inode;
  x = MAPx(anode[inode].x);
  y = MAPy(anode[inode].y);
  GRSetColor(GR_BLUE);
  for(i = 0; i < cvtx; ++i)
  { ivtxR = avtx[ivtx].ivtxR;
    inodeR = avtx[ivtxR].inode;
    xr = MAPx(anode[inodeR].x);
    yr = MAPy(anode[inodeR].y);
    GRLine(x,y,xr,yr);
    GRPutPixel(x,y,GR_YELLOW);
    GRPutPixel(xr,yr,GR_YELLOW);
    ivtx = ivtxR;
    x = xr;
    y = yr;
  }

  return(0);

}

int nDrawTriangle(PNODE anode, INODE inodeL, INODE inodeM, INODE inodeR)
{
  int x1,y1;
  int x2,y2;
  int x3,y3;

  x1 = MAPx(anode[inodeL].x);
  y1 = MAPy(anode[inodeL].y);
  x2 = MAPx(anode[inodeM].x);
  y2 = MAPy(anode[inodeM].y);
  x3 = MAPx(anode[inodeR].x);
  y3 = MAPy(anode[inodeR].y);

  GRSetColor(GR_WHITE);
  GRLine(x1,y1,x2,y2);
  GRLine(x2,y2,x3,y3);
  GRLine(x3,y3,x1,y1);
  return(0);

}


int nDrawRectangle(PNODE anode, INODE inode1, INODE inode2,
				INODE inode3, INODE inode4)
{
  int x1,y1;
  int x2,y2;
  int x3,y3;
  int x4,y4;

  x1 = MAPx(anode[inode1].x);
  y1 = MAPy(anode[inode1].y);
  x2 = MAPx(anode[inode2].x);
  y2 = MAPy(anode[inode2].y);
  x3 = MAPx(anode[inode3].x);
  y3 = MAPy(anode[inode3].y);
  x4 = MAPx(anode[inode4].x);
  y4 = MAPy(anode[inode4].y);

  GRSetColor(GR_WHITE);
  GRLine(x1,y1,x2,y2);
  GRLine(x2,y2,x3,y3);
  GRLine(x3,y3,x4,y4);
  GRLine(x4,y4,x1,y1);
  return(0);

}


int nPlotPoint(PNODE anode, INODE inode, BOOL fColor)
{
  int x,y;

  x = MAPx(anode[inode].x);
  y = MAPy(anode[inode].y);

  GRPutPixel(x,y,(fColor) ? GR_RED : GR_GREEN);
  return(0);

}


int nFillPoly(PNODE anode, INODE *ainode, WORD cnode)
{
  int nPolyPnts[32];
  int i;

  for(i = 0; i < cnode; ++i)
  { nPolyPnts[i*2]   = MAPx(anode[ainode[i]].x);
    nPolyPnts[i*2+1] = MAPy(anode[ainode[i]].y);
  }
  nPolyPnts[2*cnode]   = nPolyPnts[0];
  nPolyPnts[2*cnode+1] = nPolyPnts[1];

  GRFillPoly(cnode,nPolyPnts);
  GRSetColor(GR_YELLOW);
  GRDrawPoly(cnode+1,nPolyPnts);
  return(0);

}


int nTriangles(PNODE anode, INODE *ainode, WORD cnode, INODE inode)
{
  int nPolyPnts[32];
  int x,y;
  int i;

  for(i = 0; i < cnode; ++i)
  { nPolyPnts[i*2]   = MAPx(anode[ainode[i]].x);
    nPolyPnts[i*2+1] = MAPy(anode[ainode[i]].y);
  }
  nPolyPnts[2*cnode]   = nPolyPnts[0];
  nPolyPnts[2*cnode+1] = nPolyPnts[1];
  x = MAPx(anode[inode].x);
  y = MAPy(anode[inode].y);

  GRSetColor(GR_WHITE);
  GRDrawPoly(cnode+1,nPolyPnts);
  for(i = 0; i < cnode; ++i)
    GRLine(x,y,nPolyPnts[i*2],nPolyPnts[i*2+1]);

  return(0);

}


int
nFillTriangle(
PNODE anode,
INODE inodeA,
INODE inodeB,
INODE inodeC,
int   nColor)
{
  int color;
  int nPolyPnts[8];

  switch(nColor)
  { case 0:  color = GR_BLACK;	 break;
    case 1:  color = GR_YELLOW;	 break;
    case 2:  color = GR_RED;	 break;
  }

  nPolyPnts[0] = MAPx(anode[inodeA].x);
  nPolyPnts[1] = MAPy(anode[inodeA].y);
  nPolyPnts[2] = MAPx(anode[inodeB].x);
  nPolyPnts[3] = MAPy(anode[inodeB].y);
  nPolyPnts[4] = MAPx(anode[inodeC].x);
  nPolyPnts[5] = MAPy(anode[inodeC].y);
  nPolyPnts[6] = nPolyPnts[0];
  nPolyPnts[7] = nPolyPnts[1];

  GRSetFillColor(color);
  GRFillPoly(3,nPolyPnts);
  GRSetColor(GR_WHITE);
  GRDrawPoly(4,nPolyPnts);
  return(0);

}


int
nFillRectangle(
PNODE anode,
INODE inodeA,
INODE inodeB,
INODE inodeC,
INODE inodeD,
int   nColor)
{
  int color;
  int nPolyPnts[10];

  switch(nColor)
  { case 0:  color = GR_BLACK;	 break;
    case 1:  color = GR_YELLOW;	 break;
    case 2:  color = GR_RED;	 break;
  }

  nPolyPnts[0] = MAPx(anode[inodeA].x);
  nPolyPnts[1] = MAPy(anode[inodeA].y);
  nPolyPnts[2] = MAPx(anode[inodeB].x);
  nPolyPnts[3] = MAPy(anode[inodeB].y);
  nPolyPnts[4] = MAPx(anode[inodeC].x);
  nPolyPnts[5] = MAPy(anode[inodeC].y);
  nPolyPnts[6] = MAPx(anode[inodeD].x);
  nPolyPnts[7] = MAPy(anode[inodeD].y);
  nPolyPnts[8] = nPolyPnts[0];
  nPolyPnts[9] = nPolyPnts[1];

  GRSetFillColor(color);
  GRFillPoly(4,nPolyPnts);
  GRSetColor(GR_WHITE);
  GRDrawPoly(5,nPolyPnts);
  return(0);

}


int
nQualReport(
WORD cnode,
WORD cedge,
WORD ctri,
WORD crect,
WORD cObtuse)
{
  int  x,y;

  switch (cmdargs.nResolution)
  { case RES_640x480:	x = 480;  y = 10;  break;
    case RES_800x600:	x = 600;  y = 10;  break;
    case RES_1024x768:	x = 768;  y = 10;  break;
  }

  GRSetColor(GR_WHITE);
  sprintf(szGlobalBuf, "%d nodes", cnode);
  GRText(x+20, y,    szGlobalBuf);
  sprintf(szGlobalBuf, "%d edges", cedge);
  GRText(x+20, y+20, szGlobalBuf);
  sprintf(szGlobalBuf, "%d triangles", ctri);
  GRText(x+20, y+40, szGlobalBuf);
  sprintf(szGlobalBuf, "%d rectangles", crect);
  GRText(x+20, y+60, szGlobalBuf);
  sprintf(szGlobalBuf, "%d obtuse angles", cObtuse);
  GRText(x+20, y+80, szGlobalBuf);
  return(0);
}



