/*-----------------------------------------------------------------------------

 FILE:	   SKEL.H

 PROJECT:  TRIANGULAR GRID GENERATOR

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 10/19/93    KMK      initial implementation
 10/28/93    KMK      second rewrite

 DESCRIPTION:

 This module contains the data structures used to describe the skeleton of
 the simulation domain.

-----------------------------------------------------------------------------*/


#ifndef HEADER_SKEL
#define HEADER_SKEL


/******************************************************************************/
/***********************  PUBLIC TYPE DEFINITION  *****************************/
/******************************************************************************/


typedef WORD IPNT;			/* ipnt */
typedef struct tagPNT			/* pnt */
{ REAL	x;				/* x coordinate */
  REAL	y;				/* y coordinate */
} PNT;
typedef PNT *PPNT;			/* ppnt */

typedef WORD ISEG;			/* iseg */
typedef struct tagSEG			/* seg */
{ WORD	wLabel;				/* segment label */
  IPNT	ipntA;				/* index of point A */
  IPNT	ipntB;				/* index of point B */
  IPNT	ipntM;				/* index of arc midpoint */
  REAL	xCenter;			/* x coordinate of arc's center */
  REAL	yCenter;			/* y coordinate of arc's center */
  REAL	Radius;				/* radius of arc */
  REAL	ThetaA;				/* angle of point A */
  REAL	ThetaB;				/* angle of point B */
  IPNT *aipnt;				/* array of edge points */
  WORD	cipnt;				/* number of edge points */
  WORD	cSegment;			/* number of segments */
  REAL	InitialSpacing;			/* initial spacing */
  REAL	SpacingGrade;			/* spacing grade */
} SEG;
typedef SEG *PSEG;			/* pseg */

typedef WORD IREG;			/* ireg */
typedef struct tagREG			/* reg */
{ REAL	Hmax;				/* maximum edge length */
  WORD	wLabel;				/* region label */
  ISEG *aiseg;				/* segment array */
  WORD	ciseg;				/* size of segment array */
  BOOL	fRect;				/* rectangular elements */
} REG;
typedef REG *PREG;			/* preg */

typedef struct tagSKEL			/* skel */
{ PSTR	pszName;			/* domain name */
  PSTR	pszUnit;			/* unit of distance */
  REAL	xMin;				/* minimum x coordinate in skeleton */
  REAL	yMin;				/* minimum y coordinate in skeleton */
  REAL	xMax;				/* maximum x coordinate in skeleton */
  REAL	yMax;				/* maximum y coordinate in skeleton */
  PPNT	apnt;				/* array of points */
  PSEG	aseg;				/* array of segments */
  PREG	areg;				/* array of regions */
  WORD	capnt;				/* size of point array */
  WORD	caseg;				/* size of segment array */
  WORD	careg;				/* size of region array */
} SKEL;
typedef SKEL *PSKEL;			/* pskel */


/******************************************************************************/
/*********************	PUBLIC CONSTANTS AND MACROS  **************************/
/******************************************************************************/


#define ppntGetPoint(pskel,ipnt)    (pskel->apnt) + ipnt
#define psegGetSegment(pskel,iseg)  (pskel->aseg) + iseg
#define pregGetRegion(pskel,ireg)   (pskel->areg) + ireg
#define NOT_DEFINED -1

/******************************************************************************/
/**************************  PUBLIC FUNCTIONS  ********************************/
/******************************************************************************/


#if defined(ANSIPROTO)
PSKEL	pskelCreate   (PSTR);
#else
PSKEL	pskelCreate   ();
#endif


#endif
