/*-----------------------------------------------------------------------------

 FILE:	   TYPEDEFS.H

 PROJECT:  TRIANGULAR GRID GENERATOR

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 10/19/93    KMK      initial implementation
 10/30/93    KMK      second rewrite

 DESCRIPTION:

 This module contains type definitions and constants used by all of the
 SGLAT modules.

-----------------------------------------------------------------------------*/


#ifndef	 HEADER_TYPEDEFS
#define	 HEADER_TYPEDEFS


/* constant definitions */
#define TRUE		    1	    /* boolean true	     */
#define FALSE		    0	    /* boolean false	     */
#define LEN_IDENT	    15	    /* length of identifiers */


/* type definitions */
typedef unsigned char	  BYTE;		/* b   */
typedef unsigned char	  UCHAR;	/* uch */
typedef int		  BOOL;		/* f   */
typedef int		  BOOLEAN;	/* f   */
typedef unsigned int	  WORD;		/* w   */
typedef unsigned int	  UINT;		/* u   */
typedef unsigned long	  DWORD;	/* dw  */
typedef char *		  PSTR;		/* psz */
typedef int  *		  PINT;		/* pn  */
typedef int  *		  PBOOL;	/* pf  */
typedef unsigned int *	  PWORD;	/* pw  */
typedef long *		  PLONG;	/* pl  */
typedef unsigned long *	  PDWORD;	/* pdw */
typedef void *		  PVOID;	/* pv  */
typedef double		  REAL;		/*     */
typedef double *	  PREAL;	/*     */


/* definition of command line argument structure */
#define ERRDEV_STDERR	    1	    /* errors sent to standard error */
#define ERRDEV_STDOUT	    2	    /* errors sent to standard output */
#define ERRDEV_FILE	    3	    /* errors sent to specified file */
#define VISUAL_CONSTRUCT    1	    /* visualize mesh construction */
#define VISUAL_RATING	    2	    /* visualize mesh rating */
#define VISUAL_NONE	    3	    /* no graphics */
#define RES_640x480	    1	    /* 640 x 480 resolution */
#define RES_800x600	    2	    /* 800 x 600 resolution */
#define RES_1024x768	    3	    /* 1024 x 768 resolution */


typedef struct tagCMDARGS	    /* cmdargs */
{ PSTR	  pszSource;		    /* source file */
  PSTR	  pszAsciiMesh;		    /* ascii mesh file name */
  WORD	  wErrDev;		    /* standard error / standard out */
  WORD	  cQualEnh1;		    /* number of quality enhancement iter. */
  WORD	  cQualEnh2;		    /* number of quality enhancement iter. */
  WORD	  cQualEnh3;		    /* number of quality enhancement iter. */
  WORD	  wVisual;		    /* visulation level */
  BOOL	  fExhaustive;		    /* exhaustive search */
  BOOL	  fGraded;		    /* graded spacing */
  BOOL	  fAutoSpace;		    /* auto mesh spacing */
  BOOL	  fNoCentroid;		    /* no centroid triangulation */
  BOOL	  fKeepAspect;		    /* keep x and y aspect ration */
  BOOL	  fGeneratePSFile;	    /* generate a postscript file */
  BOOL	  fNoQualReport;	    /* do not generate a quality report */
  REAL	  Grade;		    /* mesh spacing grade */
  REAL	  Hmax;			    /* maximum mesh spacing */
  REAL	  Qmin;			    /* minimum permissible quality */
  int	  nResolution;		    /* graphics screen resolution */
  int	  cNoPress;
  int	  cPresses;
} CMDARGS;
extern CMDARGS cmdargs;

#define PI   3.141592653589793
#define EPS  1.0e-12
#define GLOBAL_BUFSIZE	 256
extern char  szGlobalBuf[GLOBAL_BUFSIZE];

/* return code definitions */
#define RETCODE_SUCCESS	    0	    /* successful execution */
#define RETCODE_CMDLINE	    1	    /* command line error */
#define RETCODE_ERROR	    2	    /* system or parsing error */
#define RETCODE_PROGERROR   3	    /* programmer's error */

/* global macros */
#ifdef max
#undef max
#endif
#ifdef min
#undef min
#endif
#define max(x,y)	    (((x) > (y)) ? (x) : (y))
#define min(x,y)	    (((x) < (y)) ? (x) : (y))
#define MEMCHECK(p,t1,t2)   { if (p == NULL)				    \
	vErrorHandler(ERROR_OUTOFMEM, t1, t2, TRUE); }

/* global function prototypes */
#if defined(ANSIPROTO)
void	vExitFunction	(void);
PSTR	pszLowerCase	(PSTR);
PSTR	pszUpperCase	(PSTR);
double	sq		(double);
#else
void	vExitFunciton	();
PSTR	pszLowerCase	();
PSTR	pszUpperCase	();
double	sq		();
#endif


#define DEBUG_INFO(s) fprintf(stdout, s);


#if defined(UNIX)
#define getch getchar
#endif


#endif
