/*-----------------------------------------------------------------------------

 PROJECT:  SIMULATION GENERATION FRAMEWORK (SGFramework)

 SUBSYSTEM:  grid.exe
 FILE:	     xlatinfo.h
 AUTHOR:     Kevin M. Kramer

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 11/02/94    KMK      initial implementation

 DESCRIPTION:

 This module contains the xlat structures that are shared among header files

-----------------------------------------------------------------------------*/


#ifndef __xlatinfo_h
#define __xlatinfo_h


/* skeleton header */
typedef struct tagSKELHEAD
{ char szLogo[64];			/* text logo */
  char szSign[16];			/* signature */
  char szCoord1[LEN_IDENT+1];		/* name of 1st coordinate */
  char szCoord2[LEN_IDENT+1];		/* name of 2nd coordinate */
  UINT cMinDivisions;			/* minimum divisions */
  UINT cMaxDivisions;			/* maximum divisions */
  UINT cConstant;			/* number of constants */
  UINT cLabel;				/* number of labels */
  UINT cRefine;				/* number of refinements */
  UINT cRegLabel;			/* number of region labels */
  UINT cPoint;				/* number of points */
  UINT cEdge;				/* number of edges */
  UINT cRegion;				/* number of regions */
} SKELHEAD;


/* constant structure definition */
typedef struct tagXLATCONST
{ char szName[LEN_IDENT+1];		/* constant name */
  UINT uType;				/* constant type */
  union
  { REAL r;				/* value of real constant */
    int	 n;				/* value of integer constant */
  } data;
} XLATCONST;


/* label structure definition */
typedef struct tagXLATLABEL
{ char szName[LEN_IDENT+1];		/* label name */
  UINT iIndex;				/* label index */
  UINT uType;				/* label type */
  UINT cNode;				/* number of nodes */
} XLATLABEL;


/* refinement structure definition */
typedef struct tagXLATREFINE
{ char szName[LEN_IDENT+1];		/* refinement variable name */
  UINT iIndex;				/* refinement variable index */
  UINT uMeasure;			/* type of measurement */
  REAL Distance;			/* minimum "distance" */
} XLATREFINE;


/* point structure definition */
typedef struct tagXLATPOINT
{ REAL x;				/* x coordinate */
  REAL y;				/* y coordinate */
} XLATPOINT;


/* edge structure definition */
typedef struct tagXLATEDGE
{ REAL InitialSpacing;			/* initial spacing */
  REAL SpacingGrade;			/* spacing grade */
  UINT iLabel;				/* label index */
  UINT iPointA;				/* first endpoint */
  UINT iPointB;				/* midpoint (may not be present) */
  UINT iPointC;				/* second endpoint */
  UINT cSegment;			/* number of segments */
} XLATEDGE;


/* region structure definition */
typedef struct tagXLATREGION
{ REAL Hmax;				/* maximum edge length */
  UINT iLabel;				/* label index */
  UINT cEdge;				/* number of edges */
  BOOL fRect;				/* rectangular elements */
} XLATREGION;


#endif
