/*-----------------------------------------------------------------------------

 PROJECT:  SIMULATION GENERATION FRAMEWORK (SGFramework)

 SUBSYSTEM:  mesh.exe
 FILE:	     codebuf.cpp
 AUTHOR:     Kevin M. Kramer

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 09/10/94    KMK      initial implementation
 10/20/94    KMK      modified for use in command line version

 DESCRIPTION:

 This module implements the code buffer class.

-----------------------------------------------------------------------------*/


#ifndef __codebuf_h
#define __codebuf_h

#define LEN_BUFFER 512
#define LEN_PREFIX 3

#ifndef __mathnode_h
#include "mathnode.h"
#endif


// code buffer class
class CodeBuffer
{ protected:
    char szFileName[LEN_FILENAME];	// file name
    char szUCPrefix[LEN_PREFIX + 1];	// uppercase prefix
    char szLCPrefix[LEN_PREFIX + 1];	// lowercase prefix
    char szBuffer[LEN_BUFFER];		// buffer
    int	 nFile;				// file handle
    UINT iLeftMargin;			// left margin
    UINT iRightMargin;			// right margin
    UINT cPrecision;			// number of significant digits
    UINT iBuffer;			// buffer index

  protected:
    void Flush ();			// flush the buffer
    void AddToken (char *);		// add token to buffer

  public:
    inline void SetRightMargin (UINT u)	    { iRightMargin = u; }
    inline void AddMathNode (MathNode *pmn) { pmn->WriteCode((void *) this); }
    void SetLeftMargin (UINT);
    void AddChar (char);
    void AddInteger (int, BOOL = TRUE);
    void AddUInteger (UINT, BOOL = TRUE);
    void AddReal (REAL, BOOL = TRUE);
    void AddCoordinate (UINT, BOOL = TRUE);
    void AddFunction (char *, UINT, BOOL = FALSE);
    void AddFunction (UINT, UINT, BOOL = FALSE);
    void AddString (char *);
    void AddText (char *);
    void AddEquation (UINT, BOOL = FALSE);

  public:
    CodeBuffer (char *, char *, UINT = 7, UINT = 0, UINT = 78);
   ~CodeBuffer ();
};


#endif
