/*-----------------------------------------------------------------------------

 PROJECT:  SIMULATION GENERATION FRAMEWORK (SGFramework)

 FILE:	     error.h
 AUTHOR:     Kevin M. Kramer
             J. Joshua Feng

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 02/25/94    KMK      initial implementation
 07/06/00    JJF      modify AddMessage

 DESCRIPTION:

 This module implements the SGFramework error handler.

-----------------------------------------------------------------------------*/


#ifndef __error_h
#define __error_h

class ErrorHandler
{ private:
    UINT cMaxErrors;			// maximum number of errors
    UINT cMaxWarnings;			// maximum number of warnings
    UINT cErrors;			// number of errors in message array
    UINT cWarnings;			// number of warnings in message array
    UINT cFatal;			// number of fatal errors
    UINT cTotal;			// total number of messages

    typedef struct tagMESSAGE		// msg
    { char *psz;			// pointer to message
      UINT  iLine;			// line index
      UINT  iCol;			// column index
    } MESSAGE;

    MESSAGE *amsg;			// message array
    void *p;				// corresponding source client
    UINT iBuf;				// buffer index

    UINT uMessagePrefix (UINT, UINT, UINT &);

  public:
    inline BOOL fFatal	    ()	     { return cFatal
                                           || (cErrors >= cMaxErrors)
                                           || (cWarnings >= cMaxWarnings); }
    inline BOOL fRunTime    ()	     { return cTotal; }
    inline BOOL fError	    ()	     { return (cErrors) ? TRUE : FALSE; }
    inline BOOL fWarning    ()	     { return (cWarnings) ? TRUE : FALSE; }
    inline UINT cNumMsg	    ()	     { return cErrors + cWarnings + cFatal; }
    inline void SetBufPos   (UINT i) { iBuf = i; }
    // inline void * GetSource ()	     { return p; }

    void Initialize (void *);
    void AddMessage (UINT,		               UINT = -1u);
    void AddMessage (UINT, const char *,	       UINT = -1u);
    void AddMessage (UINT, const char *, const char *, UINT = -1u);
    void ShowMessages ();

  public:
     ErrorHandler (UINT = 25, UINT = 25);
    ~ErrorHandler ();
};


// global reference of error instance
extern ErrorHandler *error;


#define MSG_WARNING   0
#define MSG_ERROR     1
#define MSG_FATAL     2


typedef struct tagEHM			// ehm
{ UINT uType;				// type of message
  BOOL fLineInfo;			// include line information
  char *pszMsg;				// message text
} EHM;

extern EHM aehm[];

#define EHM_STRLF	   0		// string cannot span multiple lines
#define EHM_UNEXPEOF	   1		// unexpected end of file
#define EHM_SYNTAX	   2		// %s expected
#define EHM_SYMREDEF	   3		// redefinition of symbol "?"
#define EHM_SYMNOTDEF	   4		// symbol "?" not defined
#define EHM_SYMEXP	   5		// "?" is not a ? symbol
#define EHM_NOTDEC	   6		// ? not declared
#define EHM_NOTDEC1	   7		// ? symbol "?" not declared
#define EHM_NOTDEC2	   8		// coordinates not declared
#define EHM_NOTDEC3	   9		// function "?" not declared
#define EHM_CONST	  10		// "?" does not evaluate to a constant
#define EHM_POSCONST	  11		// ? does not evaluate to a + const
#define EHM_POSICONST	  12		// ? does not evaluate to a + int const
#define EHM_NUMARGS	  13		// func "?" has the wrong # of args
#define EHM_DIVBYZERO	  14		// division by zero in ?
#define EHM_DOMAIN	  15		// domain error in ? function
#define EHM_OVERFLOW	  16		// overflow in ? function
#define EHM_UNDERFLOW	  17		// underflow in ? function
#define EHM_INTOVER	  18		// integer overflow
#define EHM_FILEOPEN	  19		// cannot open file "?"
#define EHM_NORECTREG	  20		// cannot subdivide reg into rect elem

#endif
