/*-----------------------------------------------------------------------------

 PROJECT:  SIMULATION GENERATION FRAMEWORK (SGFramework)

 SUBSYSTEM:  mesh.exe
 FILE:	     global.h
 AUTHOR:     Kevin M. Kramer

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 11/01/94    KMK      initial implementation

 DESCRIPTION:

 This module contains type definitions and constants used by all of the
 SGLAT modules.

-----------------------------------------------------------------------------*/


#ifndef __typedefs_h
#define __typedefs_h


///////////////////////////////////////////////////////////////////////////////
////////////////////////////  TYPE DEFINITIONS	///////////////////////////////
///////////////////////////////////////////////////////////////////////////////


// boolean constants
#define TRUE			1
#define FALSE			0


// global constants
#define LEN_IDENT		15
#define LEN_FILENAME		128
#define GLOBAL_BUFSIZE		512


// refinement measurement types
#define DISTANCE_LINEAR	      0		// measure distance linearly
#define DISTANCE_LOG	      1		// measure distance logarithmically
#define DISTANCE_SLOG	      2		// measure distance w/ signed log


// global type definitions
typedef int		  BOOL;				      // f
typedef unsigned int	  UINT;				      // u
typedef unsigned char	  UCHAR;			      // uch
typedef int		  I3ARRAY[3];			      // i3a
typedef void *		  PVOID;			      // pv
typedef double		  REAL;				      // r
typedef double *	  PREAL;			      // pr
typedef void *		  PRF;				      // prf
typedef double		  (*PRF1)(double);		      // prf1
typedef double		  (*PRF2)(double, double);	      // prf2
typedef double		  (*PRF3)(double, double, double);    // prf3


// global definition
extern char* pszGlobalBuf;
extern UINT uRefineMeasure;
extern REAL rRefineDistance;
extern void *pMinEdgeLen;
extern void *pMaxEdgeLen;
extern void *pMaxXEdgeLen;
extern void *pMaxYEdgeLen;
extern UINT cMaxDivisions;
extern UINT cMinDivisions;


// command line parser
typedef struct tagCMDARGS		// cmdargs
{ UINT cPrecision;			// output precision
} CMDARGS;
extern CMDARGS cmdargs;


#if defined(NODYNCAST)
#define TYPESAFE_DOWNCAST(object,toClass) ((toClass *) (object))
#else
#define TYPESAFE_DOWNCAST(object,toClass) dynamic_cast<toClass *>(object)
#endif


#if defined(UNIX)
char *strlwr(char *);
char *strupr(char *);
double pow10(double);
int stricmp(char *, char *);
#define O_BINARY 0x0000
#endif


#endif

