/*-----------------------------------------------------------------------------

 PROJECT:  SIMULATION GENERATION FRAMEWORK (SGFramework)

 SUBSYSTEM:  mesh.exe
 FILE:	     hash.h
 AUTHOR:     Kevin M. Kramer

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 10/31/94    KMK      initial implementation

 DESCRIPTION:

 This module implements a hash table.

-----------------------------------------------------------------------------*/


#ifndef __hash_h
#define __hash_h


// forward declaration
class HashTable;


// This class implements a hash node.
//   szName[] is the node's name which serves as the search key
//   uType is the node's type
//   phnodeNext() returns the next node
//   link() links the passed node to this node
class HashNode
{ protected:
    HashNode *phnode;			// link to next node
    inline HashNode *phnodeNext ()	{ return phnode; }
    inline void Link (HashNode *p)	{ phnode = p; }

  public:
    char szName[LEN_IDENT + 1];
    UINT uType;

  public:
    HashNode (char *, UINT);
    virtual ~HashNode ();

  friend class HashTable;
};


// this class is a dynamic array of hash nodes ...
// this was original defined as a template but was recoded since templates
// are not implemented the same way across platforms.
class DAPHashNode
{ private:
    UINT cSize;				// total number of elements
    UINT uInc;				// increment size
    UINT cElements;			// number of elements in array
    BOOL fManaged;			// managed array
    HashNode **dynarray;		// dynamic array

  public:
    void SetManaged (BOOL f)		{ fManaged = f; }
    UINT GetElements ()			{ return cElements; }
    UINT Add (HashNode *t);		// add an item to the array
    void RemoveLast ();			// remove last element
    void Remove(UINT i);		// remove an element
    void Flush ();			// remove all pointers from the array
    HashNode * operator[](UINT i)	{ return dynarray[i]; }

  public:
     DAPHashNode (UINT = 16);
    ~DAPHashNode ();
};


// This class implements a chain bucket hash table
//   Store() stores a node to the hash table
//   Retrieve() retrieves a node from the hash table
class HashTable
{ protected:
    UINT  cSize;			// size of the hash table
    HashNode **aphnode;			// hash table
    UINT cAlphabet;			// size of the search key alphabet
    UINT auXlat['z' + 1];		// ASCII to alphabet translation table
    UINT uHash (char *);		// hash function

  public:
    void Store	(HashNode *);
    void Remove (char *);
    HashNode *Retrieve (char *);

  public:
    HashTable (UINT = 397);
    virtual ~HashTable ();
};


#endif


