//
// AUTHOR:      Kevin M. Kramer
//              Eric R. Keiter
//              J. Joshua Feng
// 
// Description: No comment (//) is allowed
//              grammar is described as commented by //
// ---------------------------------------------------------------------------
// 10 9                        // # of x grid points, # of y grid points
// 0.0 0.0                     // (x,y) of the upper-left corner
//
// 3.125e+0 + (3.125e-1, 0.0)  // x data:  length + (division_len, grade)
// 1.250e+0 + (3.125e-1, 0.0)
// 6.250e-1 + (3.125e-1, 0.0)
// 1.250e+0 + (3.125e-1, 0.0)
// 6.250e-1 + (3.125e-1, 0.0)
// 1.250e+0 + (3.125e-1, 0.0)
// 6.250e-1 + (3.125e-1, 0.0)
// 1.250e+0 + (3.125e-1, 0.0)
// 6.250e-1 + (3.125e-1, 0.0)
// 5.000e-1 + (5.00e+0, 0.0)
//
// 5.000e-1 + (5.00e-1, 0.0)   // y data:  length + (division_len, grade)
// 8.000e+0 + (5.00e-1, 0.0)
// 5.000e-1 + (5.00e-1, 0.0)
// 5.000e-1 + (5.00e-1, 0.0)
// 2.000e+0 + (5.00e-1, 0.0)
// 1.350e+1 + (5.00e-1, 0.0)
// 5.000e-1 + (5.00e-1, 0.0)
// 2.000e+0 + (5.00e-1, 0.0)
// 1.000e+0 + (5.00e-1, 0.0)
//
// 9                           // # of regions
// P PLASMA R                  // graph_symbol region_label rec/tri
// C COIL R
// D DIELECTRIC R
// V VACUUM R
// M METAL R
// W WAFER R
// H CHUCK R
// L PLASMA2 R
// A PLASMA3 R
//
// 3                           // # of boundary edges
// 1 NOFLUX                    // graph_symbol edge_label
// 2 WALL1
// 3 WALL2
//                             // (block) graph of the grid
//       0 0 0 0 0 0 0 0 0 1        // skip the first line 
//       1 2 3 4 5 6 7 8 9 0        // skip the second line 
//                                  // skip the third line
//       2 2 2 2 2 2 2 2 2 2        // skip first 4 char's of each line follows 
// 01  1 M M M M M M M M M M 2 
// 02  1 V V V V V V V V V M 2
// 03  1 V C V C V C V C V M 2
// 04  1 V V V V V V V V V M 2
// 05  1 D D D D D D D D D M 2
// 06  1 P P P P P P P P P M 2
// 07  1 W W W H H H H L L M 2
// 08  1 H H H H H H H L L M 2
// 09  1 H H H H H H H A A M 2
//       2 2 2 2 2 2 2 3 3 2
//-----------------------------------------------------------------------------
 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

class RectSpec
{ protected:
    FILE *fileIn;                       // input file
    FILE *fileOut;                      // output file

  protected:
    int  cx;                            // number of horizontal cells
    int  cy;                            // number of vertical cells
    double *ax;                         // x coordinates
    double *ay;                         // y coordinates
    int *afx;                           // cell edge from left to right point
    int *afy;                           // cell edge from top to bottom point
    char **apszx;                       // x spacing information
    char **apszy;                       // y spacing information

  protected:
    int cEL;                            // number of edge labels
    int cRL;                            // number of region labels
    char **apszEL;                      // edge labels
    char **apszRL;                      // region labels
    char **apszRLtr;                    // triangle or rectangle indicator

  protected:
    char  **aach;                       // cell information

  protected:
    void ReadNumberOfCells ();
    void InitializeFirstCoordinates ();
    void ReadHorizontalCellInformation ();
    void ReadVerticalCellInformation ();
    void ReadRegionLabels ();
    void ReadEdgeLabels ();
    void ReadCellInformation ();

  protected:
    inline int Index(int i, int j)      { return j*(2*cx+1)+i; }

  public:
    void PrintBlockDiagram ();
    void WritePoints ();
    void WriteEdges ();
    void WriteRegions ();
    void WriteRefineInformation ();

  public:
    RectSpec (FILE *, FILE *);
   ~RectSpec ();
};

char sz[512];

///////////////////////////////////////////////////////////////////////////////
// Description: Constructor
RectSpec::RectSpec ( FILE *fin, FILE *fout )
{
  fileIn = fin;
  fileOut = fout;
  ReadNumberOfCells();
  InitializeFirstCoordinates();
  ReadHorizontalCellInformation();
  ReadVerticalCellInformation();
  ReadRegionLabels();
  ReadEdgeLabels();
  ReadCellInformation();
}


///////////////////////////////////////////////////////////////////////////////
// Description: Destructor
RectSpec::~RectSpec ( )
{
  int i;
  delete [] ax;
  delete [] ay;
  delete [] afx;
  delete [] afy;
  for(i = 0; i < cx; ++i) delete [] apszx[i];
  for(i = 0; i < cy; ++i) delete [] apszy[i];
  for(i = 0; i < cEL; ++i) delete [] apszEL[i];
  for(i = 0; i < cRL; ++i) delete [] apszRL[i];
  for(i = 0; i < cRL; ++i) delete [] apszRLtr[i];
  for(i = 0; i < 2*cy+1; ++i) delete [] aach[i];
  delete [] apszx;
  delete [] apszy;
  delete [] apszEL;
  delete [] apszRL;
  delete [] apszRLtr;
  delete [] aach;
}

///////////////////////////////////////////////////////////////////////////////
void RectSpec::ReadNumberOfCells ( )
{
  // read the number of horizontal and vertical blocks
  fgets(sz, sizeof(sz), fileIn);
  sscanf(sz, "%d %d", &cx, &cy);

  // allocate the data structures
  ax = new double[cx+1];
  ay = new double[cy+1];
  afx = new int[cx];
  afy = new int[cy];
  apszx = new char*[cx];
  apszy = new char*[cy];
  aach = new char*[2*cy+1];
  for(int i = 0; i < 2*cy+1; ++i)
    aach[i] = new char[2*cx+2];
}

///////////////////////////////////////////////////////////////////////////////
void RectSpec::InitializeFirstCoordinates ( )
{
  // read the x and y coordinates of the upperleft hand block
  // read the number of horizontal and vertical blocks
  fgets(sz, sizeof(sz), fileIn);
  sscanf(sz, "%lg %lg", ax, ay);
  fgets(sz, sizeof(sz), fileIn);
}

///////////////////////////////////////////////////////////////////////////////
void RectSpec::ReadHorizontalCellInformation ( )
{
  for(int i = 0; i < cx; ++i)
  { // initialize the spacing information
    fgets(sz, sizeof(sz), fileIn);
    char *psz  = strchr(sz, '(');
    char *pszT = (psz != NULL) ? strchr(psz, ')') : (char *) NULL;
    if (pszT)
    { pszT[1] = '\0';
      apszx[i] = new char[strlen(psz)+2];
      apszx[i][0] = ' ';
      strcpy(apszx[i]+1, psz);
    }
    else
    { apszx[i] = new char[1];
      apszx[i] = '\0';
    }

    // initialize the coordinate (read the manual page of strtok, please.)
    psz = strtok(sz, " \n");
    ax[i+1] = ax[i] + atof(psz);
    if ((ax[i+1] > -1.0e-12) && (ax[i+1] < 1.0e-12)) ax[i+1] = 0.0;

    // initialize the cell edge flag
    psz = strtok(NULL, " \n");
    afx[i] = (psz && (*psz == '+')) ? 1 : 0;
  }
  fgets(sz, sizeof(sz), fileIn);
}

///////////////////////////////////////////////////////////////////////////////
void RectSpec::ReadVerticalCellInformation ( )
{
  for(int i = 0; i < cy; ++i)
  { // initialize the spacing information
    fgets(sz, sizeof(sz), fileIn);
    char *psz = strchr(sz, '(');
    char *pszT = (psz != NULL) ? strchr(psz, ')') : (char *) NULL;
    if (pszT)
    { pszT[1] = '\0';
      apszy[i] = new char[strlen(psz)+2];
      apszy[i][0] = ' ';
      strcpy(apszy[i]+1, psz);
    }
    else
    { apszy[i] = new char[1];
      apszy[i] = '\0';
    }

    // initialize the coordinate
    psz = strtok(sz, " \n");
    ay[i+1] = ay[i] - atof(psz);
    if ((ay[i+1] > -1.0e-12) && (ay[i+1] < 1.0e-12)) ay[i+1] = 0.0;

    // initialize the cell edge flag
    psz = strtok(NULL, " \n");
    afy[i] = (psz && (*psz == '+')) ? 1 : 0;
  }
  fgets(sz, sizeof(sz), fileIn);
}

///////////////////////////////////////////////////////////////////////////////
void RectSpec::ReadRegionLabels ( )
{
  fgets(sz, sizeof(sz), fileIn);
  cRL = atoi(sz);
  apszRL = new char*[cRL];
  apszRLtr = new char*[cRL];
  for(int i = 0; i < cRL; ++i)
  { apszRL[i] = new char[32];
    apszRLtr[i] = new char[32];
    fgets(sz, sizeof(sz), fileIn);
    apszRLtr[i][0] = sz[strlen(sz)-2];
    apszRLtr[i][1] = '\0';
    sz[strlen(sz)-3] = '\0';
    strcpy(apszRL[i], sz);
    // fprintf(stderr,"Region Label %d: %s  ", i, sz);
    // fprintf(stderr,"TriRect: %s\n", apszRLtr[i]);
  }
  fgets(sz, sizeof(sz), fileIn);
}

///////////////////////////////////////////////////////////////////////////////
void RectSpec::ReadEdgeLabels ( )
{
  fgets(sz, sizeof(sz), fileIn);
  cEL = atoi(sz);
  apszEL = new char*[cEL];
  for(int i = 0; i < cEL; ++i)
  { apszEL[i] = new char[32];
    fgets(sz, sizeof(sz), fileIn);
    sz[strlen(sz)-1] = '\0';
    strcpy(apszEL[i], sz);
  }
  fgets(sz, sizeof(sz), fileIn);
}

///////////////////////////////////////////////////////////////////////////////
void RectSpec::ReadCellInformation ( )
{
  int i,j;
  char **aachT = new char*[cy+2];           // Temperoty buffer
  fgets(sz, sizeof(sz), fileIn);
  fgets(sz, sizeof(sz), fileIn);
  fgets(sz, sizeof(sz), fileIn);

  // read the block information
  for(j = 0; j < cy+2; ++j)
  { aachT[j] = new char[cx+2];
    fgets(sz, sizeof(sz), fileIn);
    for(i = 0; i < cx+2; ++i)
      aachT[j][i] = sz[2*i+4];              // skip first 4 chars of each line
  }

  // initialize aach array
  for(j = 0; j < 2*cy+1; ++j)
  { for(i = 0; i < 2*cx+1; ++i) aach[j][i] = ' ';
    aach[j][2*cx+1] = '\0';
  }

  // contrust  * ?  *     ex. *1*1*1
  //           ? ch ?         2M|M|V
  //           * ?  *         *-*-*-
  // if ? = ch, it means hitting the wall
  char ch;
  for(j = 1; j <= cy; ++j)
    for(i = 1; i <= cx; ++i)
      if (isalpha(ch = aachT[j][i]))
      { aach[2*j-1][2*i-1] = ch;
        aach[2*j-2][2*i-2] = '*';
        aach[2*j-2][2*i-0] = '*';
        aach[2*j-0][2*i-2] = '*';
        aach[2*j-0][2*i-0] = '*';
        aach[2*j-1][2*i-2] = (isalpha(ch = aachT[j][i-1])) ? '|' : ch;
        aach[2*j-1][2*i-0] = (isalpha(ch = aachT[j][i+1])) ? '|' : ch;
        aach[2*j-2][2*i-1] = (isalpha(ch = aachT[j-1][i])) ? '-' : ch;
        aach[2*j-0][2*i-1] = (isalpha(ch = aachT[j+1][i])) ? '-' : ch;
      }

  for(j = 0; j < cy+2; ++j) delete [] aachT[j];
  delete [] aachT;
}

///////////////////////////////////////////////////////////////////////////////
void RectSpec::WritePoints ( )
{
  for(int j = 0; j < 2*cy+1; j += 2)
    for(int i = 0; i < 2*cx+1; i += 2)
      if (aach[j][i] == '*')
      { fprintf(fileOut, "point p%04x = (%.6le, %.6le);\n",
                Index(i,j), ax[i/2], ay[j/2]);
        aach[j][i] = ' ';
      }
  fprintf(fileOut, "\n");
}

///////////////////////////////////////////////////////////////////////////////
void RectSpec::WriteEdges ( )
{
  char ch;
  int i, j, k;

  for(j = 0; j < 2*cy+1; ++j)
    for(i = 0; i < 2*cx+1; ++i)
      if (isdigit(ch = aach[j][i]))
      { for(k = 0; k < cEL; ++k)
          if (ch == apszEL[k][0])
            break;

        // horizontal edgee
        if (i % 2)
        { int p1 = Index(i-1,j);
          int p2 = Index(i+1,j);
          if (!afx[(i-1)/2]) { int p3 = p1; p1 = p2; p2 = p3; }
          fprintf(fileOut, "edge e%04x = %s [p%04x, p%04x]%s;\n",
                  Index(i,j), apszEL[k]+2, p1, p2, apszx[(i-1)/2]);
        }

        // vertical edge
        else
        { int p1 = Index(i,j-1);
          int p2 = Index(i,j+1);
          if (!afy[(j-1)/2]) { int p3 = p1; p1 = p2; p2 = p3; }
          fprintf(fileOut, "edge e%04x = %s [p%04x, p%04x]%s;\n",
                  Index(i,j), apszEL[k]+2, p1, p2, apszy[(j-1)/2]);
        }

        aach[j][i] = ' ';
      }


  for(j = 0; j < 2*cy+1; ++j)
    for(i = 0; i < 2*cx+1; ++i)
    { ch = aach[j][i];
      if (ch == '-')
      { fprintf(fileOut, "edge e%04x = ", Index(i,j));
        char chT = aach[j-1][i];
        char chB = aach[j+1][i];
        if (chT != chB)
        { for(k = 0; k < cRL; ++k)
            if ((apszRL[k][0] == chT) || (apszRL[k][0] == chB))
              fprintf(fileOut, apszRL[k]+2);
          fprintf(fileOut, " ");
        }
        int p1 = Index(i-1,j);
        int p2 = Index(i+1,j);
        if (!afx[(i-1)/2]) { int p3 = p1; p1 = p2; p2 = p3; }
        fprintf(fileOut, "[p%04x, p%04x]%s;\n", p1, p2, apszx[(i-1)/2]);
        aach[j][i] = ' ';
      }
      else if (ch == '|')
      { fprintf(fileOut, "edge e%04x = ", Index(i,j));
        char chL = aach[j][i-1];
        char chR = aach[j][i+1];
        if (chL != chR)
        { for(k = 0; k < cRL; ++k)
            if ((apszRL[k][0] == chL) || (apszRL[k][0] == chR))
              fprintf(fileOut, apszRL[k]+2);
          fprintf(fileOut, " ");
        }
        int p1 = Index(i,j-1);
        int p2 = Index(i,j+1);
        if (!afy[(j-1)/2]) { int p3 = p1; p1 = p2; p2 = p3; }
        fprintf(fileOut, "[p%04x, p%04x]%s;\n", p1, p2, apszy[(j-1)/2]);
        aach[j][i] = ' ';
      }
    }
  fprintf(fileOut, "\n");
}

///////////////////////////////////////////////////////////////////////////////
void RectSpec::WriteRegions ( )
{
  int i, j, k;
  char ch;
  for(j = 1; j < 2*cy+1; j += 2)
    for(i = 1; i < 2*cx+1; i += 2)
      if (isalpha(ch = aach[j][i]))
      { for(k = 0; k < cRL; ++k)
          if (ch == apszRL[k][0])
            break;

        if( (apszRLtr[k][0] == 'T') || (apszRLtr[k][0] == 't') )
        { fprintf(fileOut,
                  "region r%04x = %s {e%04x, e%04x, e%04x, e%04x};\n",
                  Index(i,j), apszRL[k]+2, Index(i-1,j), Index(i,j+1),
                  Index(i+1,j), Index(i,j-1));
        }
        else
        { fprintf(fileOut,
                  "region r%04x = %s {e%04x, e%04x, e%04x, e%04x} RECTANGLES;\n",
                  Index(i,j), apszRL[k]+2, Index(i-1,j), Index(i,j+1),
                  Index(i+1,j), Index(i,j-1));
        }

        aach[j][i] = ' ';
      }

  fprintf(fileOut, "\n");
}

///////////////////////////////////////////////////////////////////////////////
void RectSpec::WriteRefineInformation ( )
{
  // write down the remain 'refine information' of the file
  while (fgets(sz, sizeof(sz), fileIn))
    fputs(sz, fileOut);
}

///////////////////////////////////////////////////////////////////////////////
void RectSpec::PrintBlockDiagram ( )
{
  int i;
  for(i = 0; i < 2*cy+1; ++i) fprintf(fileOut, "// %s\n", aach[i]);
  fprintf(fileOut, "\n");
  return;
}

///////////////////////////////////////////////////////////////////////////////
/////////////////////////////  MAIN PROGRAM  //////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
int main ( int cargs, char *arg[] )
{
  FILE *fp1, *fp2;

  if (cargs < 2)
  { printf("Usage: r2sk *.r *.sk [-p]\n");
    exit(0);
  }

  printf("r2sk: Program to convert from *.r to *.sk file\n");

  fp1 = fopen(arg[1], "r");

  if( (fp2 = fopen(arg[2], "r")) == NULL )
  { printf("converting %s to %s\n", arg[1], arg[2]);
  }
  else
  { printf("%s already exists.  (n/N to quit) or new name:", arg[2]);
    char name[100];
    fscanf(stdin,"%s", name);
    if(name[0] == 'n' || name[0] == 'N') 
    { printf("goodbye.\n"); 
      exit(0);
    }
    printf("converting %s to %s\n", arg[1], arg[2]); 
    fclose(fp2);
  }

  fp2 = fopen(arg[2], "w");

  RectSpec rectspec(fp1, fp2);

  if ((cargs == 4) && (arg[3][0] == '-')) rectspec.PrintBlockDiagram();
  rectspec.WritePoints();
  rectspec.WriteEdges();
  rectspec.WriteRegions();
  rectspec.WriteRefineInformation();

  fclose(fp1);
  fclose(fp2);
  return 0;
}
