/*-----------------------------------------------------------------------------

 PROJECT:  SIMULATION GENERATION FRAMEWORK (SGFramework)

 SUBSYSTEM:  mesh.exe
 FILE:	     source.cpp
 AUTHOR:     Kevin M. Kramer

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 11/01/94    KMK      initial implementation

 DESCRIPTION:

 This module implements a scanner and parser.

-----------------------------------------------------------------------------*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "global.h"

#ifndef UNIX
#pragma hdrstop
#endif

#if defined(UNIX)
#include <unistd.h>
#else
#include <io.h>
#endif

#include <fcntl.h>
#include "error.h"
#include "symbol.h"
#include "fileio.h"
#include "source.h"


// ****************************************************************************
// *****************************  MEMBER FUNCTIONS  ***************************
// ****************************************************************************


///////////////////////////////////////////////////////////////////////////////
// Description:
Source::Source ( char *pszFileName )
{
  // copy the filename into the buffer and open the file
  strncpy(szFileName, pszFileName, LEN_FILENAME);
}


///////////////////////////////////////////////////////////////////////////////
// Description:
Source::~Source ( )
{
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function reads the source file from scratch.
void Source::ReadSourceFile ( )
{
  pszBuf = NULL;
  cLines = 0;
  cChars = 0;

  // open the file
  int  nFile = open(szFileName, O_RDONLY | O_BINARY);
  if (nFile == -1)
  { error->AddMessage(EHM_FILEOPEN, szFileName);
    return;
  }

  // determine the length of the file
  cChars  = (UINT) lseek(nFile, (off_t) 0L, SEEK_END);
  lseek(nFile, (off_t) 0L, SEEK_SET);

  // create source buffer and read source into buffer
  pszBuf = new char[cChars + 1];
  read(nFile, pszBuf, cChars);
  pszBuf[cChars] = '\0';

  // form the line index array
  BOOL fNewLine = TRUE;
  char *psz = pszBuf;
  for(UINT i = 0; i < cChars; ++i, ++psz)
  { if (fNewLine) { DALineIndex.Add(i); fNewLine = FALSE; ++cLines; }
    if (*psz == '\n') fNewLine = TRUE;
  }
  DALineIndex.Add(cChars);

  // close the file
  close(nFile);
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function returns the index of the line that contains
//		 the specified position.
UINT Source::GetLineFromPos ( UINT iPos )
{
  for(UINT i = 0; i <= cLines; ++i)
    if (iPos <= DALineIndex[i + 1])
      return i;

  return -1u;
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function returns the buffer index of the specified line.
UINT Source::GetLineIndex ( UINT iLine )
{
  return DALineIndex[iLine];
}


///////////////////////////////////////////////////////////////////////////////
// Description:	 This function translates the specification file
void Source::Translate ( )
{
  // set buffer pointers
  iPtr = 0;
  iRef = 0;
  iLine = 0;

  // allocate data structures needed for the scanner and parser
  // initialize the error routine
  ptt = new TokenTank(512);
  stackInput = new TokenStack(16);
  stackToken = new TokenStack(256);
  stackState = new UINTStack(256);
  symTable = new SymbolTable(397, TRUE);
  error->Initialize(this);

  // read the source file into the buffer
  ReadSourceFile();

  // translate the file and process the main procedure and equation statements
  if (pszBuf)
  { BOOL fOK = (uParse() == 0);
    if (fOK)
    { FileInputOutput fileIO;
      fileIO.nWriteRefineCode(szFileName);
      fileIO.nWriteSkeletonFile(szFileName);
    }
  }

  // show messages
  error->ShowMessages();

  // deallocate the scanner and parser data structures
  delete ptt;
  delete stackInput;
  delete stackToken;
  delete stackState;
  delete symTable;
  delete [] pszBuf;
}



