/*-----------------------------------------------------------------------------

 PROJECT:  SIMULATION GENERATION FRAMEWORK (SGFramework)

 SUBSYSTEM:  mesh.exe
 FILE:	     types.h
 AUTHOR:     Kevin M. Kramer

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 11/01/94    KMK      initial implementation

 DESCRIPTION:

 This module contains the symbol, math node and operator type defintions.

-----------------------------------------------------------------------------*/


#ifndef __types_h
#define __types_h


#define TYPE_RCONST	     1		// real constant
#define TYPE_ICONST	     2		// integer constant
#define TYPE_FUNC	     3		// function
#define TYPE_COORD	     4		// coordinate
#define TYPE_LABEL	     5		// edge or region label
#define TYPE_POINT	     6		// point
#define TYPE_EDGE	     7		// edge
#define TYPE_REGION	     8		// region
#define TYPE_REFINE	     9		// refine

#define SYMBOLS		    10		// number of symbol types
#define TYPE_ADD	    11		// addition math node
#define TYPE_SUB	    12		// subtraction math node
#define TYPE_MUL	    13		// multiplication math node
#define TYPE_DIV	    14		// division math node
#define TYPE_AND	    15		// logical and math node
#define TYPE_OR		    16		// logical or math node
#define TYPE_XOR	    17		// logical exclusive or math node
#define TYPE_GT		    18		// greater than math node
#define TYPE_GE		    19		// greater than and equal to math node
#define TYPE_LT		    20		// less than math node
#define TYPE_LE		    21		// less than and equal to math node
#define TYPE_EQ		    22		// equal to math node
#define TYPE_NE		    23		// not equal to math node
#define TYPE_NEG	    24		// negation math node
#define TYPE_NOT	    25		// logical not math node
#define TYPE_REAL	    26		// convert to real math node

#define FLAG_REAL	    0x0001	// real flag
#define FLAG_INT	    0x0002	// integer flag
#define FLAG_CONST	    0x0004	// constant flag
#define FLAG_VAR	    0x0008	// variable flag
#define FLAG_ZEROCHECK	    0x4000	// "0.0 *" check
#define FLAG_DIFF	    0x8000	// differentiation flag


#endif
