/*-----------------------------------------------------------------------------

 PROJECT:  SIMULATION GENERATION FRAMEWORK (SGFramework)

 SUBSYSTEM:  refine.exe
 FILE:	     global.h
 AUTHOR:     Kevin M. Kramer

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 11/04/94    KMK      initial implementation

 DESCRIPTION:

 This module contains type definitions and constants used by all of the
 SGLAT modules.

-----------------------------------------------------------------------------*/


#ifndef __typedefs_h
#define __typedefs_h


///////////////////////////////////////////////////////////////////////////////
////////////////////////////  TYPE DEFINITIONS	///////////////////////////////
///////////////////////////////////////////////////////////////////////////////


// boolean constants
#define TRUE			1
#define FALSE			0


// global constants
#define LEN_IDENT		15
#define LEN_FILENAME		128
#define GLOBAL_BUFSIZE		512
#define TOLERANCE		1.0e-3
#define EPS			1.0e-12
#define SCALE			1.732050807568877/2.0


// global type definitions
typedef int		  BOOL;				      // f
typedef unsigned int	  UINT;				      // u
typedef unsigned char	  UCHAR;			      // uch
typedef int		  I3ARRAY[3];			      // i3a
typedef void *		  PVOID;			      // pv
typedef int *		  PBOOL;			      // pf
typedef double		  REAL;				      // r
typedef double *	  PREAL;			      // pr
typedef void *		  PRF;				      // prf
typedef double		  (*PRF1)(double);		      // prf1
typedef double		  (*PRF2)(double, double);	      // prf2
typedef double		  (*PRF3)(double, double, double);    // prf3


// global buffer definition
extern char szGlobalBuf[GLOBAL_BUFSIZE];


/* definition of command line argument structure */
#define VISUAL_CONSTRUCT    1	    // visualize mesh construction
#define VISUAL_RATING	    2	    // visualize mesh rating
#define VISUAL_NONE	    3	    // no graphics
#define RES_640x480	    1	    /* 640 x 480 resolution */
#define RES_800x600	    2	    /* 800 x 600 resolution */
#define RES_1024x768	    3	    /* 1024 x 768 resolution */


// command line parser
typedef struct tagCMDARGS	    // cmdargs
{ char	 *pszAsciiMesh;		    // ascii mesh file name
  REAL	  MinAngle;		    // minimum angle
  REAL	  MaxRatio;		    // maximum rectangle ratio
  REAL	  MaxGrade;		    // maximum rectangle grade
  UINT	  cMinDivisions;	    // minimum number of divisions
  UINT	  cMaxDivisions;	    // maximum number of divisions
  UINT	  cDivAdj;		    // divide adjacency number
  UINT	  cQualEnh1;		    // number of quality enhancement iterations
  UINT	  cQualEnh2;		    // number of quality enhancement iterations
  UINT	  cQualEnh3;		    // number of quality enhancement iterations
  UINT	  uVisual;		    // visulation level
  BOOL	  fCylGeom;		    // cylindrical geometry flag
  BOOL	  fGeneratePSFile;	    // generate a postscript file
  BOOL	  fKeepAspect;		    // keep x and y aspect ratio
  BOOL	  fNoQualReport;	    // no quality report
  BOOL	  fNoPosRef;		    // do not use RS on elements above y = 0.
  int	  nResolution;		    // graphics screen resolution
  int	  cNoPress;		    // debug mode info
  int	  cPresses;		    // debug mode info
} CMDARGS;
extern CMDARGS cmdargs;


// global function definition
int nRefineMesh(UINT cArg, char *apszArg[], char *pszFileName);

/* global macros */
#ifdef max
#undef max
#endif
#ifdef min
#undef min
#endif
#define max(x,y)	    (((x) > (y)) ? (x) : (y))
#define min(x,y)	    (((x) < (y)) ? (x) : (y))

#if defined(BOR)
#define TYPESAFE_DOWNCAST(object,toClass) dynamic_cast<toClass *>(object)
#elif defined(GNU)
#define TYPESAFE_DOWNCAST(object,toClass) ((toClass *) (object))
#endif


#if defined(UNIX)
double pow10(double);
#define O_BINARY 0x0000
#define getch getchar
#endif


#endif
