#ifndef __sgfplot_h
#define __sgfplot_h

///////////////////////////////////////////////////////////////////////////////
///////////////  MISCELLANEOUS CONSTANTS, MACROS AND VARIABLES  ///////////////
///////////////////////////////////////////////////////////////////////////////


#define LEN_IDENT           15          // identifier length
#define LEN_FILENAME        128         // maximum file name length
#define MAX_NORM            20          // maximum norm degree
#define TYPE_RCONST         3           // real constant
#define TYPE_ICONST         4           // integer constant

typedef struct tagRESHEAD               // reshead
{ char szLogo[64];                      // text logo
  char szSign[16];                      // signature
  char szMeshFile[128];                 // name of mesh file
  UINT cConstant;                       // number of constants
  UINT cVariable;                       // number of variables
  UINT cArray;                          // number of arrays
  UINT cElement;                        // number of elements
  UINT c1DArray;                        // number of 1D arrays
  UINT c2DArray;                        // number of 2D arrays
  UINT c3DArray;                        // number of 3D arrays
  UINT cSet;                            // number of data sets
} RESHEAD;

typedef struct tagMESHHEAD
{ char szLogo[64];                      // text logo
  char szSign[16];                      // signature
  UINT cConstant;                       // number of constants
  UINT cLabel;                          // number of labels
  UINT cArray;                          // number of 1D arrays
  UINT cRegLabel;                       // number of region labels
  UINT cBndryNode;                      // number of boundary nodes
  UINT cNode;                           // number of nodes
  UINT cEdge;                           // number of edges
  UINT cTriangle;                       // number of triangles
  UINT fCylGeom;                        // cylindrical geometry flag
} MESHHEAD;

typedef struct tagXLATCONST
{ char szName[LEN_IDENT+1];             // constant name
  UINT uType;                           // constant type
  union                                 //
  { REAL r;                             // value of real constant
    int  n;                             // value of integer constant
  } data;
} XLATCONST;

typedef struct tagXLATVAR
{ char szName[LEN_IDENT+1];             // variable name
  UINT uOffset;                         // variable offset
} XLATVAR;

typedef struct tagXLATARRAY
{ char szName[LEN_IDENT+1];             // array name
  UINT uOffset;                         // array offset
  UINT cDim;                            // number of dimensions
  UINT acElements[3];                   // elements per dimensions
} XLATARRAY;

typedef struct tagXLATLABEL
{ char szName[LEN_IDENT+1];             // label name
  UINT iIndex;                          // label index
  UINT uType;                           // label type
  UINT cNode;                           // number of nodes
} XLATLABEL;

typedef UINT INODE;                     // inode
typedef UINT IEDGE;                     // iedge
typedef UINT ITRI;                      // itri

typedef struct tagNODE                  // node
{ REAL  x;                              // x coordinate
  REAL  y;                              // y coordinate
} NODE;

typedef struct tagEDGE                  // edge
{ UINT  uLabel;                         // edge label
  INODE inodeA;                         // index of node A
  INODE inodeB;                         // index of node B
} EDGE;

#define iedgeCD iedgeAC
#define itriCD  itriAC
typedef struct tagTRI                   // tri
{ UINT  uLabel;                         // region label
  IEDGE iedgeAB;                        // index of edge AB
  IEDGE iedgeBC;                        // index of edge BC
  IEDGE iedgeAC;                        // index of edge AC or CD(iedgeAD!=-1)
  IEDGE iedgeAD;                        // index of edge AD
  ITRI  itriAB;                         // index of tri. adj. to edge AB
  ITRI  itriBC;                         // index of tri. adj. to edge BC
  ITRI  itriAC;                         // index of tri. adj. to edge AC or CD
  ITRI  itriAD;                         // index of tri. adj. to edge AD
} TRI;

#endif
