///////////////////////////////////////////////////////////////////////////////
////////////////////////////  TYPE DEFINITIONS	///////////////////////////////
///////////////////////////////////////////////////////////////////////////////


// boolean constants
#define TRUE			1
#define FALSE			0

// global type definitions
typedef int		  BOOL;				      // f
typedef unsigned int	  UINT;				      // u
typedef unsigned char	  UCHAR;			      // uch
typedef void *		  PVOID;			      // pv
typedef int *		  PBOOL;			      // pf
typedef double		  REAL;				      // r
typedef double *	  PREAL;			      // pr


///////////////////////////////////////////////////////////////////////////////
///////////////	 MISCELLANEOUS CONSTANTS, MACROS AND VARIABLES	///////////////
///////////////////////////////////////////////////////////////////////////////


#define LEN_IDENT	    15		// identifier length
#define LEN_FILENAME	    128		// maximum file name length
#define MAX_NORM	    20		// maximum norm degree
#define TYPE_RCONST	    3		// real constant
#define TYPE_ICONST	    4		// integer constant

typedef struct tagRESHEAD		// reshead
{ char szLogo[64];			// text logo
  char szSign[16];			// signature
  char szMeshFile[128];			// name of mesh file
  UINT cConstant;			// number of constants
  UINT cVariable;			// number of variables
  UINT cArray;				// number of arrays
  UINT cElement;			// number of elements
  UINT c1DArray;			// number of 1D arrays
  UINT c2DArray;			// number of 2D arrays
  UINT c3DArray;			// number of 3D arrays
  UINT cSet;				// number of data sets
} RESHEAD;

typedef struct tagMESHHEAD
{ char szLogo[64];			// text logo
  char szSign[16];			// signature
  UINT cConstant;			// number of constants
  UINT cLabel;				// number of labels
  UINT cArray;				// number of 1D arrays
  UINT cRegLabel;			// number of region labels
  UINT cBndryNode;			// number of boundary nodes
  UINT cNode;				// number of nodes
  UINT cEdge;				// number of edges
  UINT cTriangle;			// number of triangles
  UINT fCylGeom;			// cylindrical geometry flag
} MESHHEAD;

typedef struct tagXLATCONST
{ char szName[LEN_IDENT+1];		// constant name
  UINT uType;				// constant type
  union					//
  { REAL r;				// value of real constant
    int	 n;				// value of integer constant
  } data;
} XLATCONST;

typedef struct tagXLATVAR
{ char szName[LEN_IDENT+1];		// variable name
  UINT uOffset;				// variable offset
} XLATVAR;

typedef struct tagXLATARRAY
{ char szName[LEN_IDENT+1];		// array name
  UINT uOffset;				// array offset
  UINT cDim;				// number of dimensions
  UINT acElements[3];			// elements per dimensions
} XLATARRAY;

typedef struct tagXLATLABEL
{ char szName[LEN_IDENT+1];		// label name
  UINT iIndex;				// label index
  UINT uType;				// label type
  UINT cNode;				// number of nodes
} XLATLABEL;

typedef UINT INODE;			// inode
typedef UINT IEDGE;			// iedge
typedef UINT ITRI;			// itri

typedef UINT AINODE[4];			// ainode

typedef struct tagNODE			// node
{ REAL	x;				// x coordinate
  REAL	y;				// y coordinate
} NODE;

typedef struct tagEDGE			// edge
{ UINT	uLabel;				// edge label
  INODE inodeA;				// index of node A
  INODE inodeB;				// index of node B
} EDGE;

#define iedgeCD iedgeAC
#define itriCD	itriAC
typedef struct tagTRI			// tri
{ UINT	uLabel;				// region label
  IEDGE iedgeAB;			// index of edge AB
  IEDGE iedgeBC;			// index of edge BC
  IEDGE iedgeAC;			// index of edge AC or CD
  IEDGE iedgeAD;			// index of edge AD
  ITRI	itriAB;				// index of tri. adj. to edge AB
  ITRI	itriBC;				// index of tri. adj. to edge BC
  ITRI	itriAC;				// index of tri. adj. to edge AC or CD
  ITRI	itriAD;				// index of tri. adj. to edge AD
} TRI;

typedef struct tagPNT3D			// pnt3D
{ REAL x;				// x coordinate
  REAL y;				// y coordinate
  REAL z;				// z coordinate
} PNT3D;


typedef struct tagPNT2D			// pnt2D
{ int x;				// x coordinate
  int y;				// y coordinate
} PNT2D;


class TriPlot3D
{ private:
    UINT cNodes;			// number of nodes
    UINT cTriangles;			// number of triangles
    AINODE *aainode;			// node index of triangles
    PNT3D  *apnt3dInput;		// input data
    PNT3D  *apnt3dXform;		// transformed data
    PNT3D  *apnt3dMesh;			// transformed mesh data
    PNT2D  *apnt2d;			// screen data - surface plot
    PNT2D  *apnt2dMesh;			// screen data - mesh
    UINT   *aiPlotOrder;		// plotting order

  private:
    void Xform3D_3D (PNT3D &, PNT3D &); // 3d to 3d transformation
    void Xform3D_2D (PNT3D &, PNT2D &); // 2d to 3d transformation
    void XformMesh  (PNT3D &, PNT3D &); // 3d to 3d transformation for mesh
    void DeterminePlottingOrder ();	// determine the plotting order

  private:
    REAL AZ;				// azimuthal rotation angle
    REAL EL;				// elevation rotation angle
    REAL Cxx, Cxy, Cxz;			// coefficients of xform matrix
    REAL Cyx, Cyy, Cyz;			// coefficients of xform matrix
    REAL Czx, Czy, Czz;			// coefficients of xform matrix
    PNT3D pnt3dMin;			// minimum values
    PNT3D pnt3dMax;			// maximum values
    PNT3D pnt3dScale;			// scaling values

  public:
    UINT nResFactor;
    UINT MAXx;
    UINT MAXy;
    UINT cx;
    UINT cy;
    UINT cxBorder;
    UINT cyBorder;
    PNT3D pnt3dMinXform;		// minimum values after transformation
    PNT3D pnt3dMaxXform;		// maximum values after transformation
    PNT3D pnt3dScaleXform;		// scaling values after transformation

  public:
    void ReadMeshFile (char *);		// read the mesh file
    void ReadResultFile (char *);	// read the result file
    void GetViewAspect ();		// prompt user for view aspect
    void InitWinDim ();
    void Xform3DData ();		// transform data
    void Xform2DData ();		// transform data
    void DeterminePlotOrder ();		// determine plot order

  public:
    UINT GetMaxX ()			{ return MAXx; }
    UINT GetMaxY ()			{ return MAXy; }
    void PlotMesh ();			// plot mesh
    void PlotData ();			// plot data

  public:
    TriPlot3D (char *);
   ~TriPlot3D ();
};
