/*-----------------------------------------------------------------------------

 PROJECT:  SIMULATION GENERATION FRAMEWORK (SIMGEN)

 SUBSYSTEM:  mesh.exe
 FILE:	     dynarray.cpp
 AUTHOR:     Kevin M. Kramer

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 11/02/94    KMK      initial implementation


 DESCRIPTION:

 This module implements dynamic arrays of pointers.

-----------------------------------------------------------------------------*/


///////////////////////////////////////////////////////////////////////////////
/////////////////////////  DYNAMIC ARRAY OF POINTERS  /////////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Function:	 DynArrayOfPtr::DynArrayOfPtr
// Scope:	 Public, Constructor
template <class Type, UINT u, UINT f>
DynArrayOfPtr<Type, u, f>::DynArrayOfPtr ()
{
  cElements = 0;
  fManaged = f;
  dynarray = (Type**) calloc((cSize = uInc = u), sizeof(Type*));
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 DynArrayOfPtr::~DynArrayOfPtr
// Scope:	 Public, Destructor
template <class Type, UINT u, UINT f>
DynArrayOfPtr<Type, u, f>::~DynArrayOfPtr ()
{
  // if the array is managed then deallocate the stored objects
  if (fManaged)
    for(UINT i = 0; i < cElements; ++i)
      delete dynarray[i];

  free((void *) dynarray);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 DynArrayOfPtr::GetElements
// Scope:	 Public
// Description:	 This function returns the number of elements in the dynamic
//		 array.
template <class Type, UINT u, UINT f>
inline UINT DynArrayOfPtr<Type, u, f>::GetElements ()
{
  return cElements;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 DynArrayOfPtr::Add
// Scope:	 Public
// Description:	 This function adds an element to the dynamic array increases
//		 the array size if necessary.
template <class Type, UINT u, UINT f>
UINT DynArrayOfPtr<Type, u, f>::Add (Type *t)
{
  // increase the array size if necessary
  if (cElements == cSize)
  { cSize += uInc;
    dynarray = (Type **) realloc(dynarray, cSize * sizeof(Type *));
  }

  dynarray[cElements] = t;
  return cElements++;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 DynArrayOfPtr::Flush
// Scope:	 Public
// Description:	 This function adds an element to the dynamic array increases
//		 the array size if necessary.
template <class Type, UINT u, UINT f>
void DynArrayOfPtr<Type, u, f>::Flush ()
{
  // if the array is managed then deallocate the stored objects
  if (fManaged)
    for(UINT i = 0; i < cElements; ++i)
      delete dynarray[i];

  // reset the element count
  cElements = 0;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 DynArrayOfPtr::operator[]
// Scope:	 Public
// Description:	 This function overloads the [] operator.
template <class Type, UINT u, UINT f>
Type * DynArrayOfPtr<Type, u, f>::operator[] (UINT i)
{
  return dynarray[i];
}



///////////////////////////////////////////////////////////////////////////////
///////////////////////////////	 DYNAMIC ARRAY	///////////////////////////////
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Function:	 DynArray::DynArray
// Scope:	 Public, Constructor
template <class Type, UINT u>
DynArray<Type, u>::DynArray ()
{
  cElements = 0;
  dynarray = (Type *) calloc((cSize = uInc = u), sizeof(Type));
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 DynArray::~DynArray
// Scope:	 Public, Destructor
template <class Type, UINT u>
DynArray<Type, u>::~DynArray ()
{
  free((void *) dynarray);
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 DynArray::GetElements
// Scope:	 Public
// Description:	 This function returns the number of elements in the dynamic
//		 array.
template <class Type, UINT u>
inline UINT DynArray<Type, u>::GetElements ()
{
  return cElements;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 DynArray::Add
// Scope:	 Public
// Description:	 This function adds an element to the dynamic array increases
//		 the array size if necessary.
template <class Type, UINT u>
UINT DynArray<Type, u>::Add (Type t)
{
  // increase the array size if necessary
  if (cElements == cSize)
  { cSize += uInc;
    dynarray = (Type *) realloc(dynarray, cSize * sizeof(Type));
  }

  dynarray[cElements] = t;
  return cElements++;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 DynArray::Flush
// Scope:	 Public
// Description:	 This function adds an element to the dynamic array increases
//		 the array size if necessary.
template <class Type, UINT u>
void DynArray<Type, u>::Flush ()
{
  cElements = 0;
}


///////////////////////////////////////////////////////////////////////////////
// Function:	 DynArray::operator[]
// Scope:	 Public
// Description:	 This function overloads the [] operator.
template <class Type, UINT u>
Type DynArray<Type, u>::operator[] (UINT i)
{
  return dynarray[i];
}

