/*-----------------------------------------------------------------------------

 PROJECT:  SIMULATION GENERATION FRAMEWORK (SIMGEN)

 SUBSYSTEM:  mesh.exe
 FILE:	     dynarray.h
 AUTHOR:     Kevin M. Kramer

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 11/02/94    KMK      initial implementation

 DESCRIPTION:

 This module implements dynamic arrays of pointers.

-----------------------------------------------------------------------------*/


#ifndef __dynarray_h
#define __dynarray_h


template <class Type, UINT u, UINT f> class DynArrayOfPtr
{ private:
    UINT cSize;				// total number of elements
    UINT uInc;				// increment size
    UINT cElements;			// number of elements in array
    BOOL fManaged;			// managed array
    Type **dynarray;			// dynamic array

  public:
    UINT GetElements ();		// return number of elements in array
    UINT Add (Type *t);			// add an item to the array
    void Flush ();			// remove all pointers from the array
    Type * operator[](UINT i);		// array indexing

  public:
     DynArrayOfPtr ();
    ~DynArrayOfPtr ();

  private:
    DynArrayOfPtr(const DynArrayOfPtr&);	  // prevent copying
    DynArrayOfPtr& operator=(const DynArrayOfPtr&);
};


template <class Type, UINT u> class DynArray
{ private:
    UINT cSize;				// total number of elements
    UINT uInc;				// increment size
    UINT cElements;			// number of elements in array
    Type *dynarray;			// dynamic array

  public:
    UINT GetElements ();		// return number of elements in array
    UINT Add (Type t);			// add an item to the array
    void Flush ();			// remove all pointers from the array
    Type operator[](UINT i);		// array indexing

  public:
     DynArray ();
    ~DynArray ();

  private:
    DynArray(const DynArray&);		// prevent copying
    DynArray& operator=(const DynArray&);
};


#endif
