#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if defined(UNIX)
#include <unistd.h>
#define   O_BINARY 0x0000
#else
#include <io.h>
#endif

#include <fcntl.h>


///////////////////////////////////////////////////////////////////////////////
////////////////////////////  TYPE DEFINITIONS  ///////////////////////////////
///////////////////////////////////////////////////////////////////////////////


#define TRUE  1
#define FALSE 0

typedef int          BOOL;
typedef unsigned int UINT;
typedef int          I3ARRAY[3];

#if   defined(SPREC)
typedef float       REAL;
#elif defined(LPREC)
typedef long double REAL;
#else
typedef double      REAL;
#endif


///////////////////////////////////////////////////////////////////////////////
///////////////  MISCELLANEOUS CONSTANTS, MACROS AND VARIABLES  ///////////////
///////////////////////////////////////////////////////////////////////////////


#define SIGNATURE_RESULT    "@~!_RESULT_!~@"

#define LEN_IDENT      15    // identifier length

typedef struct tagRESHEAD    // reshead
{ char szLogo[64];           // text logo
  char szSign[16];           // signature
  char szMeshFile[128];      // name of mesh file
  UINT cConstant;            // number of constants
  UINT cVariable;            // number of variables
  UINT cArray;               // number of arrays
  UINT cElement;             // number of elements
  UINT c1DArray;             // number of 1D arrays
  UINT c2DArray;             // number of 2D arrays
  UINT c3DArray;             // number of 3D arrays
  UINT cSet;                 // number of data sets
} RESHEAD;

typedef struct tagXLATCONST
{ char szName[LEN_IDENT+1];  // constant name
  UINT uType;                // constant type
  union                      //
  { REAL r;                  // value of real constant
    int  n;                  // value of integer constant
  } data;
} XLATCONST;

typedef struct tagXLATVAR
{ char szName[LEN_IDENT+1];  // variable name
  UINT uOffset;              // variable offset
} XLATVAR;

typedef struct tagXLATARRAY
{ char szName[LEN_IDENT+1];  // array name
  UINT uOffset;              // array offset
  UINT cDim;                 // number of dimensions
  UINT acElements[3];        // elements per dimensions
} XLATARRAY;

#if defined(UNIX)

// compare two string w/o regard for case
int stricmp ( char *psz1, char *psz2 )
{
  char ch1, ch2;
  do
  { ch1 = *psz1;
    ch2 = *psz2;
    if ((ch1 >= 'a') && (ch1 <= 'z')) ch1 -= 'a' - 'A';
    if ((ch2 >= 'a') && (ch2 <= 'z')) ch2 -= 'a' - 'A';
    if      (ch1 > ch2) return +1;
    else if (ch1 < ch2) return -1;
    ++psz1;
    ++psz2;
  } while (ch1 && ch2);

  return 0;
}

#endif

///////////////////////////////////////////////////////////////////////////////
//
// *.res       char szLogo[64]
//             char szSign[16];              // signature
//             char szMeshFile[128];         // name of mesh file
//             UINT cConstant;               // number of constants
//             UINT cVariable;               // number of variables
//             UINT cArray;                  // number of arrays
//             UINT cElement;                // number of elements
//             UINT c1DArray;                // number of 1D arrays
//             UINT c2DArray;                // number of 2D arrays
//             UINT c3DArray;                // number of 3D arrays
//             UINT cSet;                    // number of data sets
//             ----------------------------------------------------
//                  cConstant * XLATCONST
//             ----------------------------------------------------
//                  cVariable * XLATVAR
//             ----------------------------------------------------
//                  cArray    * XLATARRAY
//             ----------------------------------------------------
//  set 0           cElement  * REAL
//  set 1           cElement  * REAL
//                  ............
//  set cset-1      cElement  * REAL
//             ----------------------------------------------------
//
void Extract ( double* ar, int iSet,
               XLATVAR* avar, XLATARRAY* aarray, UINT cVariable, UINT cArray )
{
  UINT i;
  char szFileName[32];

  if (cVariable)
  {
    sprintf(szFileName, "vars.%03d", iSet);
    FILE *file = fopen(szFileName, "wt");
    if (file)
    { fprintf(file, "Variables in the SGFramework Result File\n\n");
      for (i = 0; i < cVariable; ++i)
        fprintf(file, "%s = %11.4le\n", avar[i].szName, ar[avar[i].uOffset]);
    }
    fclose(file);
  }

  for (i = 0; i < cArray; ++i)
  {
    aarray[i].szName[8] = '\0';
    sprintf(szFileName, "%s.%03d", aarray[i].szName, iSet);
    // sprintf(szFileName, "%s.%05d", aarray[i].szName, iSet);

    FILE *file = fopen(szFileName, "wt");
    if (file)
    { UINT u  = aarray[i].uOffset;
      UINT c  = aarray[i].cDim;
      UINT c1 = aarray[i].acElements[0];
      UINT c2 = aarray[i].acElements[1];
      UINT c3 = aarray[i].acElements[2];

      // -------------------- 1 dimensional array --------------------
      if (c == 1)
      { for (UINT j = 0; j < c1; ++j)
          fprintf(file, "%11.4le\n", ar[u + j]);
      }

      // -------------------- 2 dimensional array --------------------
      else if (c == 2)
      { for(UINT j = 0; j < c1; ++j)
        { for(UINT k = 0; k < c2; ++k)
            fprintf(file, "%11.4le ", ar[u + j * c2 + k]);
          fprintf(file, "\n");
        }
      }

      // -------------------- 3 dimensional array --------------------
      else if (c == 3)
      { for(UINT j = 0; j < c1; ++j)
        { for(UINT k = 0; k < c2; ++k)
          { for(UINT l = 0; l < c3; ++l)
              fprintf(file, "%11.4le ", ar[u + (j * c2 + k) * c3 + l]);
            fprintf(file, "\n");
          }
          fprintf(file, "\n");
        }
      }

      fclose(file);
    }
  }
}


///////////////////////////////////////////////////////////////////////////////
//////////////////////////////  MAIN PROCEDURE  ///////////////////////////////
///////////////////////////////////////////////////////////////////////////////


int main ( int c, char *apsz[] )
{
  if ((c < 2) || (c > 3) || ((c == 2) && !strcmp(apsz[1], "-h")))
  { fprintf(stdout, "usage: extract filename [#|all]\n");
    fprintf(stdout, "       where filename is a result file (*.res)\n");
    fprintf(stdout, "       # is data set to extract (default = 1)\n");
    return 1;
  }

  fprintf(stdout, "SG Data Extractor   Version 1.0  (c) 1997  K. Kramer\n");
  fprintf(stdout, "Filename:  %s\n", apsz[1]);

  int handle = open(apsz[1], O_RDONLY | O_BINARY);
  if (handle == -1)
  { fprintf(stdout, "\ncannot open file \"%s\"\n", apsz[1]);
    return 2;
  }

  RESHEAD head;
  int nResult = read(handle, &head, sizeof(RESHEAD));
  if ((nResult == -1) || strcmp(head.szSign, SIGNATURE_RESULT))
  { fprintf(stdout, "\nNot a SGFramework result file\n");
    close(handle);
    return 3;
  }

  UINT cConstant = head.cConstant;
  UINT cVariable = head.cVariable;
  UINT cArray    = head.cArray;
  UINT cElement  = head.cElement;
  UINT cSet      = head.cSet;

  int fAll = 0;
  int iSet = 0;
  if (c == 3)
  { char str[] = "all";
    if (!stricmp(apsz[2],str))
      fAll = 1;
    else
    { iSet = atoi(apsz[2]) - 1;
      if ((iSet < 0) || (iSet >= cSet))
      { fprintf(stdout, "data set must be between 1 and %u\n", cSet);
        return 3;
      }
    }
  }

  fprintf(stdout, "Number of Constants:   %d\n", cConstant);
  fprintf(stdout, "Number of Variables:   %d\n", cVariable);
  fprintf(stdout, "Number of Arrays:  %d\n", cArray);
  fprintf(stdout, "Number of Variables and Array Elements:  %d\n", cElement);
  fprintf(stdout, "Number of Data Sets:   %d\n", cSet);

  XLATCONST *aconst = NULL;
  XLATVAR *avar = NULL;
  XLATARRAY *aarray = NULL;
  REAL *ar = NULL;

  if (cConstant) aconst = new XLATCONST[cConstant];
  if (cVariable)   avar = new XLATVAR[cVariable];
  if (cArray)    aarray = new XLATARRAY[cArray];
  if (cElement)      ar = new REAL[cElement];

  if (cConstant) read(handle, aconst, cConstant * sizeof(XLATCONST));
  if (cVariable) read(handle, avar, cVariable * sizeof(XLATVAR));
  if (cArray)    read(handle, aarray, cArray * sizeof(XLATARRAY));

  if (cElement)
  { if (!fAll)
    { long lOffset = (long) iSet * (long) cElement * (long) sizeof(REAL);
      lseek(handle, (off_t) lOffset, SEEK_CUR);
      read(handle, ar, cElement * sizeof(REAL));
      Extract(ar, iSet, avar, aarray, cVariable, cArray);
    }
    else
    { for(iSet = 0; iSet < cSet; ++iSet)
      { read(handle, ar, cElement * sizeof(REAL));
        Extract(ar, iSet, avar, aarray, cVariable, cArray);
      }
    }
  }
  close(handle);

  if (aconst) delete [] aconst;
  if (avar)   delete [] avar;
  if (aarray) delete [] aarray;
  if (ar)     delete [] ar;
  return 0;
}
