/*-----------------------------------------------------------------------------

 PROJECT:  SIMULATION GENERATION FRAMEWORK (SGFramework)

 FILE:	     da_uint.h
 AUTHOR:     Kevin M. Kramer

 REVISION HISTORY
 DATE	   INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 03/12/94    KMK      initial implementation
 03/25/98    JJF      modification

 DESCRIPTION:

 This module implements dynamic arrays of unsigned integers.  It is for
 compilers that have problems with templates.

-----------------------------------------------------------------------------*/


#ifndef __da_uint_h
#define __da_uint_h


class DAUINT
{ private:
    UINT cSize;		      // total number of elements
    UINT uInc;		      // increment size
    UINT cElements;	      // number of elements in array
    UINT *dynarray;	      // dynamic array
    void IncSize ()	      { if (cElements == cSize)
				{ cSize += uInc;
				  dynarray = (UINT *)
                                  realloc(dynarray, cSize*sizeof(UINT));
				}
			      }

  public:
    UINT GetElements ()	      { return cElements;    }
    UINT Add (UINT t)	      { IncSize();
                                dynarray[cElements] = t;
                                return cElements++;  }
    void RemoveLast ()	      { --cElements;	     }
    void Flush ()	      { cElements = 0;	     }
    UINT *GetPointer (UINT i) { return dynarray + i; }
    UINT operator[]  (UINT i) { return dynarray[i];  }

  public:
     DAUINT (UINT c = 4096)
     { cElements = 0; uInc = cSize = c; 
       dynarray = (UINT *) calloc((size_t) c, sizeof(UINT));
     }
    ~DAUINT ()
     { free((void *) dynarray);
     }
};


#endif

