/*-----------------------------------------------------------------------------

 PROJECT:  SIMULATION GENERATION FRAMEWORK (SGFramework)

 SUBSYSTEM:  sgxlat.exe mesh.exe refine.lib(cmdparse.o)
 FILE:       cmdparse.cpp
 AUTHOR:     Kevin M. Kramer
             J. Joshua Feng

 REVISION HISTORY
 DATE      INITIALS   DESCRIPTION
 --------  --------   --------------------------------------------------------
 10/19/94    KMK      initial implementation
 03/31/97    JJF      debug (deallocation)

 DESCRIPTION:

 This module contains the command line parsing class.

-----------------------------------------------------------------------------*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// the following hdrstop = terminates the list of header files that are
// eligible for precompilation
#ifndef UNIX
#pragma hdrstop
#endif

#include "termnlgy.h"
#include "cmdparse.h"


///////////////////////////////////////////////////////////////////////////////
// Scope:        Public, Constructor
//
CmdParse::CmdParse ( UINT c, char **apsz )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("cmdparse.cpp CmdParse::CmdParse\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT i;

  cSwitch = cPos = 0;
  char **ppsz = apsz + 1;              // SKIP the very first argument
  for(i = 1; i < c; ++i, ++ppsz)
  { char ch = **ppsz;
    if ((ch == '-') || (ch == '/')) ++cSwitch;
    else                            ++cPos;
  }

  apszSwitch = (cSwitch) ? new char*[cSwitch] : (char **) NULL;
  apszPos    = (cPos)    ? new char*[cPos]    : (char **) NULL;
  char **ppszSwitch = apszSwitch;
  char **ppszPos    = apszPos;
  ppsz = apsz + 1;
  for(i = 1; i < c; ++i, ++ppsz)
  { char ch = **ppsz;
    if ((ch == '-') || (ch == '/')) { *ppszSwitch = *ppsz;  ++ppszSwitch; }
    else                            { *ppszPos    = *ppsz;  ++ppszPos;    }
  }
}


///////////////////////////////////////////////////////////////////////////////
// Scope:        Public, Destructor
//
CmdParse::~CmdParse ( )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("cmdparse.cpp CmdParse::~CmdParse\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if (cSwitch) delete [] apszSwitch;
  if (cPos)    delete [] apszPos;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:        Public
// Description:  This function returns true if the specified switch argument
//               exists.
//
BOOL CmdParse::fArgExist ( char *psz )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("cmdparse.cpp CmdParse::fArgExist\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT cLen = strlen(psz);
  char **ppsz = apszSwitch;
  for(UINT i = 0; i < cSwitch; ++i, ++ppsz)
    if (!strncmp(psz, *ppsz + 1, cLen)) return TRUE;
  return FALSE;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:        Public
// Description:  This function returns the specified switch argument if it
//               exists; otherwise the default value is returned.
//
int CmdParse::nGetSwitchArg ( char *psz, int nDef, int nMin, int nMax )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("cmdparse.cpp CmdParse::nGetSwitchArg\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT cLen = strlen(psz);
  char **ppsz = apszSwitch;
  for(UINT i = 0; i < cSwitch; ++i, ++ppsz)
    if (!strncmp(psz, *ppsz + 1, cLen))
    { int n = atoi(*ppsz + strlen(psz) + 1);
      if (n < nMin) n = nMin;
      if (n > nMax) n = nMax;
      return n;
    }
  return nDef;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:        Public
// Description:  This function returns the specified position argument if it
//               exists; otherwise the default value is returned.
//
int CmdParse::nGetPosArg ( UINT i, int nDef, int nMin, int nMax )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("cmdparse.cpp CmdParse::nGetPosArg\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if (i < cPos)
  { int n = atoi(apszPos[i]);
    if (n < nMin) n = nMin;
    if (n > nMax) n = nMax;
    return n;
  }
  return nDef;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:        Public
// Description:  This function returns the specified switch argument if it
//               exists; otherwise the default value is returned.
//
REAL CmdParse::rGetSwitchArg ( char *psz, REAL rDef, REAL rMin, REAL rMax )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("cmdparse.cpp CmdParse::rGetSwitchArg\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT cLen = strlen(psz);
  char **ppsz = apszSwitch;
  for(UINT i = 0; i < cSwitch; ++i, ++ppsz)
    if (!strncmp(psz, *ppsz + 1, cLen))
    { REAL r = atof(*ppsz + strlen(psz) + 1);
      if (r < rMin) r = rMin;
      if (r > rMax) r = rMax;
      return r;
    }
  return rDef;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:        Public
// Description:  This function returns the specified position argument if it
//               exists; otherwise the default value is returned.
//
REAL CmdParse::rGetPosArg ( UINT i, REAL rDef, REAL rMin, REAL rMax )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("cmdparse.cpp CmdParse::rGetPosArg\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  if (i < cPos)
  { REAL r = atof(apszPos[i]);
    if (r < rMin) r = rMin;
    if (r > rMax) r = rMax;
    return r;
  }
  return rDef;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:        Public
// Description:  This function returns the specified switch argument if it
//               exists; otherwise the default value is returned.
//
char * CmdParse::pszGetSwitchArg ( char *psz, char *pszDef )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("cmdparse.cpp CmdParse::pszGetSwitchArg\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  UINT cLen = strlen(psz);
  char **ppsz = apszSwitch;
  for(UINT i = 0; i < cSwitch; ++i, ++ppsz)
    if (!strncmp(psz, *ppsz + 1, cLen))
      return *ppsz + cLen + 1;
  return pszDef;
}


///////////////////////////////////////////////////////////////////////////////
// Scope:        Public
// Description:  This function returns the specified position argument if it
//               exists; otherwise the default value is returned.
//
char * CmdParse::pszGetPosArg ( UINT i, char *pszDef )
{
#if defined(DEBUG)
  FILE *fileDEBUG = fopen("debug.out", "at");
  fputs("cmdparse.cpp CmdParse::pszGetPosArg\n", fileDEBUG);
  fclose(fileDEBUG);
#endif

  return (i < cPos) ? apszPos[i] : pszDef;
}
